// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
// 
// Copyright (C) Andreas Nicolai, Andreas.Nicolai@gmx.net
//               http://home.arcor.de/ghorwin
//               http://www.bauklimatik-dresden.de/privat/nicolai
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#ifndef IBK_algorithm_h
#define IBK_algorithm_h

#include <algorithm>
#include <string>
#include <sstream>

#include "IBK_messages.h"

/// The namespace IBK contains all functions/classes/data types of the IBK 
/// library.
/// The IBK library is written in ISO C++ and can be used with most modern 
/// compilers. It has been testet with Borland C++ Builder 5.0 and GCC 3.x.
/// The name IBK stands for Institut fuer Bauklimatik (Institute for building 
/// climatology) of the TU Dresden (Technical University of Dresden) in 
/// Germany, where most of this library was coded.
namespace IBK {

/// This algorithm copies only those elements of the range 'first','last' to the range 
/// starting with 'result' if the unary predicate pred(element) yields true.
template <typename inIt, typename outIt, typename predicate>
outIt copy_if(inIt first, inIt last, outIt result, predicate pred) {
    while (first!=last) {
        if (pred(*first))  *result++ = *first;
        ++first;
    }
    return result;
}

/// This algorithm copies only those elements of the range 'first','last' to the range 
/// starting with 'result' if they are equal to 'value'. 
template <typename inIt, typename outIt, typename T>
outIt copy_if_equal(inIt first, inIt last, outIt result, const T& value) {
    while (first!=last) {
        if (*first == value)  *result++ = *first;
        ++first;
    }
    return result;
}

/// This algorithm extracts the iterators to those elements of the range 
/// 'first', 'last', who are fulfilling the unary predicate pred.
/// Remember that depending on the container type you are using the copied 
/// iterators are invalidated when the number of elements in the container
/// changes. Therefore this function should not be used to create permanent 
/// index tables. 
template <typename inIt, typename outIt, typename predicate>
outIt iterator_copy_if(inIt first, inIt last, outIt result, predicate pred) {
    while (first != last) {
        if (pred(*first))  *result++ = first;
        ++first;
    }
    return result;
}

/// This algorithm calls the clear() member function in all elements of the
/// given range.
template <typename inIt>
void clear(inIt first, inIt last) {
	while (first != last)
		(first++)->clear();
}

/// This algorithm advances an iterator similarly to the std::advance algorithm 
/// and returns the advances iterator.
template <typename inIt>
inIt advance_iterator(inIt it, std::size_t distance) {
	std::advance(it, distance);
	return it;
}

/// @brief For each element in the range [first,last] a new element is appened to the
///        container 'cont' and the member function read() is called with the element
///        as parameter.
///
/// Remember to include the type used in the container before including delphin_algorithm.h!
template <typename const_iterator, typename container>
void read_range(const_iterator first, const_iterator last, container& cont) {
    while (first!=last) {
        typename container::value_type t;  // create new element
        try {
        	t.read(*first++);         // read data from string, might throw
        }
        catch (std::exception & ex) {
        	ERROR_MSG(ex.what());
			throw IBK_exception(IBK::format_string("Error reading entry %!") % (cont.size() + 1), "[read_range]" );
        }
        cont.push_back(t);        // if successful add to container
    }
}

/// @brief Calls the member function write(ostream, int indent) in all elements in the
///        range [first,last].
///
/// Remember to include the declarations for iterator_traits<const_it>::value_type (or
/// in other words the declaration of the type first and last are pointing to) before
/// delphin_algorithm.h!
template <class const_it>
void write_range(const_it first, const_it last, std::ostream& out, std::size_t indent) {
    while (first!=last)
        (first++)->write(out, static_cast<unsigned int>(indent));
}

/// @brief Calls the member function calcCurrentValue(t) for all elements in the range [first,last].
///
/// Remember to include the declarations for DelClimateCond before delphin_algorithm.h!
template <typename iterator>
void calc_current_climateconds(iterator first, iterator last, double t) {
    while (first!=last)
        (first++)->calcCurrentValue(t);
}

/// @brief Calls the member function assign_climate(climatelist) for each element in
///        the range [first,last].
///
/// Remember to include the declarations for DelClimateCond before delphin_algorithm.h!
template <typename iterator, typename cc_list_t>
void link_climateconds(iterator first, iterator last, cc_list_t& climatelist) {
    while (first!=last)
        (first++)->assignClimateCond(climatelist);
}

/// @brief Calls the member function assign_elements(elements) for each element in
///        the range [first,last].
///
/// Remember to include the declarations for DelElement before delphin_algorithm.h!
template <typename iterator, typename element_list_t>
void link_elements(iterator first, iterator last, element_list_t& elements) {
    while (first!=last)
        (first++)->assign_elements(elements);
}

/// @brief Appends indices like '[1]' with increasing numbers to the basename until no
///        matching names are no longer found in the given range.
///
/// Example:
/// @code
///     string base_name = "temperature output";  // the base name
///     vector<string>  existing_names;           // a vector with existing names
///     // .. assume the vector is filled with values,
///     // like "temperature output", "temperature output [1]", "relative humidity output"
///     string new_name = pick_name(base_name, existing_names.begin(), existing_names.end());
///     // new_name will be "temperature output [2]"
/// @endcode
template <typename iterator, typename const_iterator>
std::string pick_name(const std::string& basename, iterator first, const_iterator last) {
    if (first==last) return basename;
	std::string basename_mod = basename;
	std::string name;
	// remove [] if present
	int pos = basename_mod.rfind('[');
	while (pos != std::string::npos && pos > 0 && basename_mod[pos-1] == ' ') --pos;
	basename_mod = basename_mod.substr(0, pos);
	int i=0;
	do {
		std::stringstream namestrm;
		if (++i==1)  namestrm << basename_mod;
		else         namestrm << basename_mod << " [" << i << ']';
		name = namestrm.str();
	} while (std::find(first, last, name)!=last);
    return name;
}


/// @file IBK_algorithm.h
/// @brief Contains different independent algorithms that enlarge the set of
///        algorithms that come with the Standard Template Library.

/// @example algorithm_example.cpp
/// This is an example of how to use the IBK algorithms.

} // namespace IBK

#endif // IBK_algorithm_h

