// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
// 
// Copyright (C) 2003 Andreas Nicolai, Andreas.Nicolai@gmx.net
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#ifndef IBK_array_h
#define IBK_array_h

#include "IBK_configuration.h"

#include <stdexcept>
#include <algorithm>
#include "IBK_format.h"             // for format_string()
#include "IBK_message_strings.h"    // for the error messages
#include "IBK_exception.h"

namespace IBK {

/// The class template IBK::array is a replacement for the C-Array and provides 
/// a certain level of STL support.
/// The array can be used with standard C++ algorithms and should be used 
/// whenever the size of the array is known at compile time already. Since 
/// the array is created on the stack it will be faster than using the 
/// std::vector. If you define IBK_ENABLE_RANGE_CHECKING the operator[] will 
/// perform a range check and throw an exception if the index is out of range. 
/// Most of the IBK::array's is very similar to the BOOST:Array.
template <typename T, std::size_t N>
class array {
  public:
    typedef T               value_type;         ///< The data type of the values in the array.
    typedef T*              iterator;           ///< Iterator type to elements.
    typedef const T*        const_iterator;     ///< Constant iterator type to elements.
    typedef T&              reference;          ///< Reference type to elements.
    typedef const T&        const_reference;    ///< Constant reference type to elements.
    typedef std::size_t     size_type;          ///< Size type used in the array.
    typedef std::ptrdiff_t  difference_type;    ///< Difference type used in the array.
    
    /// Default constructor, creates an array whose elements are not initialised.
    array() {};
    /// Constructor which initialises all elements in the vector with the initial value 'value'.
    explicit array(const T& value) { std::fill(begin(), end(), value); };
    /// Constructor initialising the array with data from a range 'first' to 'last'.
    array(const_iterator first, const_iterator last) { std::copy(first, last, data_); };
    /// Returns a random access iterator to the first element
    iterator        begin()         { return data_; }
    /// Returns a random access iterator to the first element (read only)
    const_iterator  begin() const   { return data_; }
    /// Returns a random access iterator to the position after the last element
    iterator        end()           { return data_+N; }
    /// Returns a random access iterator to the position after the last element (read only)
    const_iterator  end() const     { return data_+N; }

#ifndef IBK_ENABLE_RANGE_CHECKING
    /// Returns the element with the index 'i'
    reference       operator[](size_type i)         { return data_[i]; }
    /// Returns the element with the index 'i' (read only)
    const_reference operator[](size_type i) const   { return data_[i]; }
#else
    /// Returns the element with the index 'i'.
    /// Throws an IBK_exception if the index is out of range.
    reference       operator[](size_type i);
    /// Returns the element with the index 'i' (read only)
    /// Throws an IBK_exception exception if the index is out of range.
    const_reference operator[](size_type i) const;
#endif // IBK_ENABLE_RANGE_CHECKING

    /// Return the reference to the first element in the array
    reference       front()                         { return data_[0]; }
    /// Return a const reference to the first element in the array.
    const_reference front() const                   { return data_[0]; }
    /// Return the reference to the last element in the array
    reference       back()                          { return data_[N-1]; }
    /// Return a const reference to the last element in the array.
    const_reference back() const                    { return data_[N-1]; }

    /// Returns the size of the array.
    static size_type size()                         { return N; }

    /// Swaps the content of the array with the other array 'arr'
    void swap (array<T,N>& arr) {
        std::swap_ranges(begin(), end(), arr.begin());
    }

    /// Returns a pointer to the storage.
    const T* data() const                           { return data_; }

    /// Assignment operator= with type conversion.
    /// The data type T2 must be convertible to T1 and the size of both arrays 
    /// has to be equal. The function returns a reference to the array.
    template <typename T2>
    array<T,N>& operator=(const array<T2,N>& rhs) {
        std::copy(rhs.begin(),rhs.end(), begin());
        return *this;
    }

  private:
    T data_[N]; // C-Array as storage member
};

#ifdef IBK_ENABLE_RANGE_CHECKING

template <typename T, std::size_t N>
typename array<T,N>::reference array<T,N>::operator[](size_type i) {
    if (i>=N)
    	throw IBK_exception(format_string(IBK_INDEX_OUT_OF_RANGE) % i % 0 % (N-1), "[IBK::array::operator[] ]");
    return data_[i];
}

template <typename T, std::size_t N>
typename array<T,N>::const_reference array<T,N>::operator[](size_type i) const {
    if (i>=N)
    	throw IBK_exception(format_string(IBK_INDEX_OUT_OF_RANGE) % i % 0 % (N-1), "[IBK::array::operator[] const]");
    return data_[i];
}

#endif // IBK_ENABLE_RANGE_CHECKING

}  // namespace IBK

/// @file IBK_array.h
/// @brief Contains the declaration of the template class array.

/// @example array_example.cpp
/// This small example demonstrates the use of the IBK::array in comparison to a normal C-Array.

#endif // IBK_array_h
