// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
//
// Copyright (C) 2004 Andreas Nicolai, Andreas.Nicolai@gmx.net
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// The band matrix solver code is mainly taken from the free CVODE
// solver library by Scott D. Cohen and Alan C. Hindmarsh @ LLNL

#ifndef IBK_bandmatrix_h
#define IBK_bandmatrix_h

#include <vector>
#include <stdexcept>
#include <numeric>
#include <utility>
#include <cmath>

#include "IBK_matrix.h"
#include "IBK_format.h"
#include "IBK_message_strings.h"
#include "IBK_exception.h"

namespace IBK {

/// A banded matrix implementation.
///
/// The implementation is aimed at easy and intuitive usage of the
/// matrix and internal storage. A decomposition and backsolving
/// algorithm is also provided. When creating or resizing a band
/// matrix you need to pass the number of elements in the matrix n and
/// the number of upper and lower off-diagonal elements mu and ml.
/// If you are using the decomposition algorithm you have to ensure
/// that there is enough storage space available, because the pivoting
/// requires a larger bandwidth than the original matrix has. The
/// parameter smu gives the storage upper bandwidth used for
/// the storage of the decomposed matrix and is calculated as in the
/// following example:
/// <pre>
/// // Defined size of matrix and number of off-diagonal elements
/// unsigned int n = 10;
/// unsigned int mu = 3;	// 3 upper off-diagonal elements
/// unsigned int ml = 2;	// 2 lower off-diagonal elements
/// // Compute the storage upper bandwidth needed for decomposition.
/// // If decomposition is not used, set msu = mu
/// unsigned int smu = std::min(n - 1, mu + ml);
/// // Now create a banded matrix of type double
/// banded_matrix&lt;double&gt; mat(n, mu, ml, smu);
/// // or resize an existing matrix object
/// mat.resize(n, mu, ml, msu);
/// </pre>
template <typename T>
class banded_matrix : public IBK::matrix<T> {
  public:
    typedef T           value_type;         ///< The type of the values.
    typedef T&          reference;          ///< Reference to one element.
    typedef const T&    const_reference;    ///< Constant reference to one element.
    typedef T*          iterator;           ///< Iterator to one element.
    typedef const T*    const_iterator;     ///< Constant iterator to one element.

    /// Default constructor (creates an empty matrix)
    banded_matrix() : n_(0), mu_(0), ml_(0), smu_(0), colsize_(0) {};
    /// Constructor, see class documentation for description of parameters.
    banded_matrix(std::size_t n, std::size_t mu, std::size_t ml, std::size_t smu)
        { resize(n, mu, ml, smu); };
	/// Copy-Constructor
	banded_matrix(const banded_matrix<T>& other);

    /// Returns an iterator to the first element of the continuous memory block
    iterator        begin()         { return data_[0]; };
    /// Returns a constant iterator to the first element of the continuous memory block
    const_iterator  begin() const   { return data_[0]; };
    /// Returns an iterator to the element after the last after the continuous memory block
    iterator        end()           { return data_[0]+n_*colsize_; };
    /// Returns a constant iterator to the element after the last after the continuous memory block
    const_iterator  end() const     { return data_[0]+n_*colsize_; };

    /// Returns a reference to the element at the coordinates [col,row].
    reference  operator() (std::size_t col, std::size_t row)
        { return data_[col][row-col+smu_]; }
    /// Returns a constant reference to the element at the coordinates [col,row].
    const_reference  operator() (std::size_t col, std::size_t row) const;
    /// Returns a reference to the element at the coordinates [col,row].
    /// If the coordinates are invalid (outside the bands or out of bounds) a runtime error
    /// exception will be thrown.
    reference  at(std::size_t col, std::size_t row);
    /// Returns a constant reference to the element at the coordinates [col,row].
    /// If the coordinates are outside the matrix a runtime error
    /// exception will be thrown. If the coordinates are outside the bands, the zero_element
    /// will be returned.
    const_reference  at(std::size_t col, std::size_t row) const;

    /// Resizes the matrix (all data will be lost).
    void resize(std::size_t n, std::size_t mu, std::size_t ml, std::size_t smu);
    /// Returns the number of rows in the matrix.
    std::size_t  rows() const { return n_; };
    /// Returns the number of columns in the matrix.
    std::size_t  cols() const { return n_; };
    /// Returns the upper band width
    std::size_t  mu() const { return mu_; };
    /// Returns the lower band width
    std::size_t  ml() const { return ml_; };
    /// Returns the extended upper band width
    std::size_t  smu() const { return smu_; };
    /// Clears the matrix.
    void clear();
    /// Fills the matrix with zeros
    void setZero() { std::memset(data_[0], 0, n_*colsize_*sizeof(T)); }
    /// Returns the raw data.
    std::vector<T*>& data() { return data_; }

    /// Performs a LU factorization of the matrix, storing the pivots in the vector 'pivots'
	/// and returns 0 if successful. If a zero element is found in the diagonal,
    /// the function returns the column/row of that element.
    int LU_fact(std::vector<unsigned int>& pivots);
    /// Solves the equation system: A x = rhs, using the pivoting vector 'pivots'
    /// from the factorisation, the solution is stored in the vector 'rhs'.
    void backsolve(std::vector<unsigned int>& pivots, std::vector<T>& rhs);

  private:
    // Return value for all elements that are not part of the bands of the matrix.
    static T  zero_element_;
    // order of the matrix
    std::size_t n_;
    // size of the upper diagonal band (without center diagonal)
    std::size_t mu_;
    // size of the lower diagonal band (without center diagonal)
    std::size_t ml_;
    // size of the upper diagonal band including space for factorisation
    std::size_t smu_;
    // number of elements per storage column
    std::size_t colsize_;
    // This vectors contains pointers to the data of the matrix, whereas the element
    // data_[0] points to a continuous memory block holding the actual data and the
    // other elements point to the begins of the columns of the matrix.
    std::vector<T*> data_;
};
// ---------------------------------------------------------------------------

template <typename T>
T banded_matrix<T>::zero_element_ = T();    // static member for a zero element
// ---------------------------------------------------------------------------

template <typename T>
banded_matrix<T>::banded_matrix(const banded_matrix<T>& other) {
	resize(other.n_, other.mu_, other.ml_, other.smu_);
	// copy data over to other matrix
	//std::copy(other.data_[0], other.data_[0]+(n_*colsize_), data_[0]);
	memcpy(data_[0], other.data_[0], sizeof(double)*n_*colsize_);
}
// ---------------------------------------------------------------------------

template <typename T>
typename banded_matrix<T>::const_reference  banded_matrix<T>::operator()(std::size_t col, std::size_t row) const {
    if (col>=n_ || row >=n_)
        throw IBK_exception( format_string("Index [%,%] out of bounds!") % col % row, "[banded_matrix<T>::at]");
    if (col+ml_<row || col>row+mu_)  return zero_element_;
    else                             return data_[col][row-col+smu_];
}
// ---------------------------------------------------------------------------

template <typename T>
typename banded_matrix<T>::reference  banded_matrix<T>::at(std::size_t col, std::size_t row) {
    if (col>=n_ || row >=n_)
        throw IBK::IBK_exception(IBK::format_string("Index [%,%] out of bounds!") % col % row, "[banded_matrix<T>::at]" );
    if (col+ml_<row || col>row+mu_)
        throw IBK::IBK_exception(IBK::format_string("Index [%,%] outside the band!") % col % row, "[banded_matrix<T>::at]" );
    return data_[col][row-col+smu_];
}
// ---------------------------------------------------------------------------

template <typename T>
typename banded_matrix<T>::const_reference  banded_matrix<T>::at(std::size_t col, std::size_t row) const {
    if (col>=n_ || row >=n_)
        throw IBK::IBK_exception(IBK::format_string("Index [%,%] out of bounds!") % col % row, "[banded_matrix<T>::at]");
    if (col+ml_<row || col>row+mu_)  return zero_element_;
    else                             return data_[col][row-col+smu_];
}
// ---------------------------------------------------------------------------

template <typename T>
void banded_matrix<T>::resize(std::size_t n, std::size_t mu, std::size_t ml, std::size_t smu) {
    if (!data_.empty())  clear();
    data_.resize(n);
    n_ = n;
    mu_ = mu;
    ml_ = ml;
    smu_ = smu;
    colsize_ = smu + ml + 1;
    try {
        // reserve continuous memory block
        data_[0] = new double[n*colsize_];
        // store pointers to columns
        for (unsigned int i=1; i<n; ++i)
            data_[i] = data_[0] + i*colsize_;
    }
    catch (...) {
        // allocation failed, reset matrix to empty matrix
        n_ = ml_ = mu_ = smu_ = colsize_ = 0;
        data_.clear();
        throw;
    }
};
// ---------------------------------------------------------------------------

template <typename T>
void banded_matrix<T>::clear() {
    if (data_.size() > 1)
        delete [] data_[0];
    data_.clear();
};
// ---------------------------------------------------------------------------

template <typename T>
int banded_matrix<T>::LU_fact(std::vector<unsigned int>& pivots) {
    unsigned int n_1 = static_cast<unsigned int>(n_) - 1;
	if (pivots.size() != n_)  pivots.resize(n_);
    unsigned int *p = &pivots[0];
    T ** a = &data_[0];

    // zero out the first smu - mu rows of the rectangular array a
    unsigned int num_rows = static_cast<unsigned int>(smu_ - mu_);
    if (num_rows > 0) {
        for (unsigned int c=0; c<n_; ++c) {
            T * a_c = a[c];
            for (unsigned int r=0; r<num_rows; ++r)
	            a_c[r] = zero_element_;
        }
    }

    // k = elimination step number
    for (unsigned int k=0; k<n_1; ++k, ++p) {
        T * col_k               = a[k];
        T * diag_k              = col_k + smu_;
        T * sub_diag_k          = diag_k + 1;

        unsigned int last_row_k = std::min<unsigned int>(n_1,k+ml_);

        // find l = pivot row number
        unsigned int l=k;
        T max = std::fabs(*diag_k);
        T * kptr = sub_diag_k;
        for (unsigned int i=k+1; i<=last_row_k; ++i, ++kptr)
            if (std::fabs(*kptr) > max) {
	            l=i;
	            max = std::fabs(*kptr);
            }

        unsigned int storage_l = l-k+smu_;  // ROW(l,k,smu) : calculate row index in column
        *p = l;                             // store row of pivot element

        // check for zero pivot element
        if (col_k[storage_l] == zero_element_)  return(k+1);

        // swap a(l,k) and a(k,k) if necessary
        bool swap = (l != k);
        if (swap)
            std::swap(col_k[storage_l], *diag_k);

        // Scale the elements below the diagonal in
        // column k by -1.0 / a(k,k). After the above swap,
        // a(k,k) holds the pivot element. This scaling
        // stores the pivot row multipliers -a(i,k)/a(k,k)
        // in a(i,k), i=k+1, ..., MIN(n-1,k+ml).

        T mult = -1.0 / (*diag_k);
        kptr = sub_diag_k;
        for (unsigned int i=k+1; i<=last_row_k; ++i, ++kptr)
            if (*kptr != zero_element_)
                *kptr *= mult;

        // row_i = row_i - [a(i,k)/a(k,k)] row_k, i=k+1, ..., MIN(n-1,k+ml)
        // row k is the pivot row after swapping with row l.
        // The computation is done one column at a time,
        // column j=k+1, ..., MIN(k+smu,n-1).

        T * col_j;
        unsigned int last_col_k = std::min<unsigned int>(k+smu_, n_1);
        unsigned int storage_k;
        T a_kj;
        for (unsigned int j=k+1; j<=last_col_k; ++j) {
            col_j = a[j];
            storage_l = l-j+smu_; // ROW(l,j,smu)
            storage_k = k-j+smu_; // ROW(k,j,smu)
            a_kj = col_j[storage_l];

            // Swap the elements a(k,j) and a(k,l) if l!=k.
            if (swap) {
	            col_j[storage_l] = col_j[storage_k];
	            col_j[storage_k] = a_kj;
            }

            // a(i,j) = a(i,j) - [a(i,k)/a(k,k)]*a(k,j)
            // a_kj = a(k,j), *kptr = - a(i,k)/a(k,k), *jptr = a(i,j)

            if (a_kj != 0.0) {
                kptr=sub_diag_k;
                // pointer to the element in row k+1 and column j: ROW(k+1,j,smu)
                T * jptr = col_j + k+1-j+smu_;
	            for (unsigned int i=k+1; i<=last_row_k; ++i, ++kptr, ++jptr)
	                *jptr += a_kj * *kptr;
            }
        }
    }

    // set the last pivot row to be n-1 and check for a zero pivot
    *p = n_1;
    if (a[n_1][smu_] == 0.0)    return static_cast<int>(n_);
    else                        return 0;
}
// ---------------------------------------------------------------------------


template <typename T>
void banded_matrix<T>::backsolve(std::vector<unsigned int>& pivots, std::vector<T>& rhs) {
    // get direct access to data memory in vectors
    unsigned int n_1 = static_cast<unsigned int>(n_)-1;
	if (pivots.size() != n_)
    	throw IBK::IBK_exception("Pivoting vector size mismatch!", "[banded_matrix::backsolve]");
    unsigned int *p = &pivots[0];
    T *b = &rhs[0];
    T **a = &data_[0];

    // Solve Ly = Pb, store solution y in b
    for (unsigned int k=0; k < n_1; ++k) {
        unsigned int l = p[k];
        T mult = b[l];
        if (l != k) {
            b[l] = b[k];
            b[k] = mult;
        }
        T *diag_k = a[k]+smu_;
        unsigned int last_row_k = std::min<unsigned int>(n_1,k+ml_);
        for (unsigned int i=k+1; i<=last_row_k; ++i)
            b[i] += mult * diag_k[i-k];
    }

    // Solve Ux = y, store solution x in b
    for (int k=n_1; k>=0; --k) {
        T* diag_k = a[k]+smu_;      // pointer to the diagonal element of column/row k
        int first_row_k = std::max<int>(0, k-static_cast<int>(smu_));
        b[k] /= (*diag_k);
        T mult = -b[k];
        for (int i=first_row_k; i<=k-1; ++i)
            b[i] += mult*diag_k[i-k];
    }
}
// ---------------------------------------------------------------------------

} // namespace IBK

#endif // IBK_bandmatrix_h
