// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
//
// Copyright (C) 2004-2006 Andreas Nicolai, Andreas.Nicolai@gmx.net
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#ifndef IBK_block_tridiag_matrix_h
#define IBK_block_tridiag_matrix_h

#include <vector>
#include <cstdlib>

#include "IBK_matrixmath.h"

namespace IBK {

/// Implementation of a block tridiagonal matrix with optimized matrix decomposition function.
template <typename T, std::size_t M>
class block_tridiag_matrix {
  public:
	/// Enumeration values used to identify the band in the matrix.
    enum band_t {
    	LOWER, // = 0,
        MAIN,  // = 1,
        UPPER  // = 2
    };

	/// Constructor, creates an empty matrix, use resize() to specify the dimensions.
	block_tridiag_matrix() : size_(0) {}
    /// Constructor (creates a matrix of dimension n) with submatrices of dimension m
    explicit block_tridiag_matrix(std::size_t n) { resize(n); }
    /// Constructor (creates a matrix of dimension n and initialises all elements with the value 'value')
    block_tridiag_matrix(std::size_t n, const T& value);
	/// Destructor.
	~block_tridiag_matrix();

	/// Resizes the matrix.
	void resize(std::size_t n);

	/// Returns a pointer to the submatrix data.
	/// The submatrix data is organized as column based linear storage array. Access
	/// an element of the submatrix using the static function element() or
	/// use submatrix_ptr[col*M + row]
	/// @param row	The row of the submatrix.
	/// @param band	The band from which the submatrix should be
	///             taken: 0 = mean lower band, 1 = main diagonal, 2 = upper band
	T * submatrix(std::size_t row, band_t band);

    static T& element(T * data, int col, int row)  { return data[col*M + row]; }

    /// Returns the size of the matrix.
    std::size_t	size() const { return size_; }
    /// Returns the size of sub-matrices.
    static std::size_t	subsize() { return M; }
	/// Returns the whole matrix as it is stored.
    std::vector< std::vector<T*> >& data() { return data_; }

    /// Solves the equation system A x = rhs, where A is the block-tridiag-matrix,
    /// rhs is a vector of size(M*n) and x is the solution.
    /// The result will be stored in rhs.
    void solve_equation_system(std::vector< T >& rhs);

	/// Helper function for testing.
	void randomize();

  private:
    // do not allow copying of the matrix
	block_tridiag_matrix(const block_tridiag_matrix&);

  	// The data storage member:
    // data_[0] contains the lower diagonal data
    // data_[1] contains the main diagonal data
    // data_[2] contains the upper diagonal data
    // The submatrices are stored in linear memory blocks.
	std::vector< T* >				data_[3];
	std::size_t						size_;			// The dimension of the block tridiag matrix.
};
// ---------------------------------------------------------------------------

template <typename T, std::size_t M>
block_tridiag_matrix<T,M>::block_tridiag_matrix(std::size_t n, const T& value) {
	size_ = n;
    // loop over all bands
	for (unsigned int i=0; i<3; ++i) {
		data_[i].resize(size_); // vectors contain pointers to submatrices
        for (unsigned int j=0; j<size_; ++j) {
        	data_[i][j] = new double[M*M]; // create submatrices dynamically and store pointers
            std::fill(data_[i][j], data_[i][j]+M*M, value);
        }
    }
}
// ---------------------------------------------------------------------------

template <typename T, std::size_t M>
block_tridiag_matrix<T,M>::~block_tridiag_matrix() {
	for (unsigned int j=0; j<size_; ++j) {
		delete[] data_[LOWER][j];
		delete[] data_[MAIN][j];
		delete[] data_[UPPER][j];
    }
}
// ---------------------------------------------------------------------------

template <typename T, std::size_t M>
void block_tridiag_matrix<T,M>::resize(std::size_t n) {
	size_ = n;
	// get rid of all old data and create new
	for (unsigned int i=0; i<3; ++i) {
		for (unsigned int j=0; j<data_[i].size(); ++j)
			delete[] data_[i][j];
		// resize data vector to new size
		data_[i].resize(size_);
		// create data storage
        for (unsigned int j=0; j<size_; ++j)
        	data_[i][j] = new double[M*M]; // create submatrices dynamically and store pointers
	}
}
// ---------------------------------------------------------------------------

template <typename T, std::size_t M>
T * block_tridiag_matrix<T,M>::submatrix(std::size_t row, band_t band) {
#ifdef IBK_RANGE_CHECKING
	if (band < 0 || band > 2)
    	throw IBK::IBK_exception("Invalid band specification!", "[block_tridiag_matrix::submatrix]");
	if (row > size_)
    	throw IBK::IBK_exception("Index out of range!", "[block_tridiag_matrix::submatrix]");
#endif // IBK_RANGE_CHECKING
	return data_[band][row];
}
// ---------------------------------------------------------------------------

template <typename T, std::size_t M>
void block_tridiag_matrix<T,M>::randomize() {
	std::srand(time(NULL));
	// create data for the main diagonal
	for (std::size_t i=0; i<size_; ++i) {
		for (unsigned int k=0; k<M*M; ++k) {
			data_[MAIN][i][k]  = -floor((std::rand()*50.0)/((double)RAND_MAX));
			data_[LOWER][i][k] = -floor((std::rand()*40.0)/((double)RAND_MAX)-20);
			data_[UPPER][i][k] = -floor((std::rand()*40.0)/((double)RAND_MAX)-20);
		}
	}
}


#ifndef IBK_TEMPLATE_OPTIMIZATION
template <typename T, std::size_t M>
void block_tridiag_matrix<T,M>::solve_equation_system(std::vector<T>& rhs) {
	// Block-Tridiagonal solution algorithm
    // Refer to article "An efficient implementation of a block-tridiagonal matrix"
    // or Wikipedia page for details.

	// helper variables
    std::vector<T>  workspace_vec(M*M);
    double * workspace = &workspace_vec[0];
    std::vector<unsigned int> pivots_vec(M);
    unsigned int * pivots = &pivots_vec[0];

	// notation:
	//   a - lower diagonal elements
	//   b - main diagonal elements
	//   c - upper diagonal elements

	// variables used:
	//   a_i, b_i and c_i are pointers to the data section of
	//   the data vectors
	//   *a_i, *b_i and *c_i yield pointers to the submatrix data
	//   d_i is a pointer to the begin of a subvector in the
	//   right hand side vector rhs
	//
	//   obtaining values:
	//   **a_i gives the value at index 0,0 of the submatrix a
	//   *d_i gives the first value of the subvector
	double ** a_i = &(data_[LOWER][0]);
	double ** b_i = &(data_[MAIN][0]);
	double ** c_i = &(data_[UPPER][0]);
	double * d_i = &rhs[0];

    // 1. first row
    // 1.1. compute b0^-1 by decomposing the matrix
    matrix_decomposition(M, *b_i, pivots, workspace);
    // 1.2. c0' = b0^-1 * c0
    matrix_matrix_inverse_mult(M, *b_i, pivots, *c_i);
    // 1.3. d0' = b0^-1 * d0
    matrix_backsolve(M, *b_i, pivots, d_i);

    // 2. loop over all remaining rows
    int last = size_ - 1;
    for (int i=1; i<last; ++i) {
    	// advance pointers
    	++a_i;
    	++b_i;
    	++c_i;
    	d_i += M; // advance pointer to next subvector in rhs vector

        // 2.1 b_i' = b_i - a_i * c_(i-1)
        matrix_matrix_multsub(M, *a_i, *(c_i-1), *b_i);
        // 2.2 calculate b_i'^-1 by decomposition
    	matrix_decomposition(M, *b_i, pivots, workspace);
    	// 2.3 c_i' = b_i'^-1 * c_i
    	matrix_matrix_inverse_mult(M, *b_i, pivots, *c_i);
        // 2.4 d_i - a_i * d_(i-1) -> store in d_i
        matrix_vector_multsub(M, *a_i, d_i-M, d_i);
        // 2.5 d_i' = b_i'^-1 * (d_i - a_i * d_(i-1) )
        matrix_backsolve(M, *b_i, pivots, d_i);
    }
    // advance address pointers so that i = n-1
	++a_i;
    ++b_i;
    ++c_i;
    d_i += M;

    // 3. Last equation
    // 3.1 b_(n-1)' = b_(n-1) - a_(n-1) * c_(n-2)
    matrix_matrix_multsub(M, *a_i, *(c_i-1), *b_i);
    // 3.2 calculate b_(n-1)'^-1 by decomposition
    matrix_decomposition(M, *b_i, pivots, workspace);
    // 3.3 d_(n-1) - a_(n-1) * d_(n-2) -> store in d_(n-1)
    matrix_vector_multsub(M, *a_i, d_i - M, d_i);
    // 3.4 d_(n-1)' = b_(n-1)'^-1 * (d_(n-1) - a_(n-1) * d_(n-2) )
    matrix_backsolve(M, *b_i, pivots, d_i);
    // d_(n-1) now contains already x_(n-1)

    // 4. Backsolving loop
    for (int i=last; i>0; --i) {
    	// decrease pointers
		--b_i;
		--c_i;
		d_i -= M;
    	// 4.1 d_i = x_i = d_i' - c_i' x_(i+1)
    	matrix_vector_multsub(M, *c_i, d_i+M, d_i);
    }

    // TODO : return error code if decomposition failed somewhere
}
// ---------------------------------------------------------------------------
#endif // IBK_TEMPLATE_OPTIMIZATION

} // namespace IBK

#endif // IBK_block_tridiag_matrix_h
