// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
// 
// Copyright (C) 2003 Andreas Nicolai, Andreas.Nicolai@gmx.net
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#include "IBK_configuration.h"

#ifdef __BORLANDC__
  #include "IBK_prehead.h"
  #pragma hdrstop
#else
  #include <fstream>
  #include <sstream>
  #include <iostream>
  #include <algorithm>
#endif
#include "IBK_algorithm.h"
#include "IBK_fileutils.h"
#include "IBK_stringutils.h"
#include "IBK_messages.h"
#include "IBK_format.h"
using namespace std;

// Include header files for directory creation depending on compiler used
#ifdef __GNUC__
  #include <sys/stat.h>
  #include <sys/types.h>
#endif

#if defined(__MINGW32__) || defined(_MSC_VER) || defined(__BORLANDC__)
  #include <cstdlib>
  #include <cstdio>
  #include <direct.h>
  #include <sys/types.h>
  #include <sys/stat.h>
#elif __GNUC__
  #include <sys/types.h> 
  #include <sys/stat.h> 
  #include <unistd.h> 
  #include <errno.h>
#endif

namespace IBK {

dir_exists_t directory_exists(const std::string& dirname) {
#if defined(__MINGW32__) || defined(_MSC_VER) || defined(__BORLANDC__)
	// implementation for MS VC
	struct _stat buf;
	int res = _stat(dirname.c_str(), &buf);
	if (res == -1) return NO_DIR;
	// check if it is a directory
	if (buf.st_mode & _S_IFDIR) return DIR_EXISTS;
	else						return IS_FILE;
#elif __GNUC__
	struct stat buf;
	int res = stat(dirname.c_str(), &buf);
	if (res == -1) {
		switch (errno) {
			case EACCES : ERROR_MSG("Cannot access file/directory '"+dirname+"'!"); break;
		}
		return NO_DIR;
	}
	// check if it is a directory
	if (S_ISDIR(buf.st_mode))	return DIR_EXISTS;
	else						return IS_FILE;
	
#else
#error "Implement this for your compiler!"
#endif 
}

bool create_directory(const std::string& dirname) {
#if defined(__MINGW32__) || defined(_MSC_VER) || defined(__BORLANDC__)
	_mkdir(dirname.c_str());
#elif __GNUC__
	int res = mkdir(dirname.c_str(), 0755);
	if (res != 0) {
		switch (errno) {
			case EACCES : ERROR_MSG("Cannot create directory '"+dirname+"', access denied!"); break;
			case EEXIST : ERROR_MSG("Cannot create directory '"+dirname+"', it alreay exists (boggle)!"); break;
			case ENOTDIR : ERROR_MSG("Cannot create directory '"+dirname+"', invalid path!"); break;
			case ENOENT : ERROR_MSG("Cannot create directory '"+dirname+"', some component of the path doesn't exist!"); ;break;
			case EFAULT : ERROR_MSG("EFAULT"); break;
			case ELOOP : ERROR_MSG("ELOOP"); break;
			case ENAMETOOLONG : ERROR_MSG("ENAMETOOLONG"); break;
			case ENOMEM : ERROR_MSG("ENOMEM"); break;
		}
		cout << errno << endl;
		return false;
	}
#else
    #error Implement this for your compiler
#endif
	return directory_exists(dirname) == DIR_EXISTS;
}

bool move_file(const std::string& source, const std::string& target) {
#if defined(__MINGW32__) || defined(__BORLANDC__) || defined(_MSC_VER) || defined(__GNUC__)
	remove(target.c_str());
	// TODO : check if target file dissappeared
	rename(source.c_str(), target.c_str());
#else
    #error Write this for your compiler
#endif // __BORLANDC__
	// TODO : test if source file disappeared and target file is present
	return true;
}

void simplify_path(std::string& path) {
	// make sure we deal with /
//    std::replace(path.begin(), path.end(), '\\', '/');
	// look for .. in the string
    size_t pos = 0;
    size_t dots_pos;
    while ((dots_pos = path.find("/..", pos)) != string::npos) {
    	// try to find a slash prior to the position with the dots
        size_t slash_pos = path.rfind("/", dots_pos-1);
        if (slash_pos != string::npos) {
        	path = path.substr(0, slash_pos) + path.substr(dots_pos+3);
			pos = slash_pos;
        }
        else
        	pos = dots_pos + 3;
        if (pos >= path.size()) break;
    }
}

std::string replace_spaces(const std::string& path) {
    // for unix, replace spaces in path with escaped spaces
	string fname;
	size_t pos = path.find(" ");
	size_t lastpos = 0;
	while (pos != string::npos) {
		fname += path.substr(lastpos, pos-lastpos) + "\\ ";
		lastpos = pos+1;
		pos = path.find(" ", lastpos);
	}
	fname += path.substr(lastpos);
	return fname;
}

// ----------------------------------------------------------------------------
std::string replaceDir(const std::string& orig_line,
    const std::string& placeholder, const std::string& replace_str)
{
    string line = orig_line;
    size_t pos  = line.find(placeholder);
    if (pos != string::npos)
        line = line.substr(0, pos) + replace_str + line.substr(pos+placeholder.length()+1,string::npos);
    return line;
}
// ----------------------------------------------------------------------------
}  // namespace IBK

