// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
// 
// Copyright (C) 2006 Andreas Nicolai, Andreas.Nicolai@gmx.net
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#ifndef IBK_flag_h
#define IBK_flag_h

#include "IBK_configuration.h"

namespace IBK {

/// The class Flag represents a named boolean flag.
class Flag {
  public:
    /// Default constructor, creates a disabled (unnamed) flag with state 'false'.
    Flag() : state(false) {};

    /// Constructor to initialise the flag.
    Flag(const std::string& n, bool s) : name(n), state(s) {};

    /// Sets the flag state and whether it is enabled or not.
    void set(const std::string& n, bool s) { name = n; state = s; }

    /// Reads a flag from the stream 'in'.
    /// The format for reading is: name state<br>
    /// @param in       The input stream.
    void read(std::istream& in);

    /// Writes a flag into the output stream.
    /// The output format is: name state<br>
    /// The optional parameters can be used to specify the layout.
    /// @param out          The output stream.
    /// @param indent       (optional) Number of spaces, that should be put in front.
    /// @param width        (optional) Width of the flag name (exclusive indentation), 
    ///                                a " = " is always appended
    void write(std::ostream& out, unsigned int indent=0, unsigned int width=0) const;

    /// Clears the flag.
    void clear() { name.clear(); state = false; };

	/// Returns the state of the flag.
	bool operator()() { return state; }

    // ****** member variables *************************************************

    /// The descriptive keyword/name of the parameter.
    std::string 	name;
    /// The state of the flag
    bool 			state;
};

}  // namespace IBK

/// @file IBK_flag.h
/// @brief Contains the declaration of the class Flag, a class for a named boolean.

#endif  // IBK_flag_h
