// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
//
// Copyright (C) 2003 Andreas Nicolai, Andreas.Nicolai@gmx.net
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#ifndef IBK_format_h
#define IBK_format_h

#include <string>
#include <sstream>
#include <iosfwd>

namespace IBK {

/// Class format_string helps assembling error/progress messages without
/// the use of the old fashioned and error prone C function sprintf.
/// Define a string to be formatted like "This is section % of file '%' with % entries".
/// Use the operator % to replace placeholders with data.<br>
/// Example:
/// <pre>
///     string filename = "inputfile.txt";
///     int line = 14;
///     throw IBK_exception( format_string("Error in line % of file '%'!") % line % filename );
/// </pre>
class format_string {
  public:
    /// Default constructor, creates an empty format string
    format_string() : m_format(""), m_insertIndex(0), m_data("") {};
    /// Constructor taking a format string as argument.
    format_string(const std::string& format) : m_format(format), m_insertIndex(0), m_data("") {};

    /// Sets a new format string of style "This is section % with % entries".
    /// The previously inserted data is deleted.
    void set_format(const std::string& format) { m_format = format; m_insertIndex=0; m_data=""; };

    /// Returns the filled in format string.
    std::string str() const;

    /// Replaces next % with content of data.
    template <typename T>
    format_string& operator%(const T& data);

    /// Specialization for type std::size_t (needed for VC8).
    format_string& operator%(std::size_t data) { return format_string::operator%<unsigned int>(static_cast<unsigned int>(data)); }

    /// Operator + for string concatenation
    std::string operator+(const std::string& s) const;

  private:
    std::string     m_format;       ///< The format string.
    unsigned int    m_insertIndex;  ///< The current parameter index.
    std::string     m_data;         ///< The already formatted data.
};

/// Stream output operator.
std::ostream& operator<<(std::ostream& strm, const format_string& fstr);

/// String concatenation operator
std::string operator+(const std::string& lhs, const format_string& rhs);

// Generic implementation
template <typename T>
format_string& format_string::operator%(const T& data) {
    size_t pos = m_format.find("%", m_insertIndex);
    if (pos==std::string::npos) return *this;
    m_data += m_format.substr(m_insertIndex, pos-m_insertIndex);
    std::stringstream strm;
    strm << data;
    m_data += strm.str();
    m_insertIndex = static_cast<unsigned int>(pos)+1;
    return *this;
};


} // namespace IBK

#endif  // IBK_format_h

