// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
// 
// Copyright (C) 2003 Andreas Nicolai, Andreas.Nicolai@gmx.net
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#ifndef IBK_functors_h
#define IBK_functors_h

#include <functional>
#include <string>
#include "IBK_keyword.h"
#include "IBK_constants.h"  // for NEAR_ZERO
#include "IBK_types.h"

#ifdef _MSC_VER
#define TOLOWER(x) tolower(x)
#else
#define TOLOWER(x) std::tolower(x)
#endif

namespace IBK {

/// Binary predicate that returns true when the characters are equal in lower case
class nocase_equal : public std::binary_function<bool, char, char> {
  public:
    bool operator() (char lhs, char rhs) { return TOLOWER(lhs)==TOLOWER(rhs); };
};

/// The unary function abs returns the absolute value of 'value'
/// (which only makes sense with numerical data types).
template <typename T>
struct abs {
    T operator()(const T& value) const { return (value > 0) ? value : -value; };
};

/// Tests if a is equal to b including a certain range for potential rounding errors.
inline bool near_equal(double a, double b) {
    return (a + NEAR_ZERO > b  &&  a < b + NEAR_ZERO);
};

/// Tests if a is less then b including a certain range for potential rounding errors.
inline bool near_le(double a, double b)     { return !near_equal(a, b) && a < b; };
/// Tests if a is less or equal to b including a certain range for potential rounding errors.
inline bool near_le_eq(double a, double b)  { return (a < b || a - NEAR_ZERO < b); };
/// Tests if a is greater then b including a certain range for potential rounding errors.
inline bool near_gr(double a, double b)     { return (a > b && !near_equal(a, b)); };
/// Tests if a is greater or equal to b including a certain range for potential rounding errors.
inline bool near_gr_eq(double a, double b)  { return (a > b || a + NEAR_ZERO > b); };


/// @file IBK_functors.h
/// @brief Contains some useful function objects.

/// @example functor_example.cpp
/// This is an example of how to use the function objects.

} // namespace IBK

#endif  // IBK_functors_h
