// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
// 
// Copyright (C) 2003 Andreas Nicolai, Andreas.Nicolai@gmx.net
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#ifndef IBK_keyword_h
#define IBK_keyword_h

#include <string>
#include <utility>  // for std::pair
#include <iosfwd>   // for the iostream prototypes
#include <functional>

namespace IBK {

/// The class Keyword encapsulates a keyword (consisting of the actual name, 
/// the category, the enumeration value, optional other keyword strings and an
/// optional description)
class Keyword {
  public:

	/// Unary predicate that returns 'true' if the passed keyword is of category 'cat'.
	/// Use this predicate as in the following example:
	/// <pre>
	///     vector<Keyword> kwvec; 
	///     // fill vector with keywords
	///     vector<Keyword> filtered_kwvec;
	///     IBK::copy_if(kwvec.begin(), kwvec.end(), back_inserter(filtered_kwvec),
	///         Keyword::has_category("SomeCategory") );
	/// </pre>
	class has_category : public std::unary_function<Keyword, bool> {
		std::string m_category;
	  public:
		has_category(const std::string& cat) : m_category(cat) {};
		bool operator() (const Keyword& kw) const { return kw.category == m_category; };
	};
	// ----------------------------------------------------------------------------
  
    /// Reads the keyword from a string. 
    void read(const std::string& keywordline);

    std::string     name;           ///< The keyword string.
    std::string     category;       ///< The keyword category (derived from the class it's defined in).
    int             index;          ///< The number of the enumeration value.
    std::string     other_ids;      ///< String containing all other (deprecated) ids (separated by spaces), that are also recognised as the keyword.
    std::string     description;    ///< Optional description of the keyword and it's meaning.
};

/// operator == for searching a keyword with a given category and name
inline bool operator==(const Keyword& lhs, const std::pair<std::string, std::string>& rhs) { return (lhs.category == rhs.first && lhs.name == rhs.second); };
/// operator != for searching a keyword with a given category and name
inline bool operator!=(const Keyword& lhs, const std::pair<std::string, std::string>& rhs) { return (lhs.category != rhs.first || lhs.name != rhs.second); };

/// removes the trailing '=' from the keyword string and the brackets '[' and ']' around it
std::string& trim_keyword(std::string& keyword);

/// @file IBK_keyword.h
/// @brief Contains the declaration of the class Keyword and the function trim_keyword().

} // namespace IBK

#endif // IBK_keyword_h
