// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
// 
// Copyright (C) 2003 Andreas Nicolai, Andreas.Nicolai@gmx.net
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#include "IBK_configuration.h"

#ifdef __BORLANDC__
  #include "IBK_prehead.h"
  #pragma hdrstop
#else
  #include <iostream>
  #include <fstream>
  #include <sstream>
  #include <stdexcept>
#endif
#include "IBK_keywordlistmanager.h"
#include "IBK_format.h"
#include "IBK_message_strings.h"
#include "IBK_messages.h"
#include "IBK_exception.h"
using namespace std;

namespace IBK {

KeywordListManager* KeywordList=NULL;

bool KeywordListManager::read(std::istream& in) {
    if (!in) return false;
    kwlist_.clear();
    return append(in);
}
// ---------------------------------------------------------------------------

bool KeywordListManager::append(std::istream& in) {
    if (!in) return false;
    string line;
    while (getline(in, line)) {
        Keyword kw;
        try {
            kw.read(line);
            kwlist_.push_back(kw);
        }
        catch (std::exception& ex) {
            MSG( string(ex.what()) +  " : skipped", MSG_WARNING, "[KeywordListManager::read]");
        }
    }
    return true;
}
// ---------------------------------------------------------------------------

int KeywordListManager::get_index(const std::string& category, const std::string& kw_name) const {
    bool no_such_category=true;
    // search for correct combination of category and keyword name
    for (KwListType::const_iterator pos=kwlist_.begin(); pos!=kwlist_.end(); ++pos) {
        if (pos->category==category) {
            no_such_category=false;
            if (pos->name==kw_name || pos->other_ids.find(kw_name)!=string::npos)  return pos->index;
        }
    }
    if (no_such_category) {
    	IBK_exception(format_string(IBK_NO_SUCH_CATEGORY) % category, "[KeywordListManager::get_index]");
    }
    throw IBK_exception(format_string(IBK_KEYWORD_NOT_FOUND) % kw_name % category, "[KeywordListManager::get_index]");
}
// ---------------------------------------------------------------------------

string KeywordListManager::get_name(const std::string& category, int index) const {
    bool no_such_category=true;
    // search for correct combination of category and index
    for (KwListType::const_iterator pos=kwlist_.begin(); pos!=kwlist_.end(); ++pos) {
        if (pos->category==category) {
            no_such_category=false;
            if (pos->index==index)  return  pos->name;
        }
    }
    if (no_such_category) {
    	IBK_exception(format_string(IBK_NO_SUCH_CATEGORY) % category, "[KeywordListManager::get_name]");
    }
    throw IBK_exception(format_string(IBK_INVALID_KEYWORD_INDEX) % index % category, "[KeywordListManager::get_name]");
}
// ---------------------------------------------------------------------------

const Keyword& KeywordListManager::get_keyword(const std::string& category, const std::string& kw_name) const {
    bool no_such_category=true;
    // search for correct combination of category and keyword name
    for (KwListType::const_iterator pos=kwlist_.begin(); pos!=kwlist_.end(); ++pos) {
        if (pos->category==category) {
            no_such_category=false;
            if (pos->name==kw_name || pos->other_ids.find(kw_name)!=string::npos)  return *pos;
        }
    }
    if (no_such_category) {
    	IBK_exception(format_string(IBK_NO_SUCH_CATEGORY) % category, "[KeywordListManager::get_keyword]");
    }
    throw IBK_exception(format_string(IBK_KEYWORD_NOT_FOUND) % kw_name % category, "[KeywordListManager::get_keyword]");
}
// ---------------------------------------------------------------------------

const Keyword& KeywordListManager::get_keyword(const std::string& category, int index) const {
    bool no_such_category=true;
    // search for correct combination of category and index
    KwListType::const_iterator pos=kwlist_.begin();
    for (KwListType::const_iterator pos=kwlist_.begin(); pos!=kwlist_.end(); ++pos) {
        if (pos->category==category) {
            no_such_category=false;
            if (pos->index==index)  return  *pos;
        }
    }
    if (no_such_category) {
    	IBK_exception(format_string(IBK_NO_SUCH_CATEGORY) % category, "[KeywordListManager::get_keyword]");
    }
    throw IBK_exception(format_string(IBK_INVALID_KEYWORD_INDEX) % index % category, "[KeywordListManager::get_keyword]");
}
// ---------------------------------------------------------------------------

bool KeywordListManager::keyword_exists(const std::string& category, const std::string& kw_name) const {
    for (KwListType::const_iterator pos=kwlist_.begin(); pos!=kwlist_.end(); ++pos)
        if (pos->category==category) {
            if (pos->name==kw_name || pos->other_ids.find(kw_name)!=string::npos) return true;
        }
    return false;
}
// ---------------------------------------------------------------------------

std::string KeywordListManager::description_for_keyword(const std::string& cat, const std::string& kw_name) {
    try {
        const Keyword& kw = get_keyword(cat, kw_name);
        if (kw.description.empty())  return kw_name;
        else                         return kw.description;
    }
    catch (...) {
        return kw_name;
    }
}
// ---------------------------------------------------------------------------

} // namespace IBK

