// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
// 
// Copyright (C) 2003 Andreas Nicolai, Andreas.Nicolai@gmx.net
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#ifndef IBK_keywordlistmanager_h
#define IBK_keywordlistmanager_h

#include <string>
#include <vector>
#include <list>
#include <iosfwd>
#include "IBK_keyword.h"
#include "IBK_algorithm.h"

namespace IBK {

/// Provides the functionality for converting a 'keyword' into the 
/// corresponding enumeration number. 
/// The keywordlist can be initialized from a string or optionally using
/// the keyword file (usually 'keyword.txt') which can be generated using 
/// the tool keylist_creator.
/// After initialising the keyword list you can freely convert keywords 
/// (given as strings) into the enumeration values and vice versa.
/// The keyword list file can be created with the keylist creator tool.
class KeywordListManager {
  public:
    /// Data type of the whole keyword list.
    typedef std::list<Keyword>                    KwListType;
    typedef std::list<Keyword>::iterator          KwListIterator;
    typedef std::list<Keyword>::const_iterator    KwListConstIterator;

/*
    ~KeywordListManager() {
    	std::cout << "[IBK::cleanup]  Destructing KeywordListManager!" << std::endl;
    }
*/

    /// Reads the keywords file 'kwfile' and thus initialises the keyword list.
    /// The keywords file should have been created by the keylist_creator tool 
    /// and each line if the keyword file should defined one keyword.
    /// If a syntactical error was encountered (keywords file corrupt) an
    /// IBK::IBK_exception will be thrown.
    /// @return Returns 'true', if the keywords file was read properly. The function
    ///         returns 'false', if the keywords file could not be found or openend.
    bool read(std::istream& in);
    bool append(std::istream& in);

    /// Returns the enumeration value of the keyword.
    /// The function searches in the keyword list for a keyword, that matches both
    /// the category and the keyword's id-string. This id-string can be either the keyword's
    /// name itself or one of the optional other ids stored for the keywords.
    /// If the ID-string or the category could not be found a std::IBK_exception
    /// exception will be thrown.
    /// @param category The keywords category, usually the full name of the enumeration type
    ///                 e.g. 'CClimateCond::condition_t'
    /// @param kw_name The keyword's ID-string, usually the name of the keyword.
    /// @return Returns the enumeration value of the keyword.
    int get_index(const std::string& category, const std::string& kwname) const;

    /// Returns the name (ID-string) of the keyword.
    /// The function searches in the keyword list for a keyword, that matches both
    /// the category and the keyword's enumeration value. 
    /// If the category could not be found or the enumeration value is too big 
    /// a std::IBK_exception exception will be thrown.
    /// @param category The keywords category, usually the full name of the enumeration type
    ///                 e.g. 'CClimateCond::condition_t'
    /// @param index The keyword's enumeration value.
    /// @return Returns the name (ID-string) of the keyword.
    std::string get_name(const std::string& category, int index) const;

    /// Returns a reference to the keyword entry in the list that matches both 'category' and 'kw_name'.
    /// Throws a IBK_exception exception if the keyword or the category cannot be found.
    const Keyword& get_keyword(const std::string& category, const std::string& kwname) const;

    /// Returns a reference to the keyword entry in the list that matches both 'category' and 'index'.
    /// Throws a IBK_exception exception if the keyword or the category cannot be found.
    const Keyword& get_keyword(const std::string& category, int index) const;

    /// Checks, whether a keyword with the name 'kw_name' exists in the given category.
    /// This function is a 'quiet' way to determine, whether a keyword exists or not. 
    /// Essentially it does the same as get_name() but does not throw an exception and 
    /// returns a boolean value instead.
    /// @return Returns 'true' when the keyword was found, otherwise 'false'.
    bool keyword_exists(const std::string& category, const std::string& kwname) const;

    /// Returns whether the keyword list is empty.
    std::size_t empty() const { return kwlist_.empty(); };
    /// Returns the number of the keywords in the list.
    std::size_t size() const { return kwlist_.size(); };

    /// Returns a constant reference to the full keyword list.
    const KwListType& get_list() const { return kwlist_; };

    /// Returns the description of the keyword 'kw_name' in the category 'cat' if available, 
    /// otherwise the keyword is returned.
    std::string description_for_keyword(const std::string& cat, const std::string& kwname);

    /// Returns a keyword list filtered by the predicate 'pred'.
    /// Only those keywords are copied in the resulting list for that the unary predicate 
    /// 'pred' returns true.
    template <typename Predicate>
    KwListType get_filtered_list(Predicate pred) const;

  private:
    KwListType kwlist_;
};

/// Gobal variable for the use of the keyword list (assuming that there will be only one
/// keyword list in the program).
/// Before the keyword list is used in the program is must be initialised by creating
/// an instance of the keyword list. Then it must be filled with keywords by reading 
/// the keyword list from file.
extern KeywordListManager* KeywordList;

template <typename Predicate>
KeywordListManager::KwListType KeywordListManager::get_filtered_list(Predicate pred) const {
    KwListType kwlist;
    IBK::copy_if(kwlist_.begin(), kwlist_.end(), std::back_inserter(kwlist), pred);
    return kwlist;
}

/// @file IBK_keywordlistmanager.h
/// Contains the class KeywordListManager which encapsulates the global keyword list and
/// thus provides the connection between keywords (text) and the enumeration values.

} // namespace IBK

#endif // IBK_keywordlistmanager_h
