// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
// 
// Copyright (C) 2003 Andreas Nicolai, Andreas.Nicolai@gmx.net
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#include "IBK_configuration.h"

#ifdef __BORLANDC__
  #include "IBK_prehead.h"
  #pragma hdrstop
#else
  #include <iostream>
  #include <sstream>
  #include <limits>
  #include <set>
  #include <functional>
  #include <stdexcept>
  #include <iomanip>
  #include <iterator>
#endif
#include "IBK_linearspline.h"
#include "IBK_algorithm.h"
#include "IBK_message_strings.h"
#include "IBK_debug.h"
#include "IBK_exception.h"
using namespace std;

namespace IBK {

bool LinearSpline::read(const std::string& x_data, const std::string& y_data, std::string& errorMsg) {
	stringstream xstrm(x_data);
    m_x.clear();
    copy(istream_iterator<double>(xstrm), istream_iterator<double>(), back_inserter(m_x) );
	stringstream ystrm(y_data);
    m_y.clear();
    copy(istream_iterator<double>(ystrm), istream_iterator<double>(), back_inserter(m_y) );
    if (m_x.empty()) {
		errorMsg = "Missing data to read!";
		return false;
	}
	if (m_x.size() != m_y.size()) {
    	errorMsg = "Number of x and y values differs!";
        m_x.resize(std::min(m_x.size(), m_y.size()));
        m_y.resize(std::min(m_x.size(), m_y.size()));
        return false;
	}
	return true;
}
// ---------------------------------------------------------------------------

void LinearSpline::read(const std::string& x_data, const std::string& y_data) {
	string errmsg;
	if (!read(x_data, y_data, errmsg))
		throw IBK_exception(errmsg, "[LinearSpline::read]");
}
// ---------------------------------------------------------------------------

/// Reads the spline from a stream (as vectors containing the x and y coordinates).
bool LinearSpline::read(std::istream& in, std::string& errorMsg) {
	string x_vec, y_vec;
    getline(in, x_vec);
    getline(in, y_vec);
    return read(x_vec, y_vec, errorMsg);
}
// ---------------------------------------------------------------------------

void LinearSpline::write(ostream& out, unsigned int indent) const {
	string istr = string(indent, ' ');
	out << istr;
	for (unsigned int i=0; i<m_x.size(); ++i)
    	out << setw(15) << left << m_x[i] << "  ";
    out << endl << istr;
	for (unsigned int i=0; i<m_y.size(); ++i)
    	out << setw(15) << left << m_y[i] << "  ";
    out << endl;
}
// ---------------------------------------------------------------------------

void LinearSpline::clear() {
    m_x.clear();
    m_y.clear();
    m_slope.clear();
}
// ---------------------------------------------------------------------------

void LinearSpline::swap(LinearSpline& spl) {
    m_x.swap(spl.m_x);
    m_y.swap(spl.m_y);
    m_slope.swap(spl.m_slope);
}
// ---------------------------------------------------------------------------

bool LinearSpline::makeSpline() {
	string errMsg;
    return makeSpline(errMsg);
}
// ---------------------------------------------------------------------------

bool LinearSpline::makeSpline(std::string& errMsg) {
	m_slope.clear();
    if (m_x.empty() || m_x.size() != m_y.size()) {
        errMsg = "[LinearSpline::makeSpline]  " + std::string(IBK_INVALID_VECTOR_DIMENSIONS);
        return false;
    }
    // check for strict monotonic increasing values, so if we find any
    // adjacent values that fulfil x[i] >= x[i+1], we have a problem
    if (std::adjacent_find(m_x.begin(), m_x.end(), std::greater_equal<double>())!=m_x.end()) {
        errMsg = "[LinearSpline::makeSpline]  " + std::string(IBK_X_MONOTONY_ERROR);
        return false;
    }
    for (unsigned int i=1; i<m_x.size(); ++i)
        m_slope.push_back( (m_y[i] - m_y[i-1])/(m_x[i]-m_x[i-1]) );
    return true;
}
// ---------------------------------------------------------------------------

double LinearSpline::value(double x) const {
	IBK_check(valid());
    if (!valid())
        throw IBK_exception("Spline not generated or invalid data!", "[LinearSpline::value]" );
    if (x <= m_x.front())  return m_y.front();
    if (x >= m_x.back())   return m_y.back();
    // m_xValues.size() = n    -> max i = n-1
    // m_slope.size() = n-1    -> max i = n-2
    // lower_bound ->  n = 0...n
	int i = static_cast<int>(std::lower_bound(m_x.begin(), m_x.end(), x)-m_x.begin() - 1);
    IBK_check(i >= 0);
    IBK_check(i < static_cast<int>(m_slope.size()));
    return m_y[i] + m_slope[i]*(x-m_x[i]);
}
// ---------------------------------------------------------------------------

double LinearSpline::non_interpolated_value(double x) const {
    if (m_x.empty())
        throw IBK_exception("X value vector is empty!", "[LinearSpline::non_interpolated_value]");
    if (x <= m_x.front())  return m_y.front();
    if (x >= m_x.back())   return m_y.back();
    // m_xValues.size() = n    -> max i = n-1
    // m_slope.size() = n-1    -> max i = n-2
    // lower_bound ->  n = 0...n
	int i = static_cast<int>(std::lower_bound(m_x.begin(), m_x.end(), x)-m_x.begin() - 1);
    IBK_check(i >= 0);
    IBK_check(i < static_cast<int>(m_slope.size()));
    return m_y[i];
}
// ---------------------------------------------------------------------------

double LinearSpline::slope(double x) const {
    if (m_x.empty())
        throw IBK_exception("X value vector is empty!","[LinearSpline::value]");
    // lower_bound -> i = 0...n   -> subtract 1 to get the slope index
    // max i = n-2 for slope
    size_t i = lower_bound(m_x.begin(), m_x.end(), x) - m_x.begin();
    if (i <= 1)                         return m_slope.front();
    else if (i < m_x.size()-1)    		return m_slope[--i];
    else                                return m_slope.back();
}
// ---------------------------------------------------------------------------

// here is the local stuff for the combination of linear splines
template <class BinaryOperation>
LinearSpline combine_spline(const LinearSpline& first, const LinearSpline& second,
    BinaryOperation op)
{
	// first insert all values into the set containing the x values
    std::set<double> x;
    std::copy(first.m_x.begin(), first.m_x.end(), inserter(x, x.begin()) );
    std::copy(second.m_x.begin(), second.m_x.end(), inserter(x, x.begin()) );
	// now create a vector for the y values
    std::vector<double> y(x.size());
    std::vector<double>::iterator yit = y.begin();
    for (std::set<double>::const_iterator xit=x.begin(); xit!=x.end(); ++xit, ++yit)
        *yit = op(first.value(*xit), second.value(*xit));
    return LinearSpline(x.begin(), x.end(), y.begin());
}
// ---------------------------------------------------------------------------


}	// namespace IBK

