// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
// 
// Copyright (C) 2003 Andreas Nicolai, Andreas.Nicolai@gmx.net
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#ifndef _IBK_linearspline_h_
#define _IBK_linearspline_h_

#include <vector>
#include <iosfwd>
#include <stdexcept>
#include <algorithm>
#include <functional>

#include "IBK_mathvector.h"

namespace IBK {

/// This class encapsulated two data vectors, which are used for linear interpolation.
/// The class LinearSpline can be thought of as a container of a tabulated function
/// (containing the data points of that function in separate vectors m_x and m_y).
/// In addition to the reading and writing functionality of such a container, it
/// provides the functionality to linearly interpolate between two data points using
/// the member function value(). In order to do that efficiently the slopes between
/// the points will be precalculated. The function makeSpline() essentially creates
/// this slopes vector and a call to this function is required if you want to use
/// the interpolation functionality of the spline. Also the first data vector must
/// be always strictly monotonic increasing if you want to use the interpolation
/// functions.
class LinearSpline {
  public:
    /// Default constructor, creates an empty spline.
    LinearSpline() {};

    /// Constructor, creates a linear spline out of two vectors.
    /// This function will test whether both vectors have the same length and 
    /// whether the xValues vector is strictly monotonic. If any of the requirements
    /// is violated an exception is thrown.
    LinearSpline(const std::vector<double> &x, const std::vector<double> &y)
		: m_x(x), m_y(y) { makeSpline(); }

    /// Constructor, creates a linear spline from two ranges.
    /// The function expects the range starting with firstY to be as long as the range
    /// firstX...lastX. If the monotonic criterion is violated, an exception is thrown.
    template <typename inputIteratorX, typename inputIteratorY>
    LinearSpline(inputIteratorX firstX, inputIteratorX lastX, inputIteratorY firstY);

	/// Reads the spline from two vectors containing the x and y coordinates.
	/// This function returns true or false thus indicating whether the reading was
	/// successful or not. If an error occurred, it stores an error message in the
	/// string errorMsg. 
	bool read(const std::string& x_data, const std::string& y_data, std::string& errorMsg);
	/// Reads the spline from two vectors containing the x and y coordinates.
	/// This function behaves very much like the function above, but instead of
	/// returning a value it throws an exception if an error occurred.
	void read(const std::string& x_data, const std::string& y_data);
	/// Reads the spline from a stream (as vectors containing the x and y coordinates).
	bool read(std::istream& in, std::string& errorMsg);
    /// Writes the spline into a stream as two data vectors (one for the x coordinates and one for
    /// the y coordinates).
    void write(std::ostream& out, unsigned int indent) const;
    /// Empties the spline.
    void clear();

	/// Swaps the content of a linear spline quickly
    void swap(LinearSpline& spl);
	/// Returns whether the spline is valid.
	bool valid() const { return !m_slope.empty(); }
	/// Creates the spline.
	bool makeSpline();
	/// Creates the spline.
	bool makeSpline(std::string& errMsg);

    /// Returns the size of the linear spline.
    std::size_t size()  const { return m_x.size(); };
    /// Returns whether the spline is empty.
    bool        empty() const { return m_x.empty(); };

    /// Returns an interpolated value y at a given point x.
    /// If x is outside the range of x value in the spline the first or last y value is returned respectively.
    double value(double x) const;
    /// Returns a non-interpolated value y at a given point x (a value at
    /// a given time point is valid, until the next time point comes).
    double non_interpolated_value(double x) const;
    /// Returns the slope at a given point x.
    double slope(double x) const;
    /// Returns a constant reference to a vector holding the slopes.
    const std::vector<double> &slopes()  const { return m_slope; };

    IBK::math_vector<double>	m_x;	///< The x values of the spline data table.
    IBK::math_vector<double>	m_y;	///< The y values of the spline data table.

  private:
    std::vector<double>		m_slope;	// The precalculated slopes for faster computation of the interpolated values.
};


/// This function combines two linear splines using the binary operation op and 
/// returns the resulting spline. 
///
/// The combination of splines works in a simple way. First the total range of
/// the splines x-values is obtained and all spline points in both splines inside
/// this range are processed. If a spline point doesn't exist on the other splines
/// range, its value is interpolated and the operation op is performed between
/// the actual value of one spline point and this interpolated point. 
/// The new spline will have first.size() &lt;= n &lt;= first.size() + second.size() 
/// data points.
template <class BinaryOperation>
LinearSpline combine_spline(const LinearSpline& first, const LinearSpline& second,
                                 BinaryOperation op);
/// This function adds two linear splines and returns the result.
/// Adding is done by adding the y-values (either tabulated or interpolated) for each x-value 
/// in both splines.
inline LinearSpline operator+(const LinearSpline& first, const LinearSpline& second) {
    return combine_spline(first, second, std::plus<double>());
}

/// This function subtracts two linear splines and returns the result.
/// The y-values of the second spline (either tabulated or interpolated) are subtracted from
/// the y-values of the first spline for each x-value in both splines.
inline LinearSpline operator-(const LinearSpline& first, const LinearSpline& second) {
    return combine_spline(first, second, std::minus<double>());
}

template <typename inputIteratorX, typename inputIteratorY>
LinearSpline::LinearSpline(inputIteratorX firstX, inputIteratorX lastX, inputIteratorY firstY) {
	while(firstX != lastX) {
    	m_x.push_back(*firstX++);
    	m_y.push_back(*firstY++);
    }
    makeSpline();
}

/// Comparison operator == between two linear splines.
/// Two splines are considered equal if the input data vectors m_x and m_y are equal
inline bool operator==(const LinearSpline& lhs, const LinearSpline& rhs) {
	return (lhs.m_x == rhs.m_x && lhs.m_y == rhs.m_y);
}

/// Comparison operator != between two linear splines.
inline bool operator!=(const LinearSpline& lhs, const LinearSpline& rhs) { return !(lhs==rhs); }


}  // namespace IBK


/// @file IBK_linearspline.h
/// @brief Contains the declaration of the class LinearSpline.

#endif // _IBK_linearspline_h_

