// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
//
// Copyright (C) 2006 Andreas Nicolai, Andreas.Nicolai@gmx.net
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#include "IBK_configuration.h"

#ifdef __BORLANDC__
  #include "IBK_prehead.h"
  #pragma hdrstop
#else
  #include <cmath>
  #include <stdexcept>
#endif
#include "IBK_math.h"
#include "IBK_exception.h"
using namespace std;

namespace IBK {

#ifdef IBK_ENABLE_SAFE_MATH

// Safe variant for std::pow function taking a double exponent
double f_pow(double base, double e, std::string * calling_fn_name) {
	// detect negative base error
	if (base <= 0) {
		throw IBK_exception("Negative base in f_pow() function!", 
			(calling_fn_name==NULL) ? string() : *calling_fn_name);
	}
	double log10b = std::log10(base);
	// detect overflow and throw exception
	if (e*log10b >= 300) {
		throw IBK_exception("  Overflow in f_pow() function!",
			(calling_fn_name==NULL) ? string() : *calling_fn_name);
	}
	return std::pow(base, e);
}

// Safe variant for std::pow function taking an integer exponent
double f_pow(double base, int e, std::string * calling_fn_name) {
	// detect negative base error
	if (base <= 0) {
		throw IBK_exception("Negative base in f_pow() function!",
			(calling_fn_name==NULL) ? string() : *calling_fn_name);
	}
	double log10b = std::log10(base);
	// detect overflow and throw exception
	if (e*log10b >= 300) {
		throw IBK_exception("Overflow in f_pow() function!",
			(calling_fn_name==NULL) ? string() : *calling_fn_name);
	}
	return std::pow(base, e);
}

// Safe variant for std::exp function
double f_exp(double x, std::string * calling_fn_name) {
	return std::exp(x);
}

// Safe variant for std::pow(10,x) function
double f_pow10(double e, std::string * calling_fn_name) {
#ifdef __GNUC__
#ifdef __MINGW32__
	return std::pow(10,e);
#else
	return pow10(e);
#endif
#else
	return std::pow(10,e);
#endif
}

// Safe variant for std::log function
double f_log(double x, std::string * calling_fn_name) {
	if (x<=0)
		throw IBK_exception("Negative or zero value in f_log() function!",
			(calling_fn_name==NULL) ? string() : *calling_fn_name);
	return std::log(x);
}

// Safe variant for std::log10 function
double f_log10(double x, std::string * calling_fn_name) {
	if (x<=0)
		throw IBK_exception("Negative or zero value in f_log10() function!",
			(calling_fn_name==NULL) ? string() : *calling_fn_name);
	return std::log10(x);
}

#endif // IBK_ENABLE_SAFE_MATH

} // namespace IBK

