// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
// 
// Copyright (C) 2003 Andreas Nicolai, Andreas.Nicolai@gmx.net
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#ifndef IBK_mathvector_h
#define IBK_mathvector_h

#include <stdexcept>
#include <algorithm>
#include <vector>
#include <string>
#include "IBK_message_strings.h"
#include "IBK_format.h"
#include "IBK_exception.h"

namespace IBK {

/// The IBK::math_vector is a std::vector enhanced by additional functions for
/// numeric operations.
/// The math_vector is a std::vector enhanced by numerous mathematical functions
/// thus making it usable as std::valarry but still providing STL support.
/// The template class math_vector only makes sense with numerical data types.
/// Usually a math vector represents a one-dimensional numerical vector.
/// The arithmetic operators are overloaded so that a very natural use
/// is possible. If you require a high performance numeric vector and you don't need
/// to use STL algoritms with it rather use std::valarray instead of math_vector.
/// IBK::math_vector does not have own member variables so passing it as a
/// std::vector does not cause slicing problems.
template <class T>
class math_vector : public std::vector<T> {
  public:
    typedef typename std::vector<T>::iterator           iterator;
    typedef typename std::vector<T>::const_iterator     const_iterator;

    /// Default constructor (creates an empty math_vector).
    math_vector() {};
    /// Constructor (creates a math_vector with 'size' elements).
    explicit math_vector(std::size_t size) : std::vector<T>(size) {};
    /// Constructor (creates a math_vector with 'size' elements with initial value 'init_value').
    math_vector(std::size_t size, const T& init_value) : std::vector<T>(size, init_value) {};
    /// Constructor (creates a math_vector and initialises it with values from the range 'first' to 'last').
    math_vector(const_iterator first, const_iterator last) : std::vector<T>(first, last) {};
    /// Constructor for intialisation with a std::vector.
    math_vector(const std::vector<T>& vec) : std::vector<T>(vec) {};

#ifdef IBK_ENABLE_RANGE_CHECKING

	/// Returns a reference to the vector element at index 'i'.
    T& operator[](std::size_t i);
	/// Returns a constant reference to the vector element at index 'i'.
    const T& operator[](std::size_t i) const;

#else // IBK_ENABLE_RANGE_CHECKING

	/// Returns a reference to the vector element at index 'i'.
    T& operator[](std::size_t i) { return std::vector<T>::operator[](i); }
	/// Returns a constant reference to the vector element at index 'i'.
    const T& operator[](std::size_t i) const { return std::vector<T>::operator[](i); }

#endif // IBK_ENABLE_RANGE_CHECKING

    /// Operator = for a single value ('value' will be set in all elements of the vector)
    const math_vector<T>& operator=(T value) {
    	std::fill(std::vector<T>::begin(), std::vector<T>::end(), value);
    	return *this;
	}

    /// Operator += for a single value ('value' will be added to every element of the vector).
    const math_vector<T>& operator+=(T value);
    /// Operator -= for a single value ('value' will be subtracted from every element of the vector).
    const math_vector<T>& operator-=(T value);
    /// Operator *= for a single value (every element of the vector will be multiplied with 'value').
    const math_vector<T>& operator*=(T value);
    /// Operator /= for a single value (every element of the vector will be divided by 'value').
    /// If the divisor is zero, the function throws a IBK::IBK_exception exception.
    const math_vector<T>& operator/=(T value);

    /// Operator + for a single value.
    /// Adds 'value' to all elements of the vector and returns the
    /// resulting vector. Instead of writing 'vec = vec + value' one should
    /// always use 'vec += value', because the creation of a temporary object
    /// is not necessary (and thus it will be faster).
    math_vector<T> operator+(T value) const;
    /// Operator - for a single value.
    /// Subtracts 'value' from all elements of the vector and returns the
    /// resulting vector. Instead of writing 'vec = vec - value' one should
    /// always use 'vec -= value', because the creation of a temporary object
    /// is not necessary (and thus it will be faster).
    math_vector<T> operator-(T value) const;
    /// Operator * for a single value.
    /// Multiplies all elements of the vector with 'value' and returns the
    /// resulting vector. Instead of writing 'vec = vec * value' one should
    /// always use 'vec *= value', because the creation of a temporary object
    /// is not necessary (and thus it will be faster).
    math_vector<T> operator*(T value) const;
    /// Operator / for a single value.
    /// Divides all elements of the vector by 'value' and returns the
    /// resulting vector. Instead of writing 'vec = vec / value' one should
    /// always use 'vec /= value', because the creation of a temporary object
    /// is not necessary (and thus it will be faster). The function throws a
    /// IBK::IBK_exception exception, if the value is zero.
    math_vector<T> operator/(T value) const;

    /// Operator += for math_vectors.
    /// Executes the operation this[i] += other[i] for all elements of the
    /// vectors. The function throws a IBK::IBK_exception exception if the
    /// vectors have different sizes.
    const math_vector<T>& operator+=(const math_vector<T>& other);
    /// Operator -= for math_vectors.
    /// Executes the operation this[i] -= other[i] for all elements of the
    /// vectors. The function throws a IBK::IBK_exception exception if the
    /// vectors have different sizes.
    const math_vector<T>& operator-=(const math_vector<T>& other);
    /// Operator *= for math_vectors.
    /// Executes the operation this[i] *= other[i] for all elements of the
    /// vectors. The function throws a IBK::IBK_exception exception if the
    /// vectors have different sizes.
    const math_vector<T>& operator*=(const math_vector<T>& other);
    /// Operator /= for math_vectors.
    /// Executes the operation this[i] /= other[i] for all elements of the
    /// vectors. The function throws a IBK::IBK_exception exception if the
    /// vectors have different sizes. 
    /// @warning This function DOES NOT test the divisor, so in case of
    ///          a divisor equal to zero the standard div function handles
    ///          that (usually by a call to _matherr). If you want to have
    ///          a check for each divisor use the member function div() instead.
    const math_vector<T>& operator/=(const math_vector<T>& other);
    /// Works like operator /= above but with checking for division by zero .
    /// In case that there is at least one divisor, which is equal to zero a
    /// IBK::IBK_exception exception will be thrown before the actual 
    /// calculation starts, so that no value of the vector will be
    /// changed in case of such an error.
    const math_vector<T>& div(const math_vector<T>& other);

    /// Operator + for math_vectors.
    /// Exceutes this[i] + other[i] for all elements and returns the resulting
    /// vector. Use vec1 += vec2 instead of vec1 = vec1 + vec2 (it's faster!).
    math_vector<T> operator+(const math_vector<T>& other) const;
    /// Operator - for math_vectors.
    /// Exceutes this[i] - other[i] for all elements and returns the resulting
    /// vector. Use vec1 -= vec2 instead of vec1 = vec1 - vec2 (it's faster!).
    math_vector<T> operator-(const math_vector<T>& other) const;
    /// Operator * for math_vectors.
    /// Exceutes this[i] * other[i] for all elements and returns the resulting
    /// vector. Use vec1 *= vec2 instead of vec1 = vec1 * vec2 (it's faster!).
    math_vector<T> operator*(const math_vector<T>& other) const;
    /// Operator / for math_vectors.
    /// Exceutes this[i] / summand[i] for all elements and returns the resulting
    /// vector. Use vec1 /= vec2 instead of vec1 = vec1 / vec2 (it's faster!).
    /// This version does NOT test for division by zero.
    math_vector<T> operator/(const math_vector<T>& other) const;
    /// Calculates and returns the sum of all elements of the vector.
    T sum() const;
    /// Calculates the mean value of all elements in the vector.
    /// Throws an exception if the vector is empty.
    T mean_value() const;
	/// Sets all values in the vector to zero (only makes sense for numerical
	/// types)
	void setZero() { std::memset(&std::vector<T>::front(), 0, std::vector<T>::size()*sizeof(T)); }
};
//*****************************************************************************

// **** Implementations of class members *****

#ifdef IBK_ENABLE_RANGE_CHECKING

template <class T>
T& math_vector<T>::operator[](std::size_t i) {
	if (i >= std::vector<T>::size())
		throw IBK::IBK_exception(format_string(IBK_INDEX_OUT_OF_RANGE) % i % std::vector<T>::size(),
			"[IBK::math_vector<T>::operator[] ]");
	return std::vector<T>::operator[](i);
}

template <class T>
const T& math_vector<T>::operator[](std::size_t i) const {
	if (i >= std::vector<T>::size())
        throw IBK::IBK_exception(format_string(IBK_INDEX_OUT_OF_RANGE) % i % std::vector<T>::size(),
			"[IBK::math_vector<T>::operator[] const]");
	return std::vector<T>::operator[](i);
}

#endif // IBK_ENABLE_RANGE_CHECKING

template <class T>
const math_vector<T>& math_vector<T>::operator+=(T value) {
    iterator it=std::vector<T>::begin();
    iterator eit=std::vector<T>::end();
    while (it!=eit)
        *it++ += value;
    return *this;
}

template <class T>
const math_vector<T>& math_vector<T>::operator-=(T value) {
    iterator it=std::vector<T>::begin();
    iterator eit=std::vector<T>::end();
    while (it!=eit)
        *it++ -= value;
    return *this;
}

template <class T>
const math_vector<T>& math_vector<T>::operator*=(T value) {
    iterator it=std::vector<T>::begin();
    iterator eit=std::vector<T>::end();
    while (it!=eit)
        *it++ *= value;
    return *this;
}

template <class T>
const math_vector<T>& math_vector<T>::operator/=(T value) {
    if (value==0)  throw IBK::IBK_exception(IBK_DIV_BY_ZERO,"[math_vector<T>::operator/=]");
    iterator it=std::vector<T>::begin();
    iterator eit=std::vector<T>::end();
    while (it!=eit)
        *it++ /= value;
    return *this;
}

template <class T>
math_vector<T> math_vector<T>::operator+(T value) const {
    math_vector<T> tmp(*this);
    return (tmp+=value);
}

template <class T>
math_vector<T> math_vector<T>::operator-(T value) const {
    math_vector<T> tmp(*this);
    return (tmp-=value);
}

template <class T>
math_vector<T> math_vector<T>::operator*(T value) const {
    math_vector<T> tmp(*this);
    return (tmp*=value);
}

template <class T>
math_vector<T> math_vector<T>::operator/(T value) const {
    math_vector<T> tmp(*this);
    return (tmp/=value);
}

template <class T>
const math_vector<T>& math_vector<T>::operator+=(const math_vector<T>& other) {
    if (std::vector<T>::size()!=other.size())
        throw IBK::IBK_exception(IBK_SIZE_MISMATCH,"[math_vector<T>::operator+=]");
    iterator it=std::vector<T>::begin();
    iterator eit=std::vector<T>::end();
    const_iterator src=other.begin();
    while (it!=eit)
        *it++ += *src++;
    return *this;
}

template <class T>
const math_vector<T>& math_vector<T>::operator-=(const math_vector<T>& other) {
    if (std::vector<T>::size()!=other.size())
        throw IBK::IBK_exception(IBK_SIZE_MISMATCH,"[math_vector<T>::operator-=]");
    iterator it=std::vector<T>::begin();
    iterator eit=std::vector<T>::end();
    const_iterator src=other.begin();
    while (it!=eit)
        *it++ -= *src++;
    return *this;
}

template <class T>
const math_vector<T>& math_vector<T>::operator*=(const math_vector<T>& other) {
    if (std::vector<T>::size()!=other.size())
        throw IBK::IBK_exception(IBK_SIZE_MISMATCH,"[math_vector<T>::operator*=]");
    iterator it=std::vector<T>::begin();
    iterator eit=std::vector<T>::end();
    const_iterator src=other.begin();
    while (it!=eit)
        *it++ *= *src++;
    return *this;
}

template <class T>
const math_vector<T>& math_vector<T>::operator/=(const math_vector<T>& other) {
    if (std::vector<T>::size()!=other.size())
        throw IBK::IBK_exception(IBK_SIZE_MISMATCH,"[math_vector<T>::operator*=]");
    iterator it=std::vector<T>::begin();
    iterator eit=std::vector<T>::end();
    const_iterator src=other.begin();
    while (it!=eit)
        *it++ /= *src++;
    return *this;
}

template <class T>
const math_vector<T>& math_vector<T>::div(const math_vector<T>& other) {
    if (std::vector<T>::size()!=other.size())
        throw IBK::IBK_exception(IBK_SIZE_MISMATCH,"[math_vector<T>::operator*=]");
    if (std::find(other.begin(), other.end(), 0.0)!=other.end())
        throw IBK::IBK_exception(IBK_DIVISOR_CONTAINS_ZERO,"[math_vector<T>::operator/=]");
    iterator it=std::vector<T>::begin();
    const_iterator src=other.begin();
    const_iterator eit=std::vector<T>::end();
    while (it!=eit)
        *it++ /= *src++;
    return *this;
}

// **** Implementations of global functions *****

template <class T>
math_vector<T> math_vector<T>::operator+(const math_vector<T>& other) const {
    math_vector<T> tmp(*this);
    return (tmp+=other);
}

template <class T>
math_vector<T> math_vector<T>::operator-(const math_vector<T>& other) const {
    math_vector<T> tmp(*this);
    return (tmp-=other);
}

template <class T>
math_vector<T> math_vector<T>::operator*(const math_vector<T>& other) const {
    math_vector<T> tmp(*this);
    return (tmp*=other);
}

template <class T>
math_vector<T> math_vector<T>::operator/(const math_vector<T>& other) const {
    math_vector<T> tmp(*this);
    return (tmp/=other);
}


// *** binary operators ***

template <class T>
math_vector<T> operator+(const T value, const math_vector<T>& vec) {
    return (vec+value);
}

template <class T>
math_vector<T> operator*(const T value, const math_vector<T>& vec) {
    return (vec*value);
}

template <class T>
T math_vector<T>::sum() const {
    const_iterator it=std::vector<T>::begin();
    const_iterator eit=std::vector<T>::end();
    T tmp=0;
    while (it!=eit)
        tmp += *(it++);
    return tmp;
}

template <class T>
T math_vector<T>::mean_value() const {
    if (std::vector<T>::empty())
        throw IBK::IBK_exception(IBK_VECTOR_IS_EMPTY,"[math_vector<T>::mean_value]");
    return sum()/std::vector<T>::size();
}

/// Calculates the Cartesian product (only defined for vectors of size 3).
/// This function throws an exception if one of the parameter vectors hasn't got a size of 3.
template <class T>
math_vector<T> cartesian_product(const math_vector<T>& vec1, const math_vector<T>& vec2) {
    if (vec1.size()!=3 || vec2.size()!=3)
        throw IBK::IBK_exception(IBK_VECTOR_SIZE_3_REQUIRED,"[math_vector<T>::cartesian_product]");
    math_vector<T> tmp(3);
    tmp[0]=vec1[1]*vec2[2] - vec2[1]*vec1[2];
    tmp[1]=vec1[2]*vec2[0] - vec2[2]*vec1[0];
    tmp[2]=vec1[0]*vec2[1] - vec2[0]*vec1[1];
    return tmp;
}

} // namespace IBK

/// @file IBK_mathvector.h
/// @brief Contains the template class IBK::math_vector.

/// @example mathvector_example.cpp
/// This is an example of how to use the IBK::math_vector class template.

#endif  // IBK_mathvector_h
