// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
//
// Copyright (C) 2003 Andreas Nicolai, Andreas.Nicolai@gmx.net
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#ifndef IBK_matrix_h
#define IBK_matrix_h

#include <iosfwd>

namespace IBK {

/// Abstract base class for all matrices of the IBK lib.
/// This class only exists for output/input purposes. So you can pass a matrix
/// regardless of its internal implementation (and perhaps optimisation for a
/// certain purpose) to another funtion which only needs element access.
/// A typical example would be output routines.
template <typename T>
class matrix {
  public:
    typedef T           value_type;         ///< Data types of the values in the matrix.
    typedef T&          reference;          ///< Reference type to elements in the matrix.
    typedef const T&    const_reference;    ///< Constant reference type to elements in the matrix.

    /// Virtual destructor, so that the destructors of derived matrix classes are called.
    virtual ~matrix() {};
    /// Returns a reference to the matrix element at [col][row].
    /// This function must be implemented by derived classes. If IBK_ENABLE_RANGE_CHECKING is
    /// defined the function should throw an exception if the indices are out of range.
    virtual reference  at(std::size_t col, std::size_t row) = 0;
    /// Returns a constant reference to the matrix element at [col][row].
    /// This function must be implemented by derived classes. If IBK_ENABLE_RANGE_CHECKING is
    /// defined the function should throw an exception if the indices are out of range.
    virtual const_reference  at(std::size_t col, std::size_t row) const = 0;
    /// Returns the number of rows in the matrix (must be reimplemented in derived classes).
    virtual std::size_t rows() const = 0;
    /// Returns the number of columns in the matrix (must be reimplemented in derived classes).
    virtual std::size_t cols() const = 0;
};
// ---------------------------------------------------------------------------

} // namespace IBK

#endif // IBK_matrix_h
