// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
//
// Copyright (C) 2004-2006 Andreas Nicolai, Andreas.Nicolai@gmx.net
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#ifndef IBK_matrixmath_h
#define IBK_matrixmath_h

#include "IBK_configuration.h"

#include <cmath> // for fabs()

#include "IBK_debug.h"

namespace IBK {

/// Performs the matrix multiplication C = A*B.
///
/// The matrices must be square matrices and must be stored column-based in
/// a linear memory block of size n*n.
template <typename T>
void matrix_matrix_mult(unsigned int n, const T * A, const T * B, T * C) {
	for (unsigned int i=0; i<n; ++i)
		for (unsigned int j=0; j<n; ++j)
			for (unsigned int k=0; k<n; ++k)
				C[i + j*n] += A[i + k*n] * B[k + j*n];
}
// ---------------------------------------------------------------------------

/// Performs the matrix multiplication C = A*B.
///
/// C++ version of the function above taking std::vector arguments.
template <typename T>
void matrix_matrix_mult(unsigned int n, const std::vector<T>& A,
	const std::vector<T>& B, std::vector<T>& C)
{
	IBK_check(A.size() == B.size() == C.size() == n*n);
	matrix_matrix_mult(n, &A[0], &B[0], &C[0]);
}
// ---------------------------------------------------------------------------


/// Performs the matrix multiplication and subtraction C = C - A*B.
///
/// The matrices must be square matrices and must be stored column-based in
/// a linear memory block of size n*n. The result of the multiplication of
/// A and B is subtracted from the matrix C.
template <typename T>
void matrix_matrix_multsub(unsigned int n, const T * A, const T * B, T * C) {
	for (unsigned int i=0; i<n; ++i)
		for (unsigned int j=0; j<n; ++j)
			for (unsigned int k=0; k<n; ++k)
				C[i + j*n] -= A[i + k*n] * B[k + j*n];
}
// ---------------------------------------------------------------------------

/// Performs the matrix multiplication and subtraction C = C - A*B.
///
/// C++ version of the function above taking std::vector arguments.
template <typename T>
void matrix_matrix_multsub(unsigned int n, const std::vector<T>& A,
	const std::vector<T>& B, std::vector<T>& C)
{
	IBK_check(A.size() == B.size() == C.size() == n*n);
	matrix_matrix_multsub(n, &A[0], &B[0], &C[0]);
}
// ---------------------------------------------------------------------------


/// Performs the matrix vector multiplication and subtraction d = d - M*b.
///
/// The matrix M must be square matrices and must be stored column-based in
/// a linear memory block of size n*n. The vectors b and d must be
/// of size n. The result of the multiplication of M and b is subtracted
/// from the vector d.
template <typename T>
void matrix_vector_multsub(unsigned int n, const T * M, const T * b, T * d) {
	for (unsigned int i=0; i<n; ++i)
		for (unsigned int k=0; k<n; ++k)
			d[i] -= M[i + k*n] * b[k];
}
// ---------------------------------------------------------------------------

/// Performs the matrix vector multiplication and subtraction d = d - M*b.
///
/// C++ version of the function above taking std::vector arguments.
template <typename T>
void matrix_vector_multsub(unsigned int n, const std::vector<T>& M,
	const std::vector<T>& b, std::vector<T>& d)
{
	IBK_check(M.size() == n*n && d.size() == b.size() == n);
	matrix_vector_multsub(n, &M[0], &b[0], &d[0]);
}
// ---------------------------------------------------------------------------


/// LU decomposition of the matrix M using partial pivoting.
///
/// The matrix M must be stored column-based in a continuous memory block
/// of size n*n. The vector pivots must be of size n and will contain the
/// pivoting information. The vector workspace must be of size n.
/// The decomposed matrix will be stored in place of the matrix M.
template <typename T>
int matrix_decomposition(unsigned int n, T * M, unsigned int * pivots, T * workspace) {
	T pivot, temp, sum;
	unsigned int imax;
	// determine scaling of rows for crout's algorithm
	for (unsigned int j=0; j<n; ++j) {
		pivot = 0.0;
		for (unsigned int i=0; i<n; ++i) {
			temp = M[i + j*n];
			if (fabs(temp) > fabs(pivot))
				pivot = temp;
		}
		if (pivot == 0.0)  return -2; // matrix singular
		workspace[j] = 1.0/pivot;
	}
	// loop over columns
	for (unsigned j=0; j<n; ++j) {
		imax = j;
		for (unsigned int i=0; i<j; ++i) {
			sum = M[i + j*n];
			for (unsigned int k=0; k<i; ++k)
				sum -= M[i + k*n]*M[k + j*n];
			M[i + j*n] = sum;
		}
		// search larges pivot element index
		pivot = 0.0;
		for (unsigned i=j; i<n; ++i) {
			sum = M[i + j*n];
			for (unsigned int k=0; k<j; ++k)
				sum -= M[i + k*n]*M[k + j*n];
			M[i + j*n] = sum;
			temp = workspace[i]*fabs(sum);
			if ( fabs(temp) > pivot) {
				pivot = fabs(temp);
				imax = i;
			}
		}
		// swap rows if necessary
		if (j != imax) {
			for (unsigned int k=0; k<n; ++k)
				std::swap(M[imax + k*n], M[j + k*n]);
			workspace[imax] = workspace[j];
		}
		pivots[j] = imax;
		if (M[j + j*n] == 0.0)  M[j + j*n]=1e-20;
		if (j != n-1) {
			temp = 1.0/M[j + j*n];
			for (unsigned int i=j+1; i<n; ++i)
				M[i + j*n] *= temp;
		}
	}
	return 0;
}
// ---------------------------------------------------------------------------

/// LU decomposition of the matrix M using partial pivoting.
///
/// C++ version of the function above taking std::vector arguments.
template <typename T>
int matrix_decomposition(unsigned int n, std::vector<T>& M,
	std::vector<unsigned int>& pivots, std::vector<T>& workspace)
{
	IBK_check(M.size() == n*n && pivots.size() == workspace.size() == n);
	return matrix_decomposition(n, &M[0], &pivots[0], &workspace[0]);
}
// ---------------------------------------------------------------------------


/// Backsolving algorithm for solving the equation system M*x = d.
///
/// M must be a matrix stored column-based in a linear memory block of size n*n
/// and must contain already the decomposed matrix (use matrix_decomposition for that).
/// The vector pivots must contain the pivoting information (again from
/// matrix_decomposition) and be of size n. The vector d must be of size n. The
/// result will be stored in place of the vector d.
template <typename T>
void matrix_backsolve(unsigned int n, const T * M, unsigned int * pivots, T * d) {
	unsigned int ip;
	int ii=-1;
	T sum;
	for (unsigned int i=0; i<n; ++i) {
		ip = pivots[i];
		sum = d[ip];
		d[ip] = d[i];
		if (ii >=0 )
			for (unsigned int j=ii; j<=i-1; ++j)
				sum -= M[i + j*n]*d[j];
		else
			if (sum) ii = i;
		d[i] = sum;
	}
	for (int i=n-1; i>=0; --i) {
		sum = d[i];
		for (unsigned int j=i+1; j<n; ++j)
			sum -= M[i + j*n]*d[j];
		d[i] = sum/M[i + i*n];
	}
}
// ---------------------------------------------------------------------------

/// Backsolving algorithm for solving the equation system M*x = d.
///
/// C++ version of the function above taking std::vector arguments.
template <typename T>
void matrix_backsolve(unsigned int n, const std::vector<T>& M,
	std::vector<unsigned int>& pivots, std::vector<T>& d)
{
	IBK_check(M.size() == n*n && pivots.size() == d.size() == n);
	matrix_backsolve(n, &M[0], &pivots[0], &d[0]);
}
// ---------------------------------------------------------------------------


/// Solves B = A^-1*B.
///
/// A must contain the decomposed matrix A. Matrices A and B must be stored in
/// column-based linear memory blocks and have the size n*n. The vector pivots
/// must contain the pivoting information from the decomposition of the matrix A.
template <typename T>
void matrix_matrix_inverse_mult(unsigned int n, const T * A, unsigned int * pivots, T * B) {
	for (unsigned int i=0; i<n; ++i)
		matrix_backsolve(n, A, pivots, B + i*n);
}
// ---------------------------------------------------------------------------

/// Solves B = A^-1*B
///
/// C++ version of the function above taking std::vector arguments.
template <typename T>
void matrix_matrix_inverse_mult(unsigned int n, const std::vector<T>& A,
	std::vector<unsigned int>& pivots, std::vector<T>& B)
{
	matrix_matrix_inverse_mult(n, &A[0], &pivots[0], &B[0]);
}
// ---------------------------------------------------------------------------

/// @example matrixmath_example.cpp
/// This is an example and test suite of how to use the matrix math routines.

} // namespace IBK

#endif // IBK_matrixmath_h
