// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
// 
// Copyright (C) 2003 Andreas Nicolai, Andreas.Nicolai@gmx.net
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#ifndef IBK_matrixview_h
#define IBK_matrixview_h

#include "IBK_configuration.h"

#include "IBK_format.h"
#include "IBK_message_strings.h"
#include "IBK_debug.h"

namespace IBK {

/// The class matrix_view is a container adapter providing a matrix view for 
/// linear array data.
/// A matrix_view is not a container itself, it merely allows you to treat
/// a linear data vector, such as the plain C array or the std::vector as a
/// matrix. When creating a matrix view you simply have to specify the 
/// dimensions and a pointer to the start of the linear data storage.
/// Note that creating and destructing a matrix view does not change the
/// data in the storage at all.
template <class T>
class matrix_view {
  public:
    typedef T           value_type;         ///< The data type of the values.
    typedef T&          reference;          ///< Reference to one element.
    typedef const T&    const_reference;    ///< Constant reference to one element.
    typedef T*          iterator;           ///< Iterator to one element.
    typedef const T*    const_iterator;     ///< Constant iterator to one element.

    /// Constructor, creates a matrix view for a linear storage container.
    /// @param cols Number of colums of the matrix.
    /// @param rows Number of rows of the matrix.
    /// @param storage Pointer to the start of the data.
    matrix_view(std::size_t cols, std::size_t rows, T* storage)
        : cols_(cols), rows_(rows), storage_(storage) {};

    /// Returns an iterator to the first element in the storage.
    iterator        begin()         { return storage_; }
    /// Returns a constant iterator to the first element in the storage.
    const_iterator  begin() const   { return storage_; }
    /// Returns an iterator to one past the last element in the storage.
    iterator        end()           { return storage_+cols_*rows_; }
    /// Returns a constant iterator to one past the last element in the storage.
    const_iterator  end() const     { return storage_+cols_*rows_; }

#ifdef IBK_ENABLE_RANGE_CHECKING

    /// Returns a reference to the element at (col,row).
    reference       operator() (std::size_t col, std::size_t row) { 
    	IBK_check(col<cols_);
    	IBK_check(row<rows_);
        return storage_[row*cols_+col]; 
    }

    /// Returns a constant reference to the element at (col,row).
    const_reference operator() (std::size_t col, std::size_t row) const { 
    	IBK_check(col<cols_);
    	IBK_check(row<rows_);
        return storage_[row*cols_+col];
    }

#else // IBK_ENABLE_RANGE_CHECKING

    /// Returns a reference to the element at (col,row).
    reference       operator() (std::size_t col, std::size_t row) { return storage_[row*cols_+col]; }
    /// Returns a constant reference to the element at (col,row).
    const_reference operator() (std::size_t col, std::size_t row) const { return storage_[row*cols_+col]; }

#endif // IBK_ENABLE_RANGE_CHECKING

    /// Changes the dimensions of the matrix view.
    void resize(std::size_t cols, std::size_t rows, T* storage=NULL);

    /// Returns the number of columns in the matrix.
    std::size_t     cols() const    { return cols_; };
    /// Returns the number of rows in the matrix.
    std::size_t     rows() const    { return rows_; };

  private:
    std::size_t cols_, rows_;
    T *storage_;
};
// ---------------------------------------------------------------------------

template <class T>
void matrix_view<T>::resize(std::size_t cols, std::size_t rows, T* storage) {
    cols_=cols;
    rows_=rows;
    storage_=storage;
}

} // namespace IBK

#endif  // IBK_matrix_view_h
