// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
// 
// Copyright (C) 2003 Andreas Nicolai, Andreas.Nicolai@gmx.net
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#ifndef IBK_messages_h
#define IBK_messages_h

#include <string>
#include "IBK_format.h"

namespace IBK {

#ifdef __BORLANDC__
#define ERROR_MSG(x)  IBK::MSG(x, IBK::MSG_ERROR, __FUNC__)
#elif __MINGW32__
#define ERROR_MSG(x)  IBK::MSG(x, IBK::MSG_ERROR, __func__)
#elif _MSC_VER
#define ERROR_MSG(x)  IBK::MSG(x, IBK::MSG_ERROR, __FUNCTION__)
#else
#define ERROR_MSG(x)  IBK::MSG(x, IBK::MSG_ERROR, __func__)
#endif //__BORLANDC__

enum msg_type_t {
    MSG_PROGRESS,
	MSG_WARNING,
    MSG_ERROR,
    MSG_DEBUG
};

/// Abstract base class for message receiver classes.
class MessageReceiver {
  public:
  	virtual ~MessageReceiver() {}
    virtual void msg(const std::string& msg, msg_type_t t = MSG_PROGRESS, const char * func_id = NULL, int verbose_level = -1) = 0;
};

class DefaultMessageReceiver : public MessageReceiver {
    virtual void msg(const std::string& msg, msg_type_t t = MSG_PROGRESS, const char * func_id = NULL, int verbose_level = -1);
};

/// This class relays messages (error, debug or progress messages)
/// from the IBK library towards the current message handling class.
class MessageHandlerImpl {
  public:
    void msg(const std::string& msg, msg_type_t t = MSG_PROGRESS, const char * func_id = NULL, int verbose_level = -1) {
    	m_receiver->msg(msg, t, func_id, verbose_level);
    };

    void setDefaultReceiver() { m_receiver = &m_defaultReceiver; }
    void setReceiver(MessageReceiver * receiver) { m_receiver = receiver; }

  private:
  	MessageHandlerImpl();
    MessageHandlerImpl(const MessageHandlerImpl&);

    DefaultMessageReceiver 	 m_defaultReceiver;
	MessageReceiver 		*m_receiver;

  friend MessageHandlerImpl& MessageHandler();
};

MessageHandlerImpl& MessageHandler();

/// Prototype for the message function.
/// The function will be called from certain IBK classes.
inline void MSG(const std::string& msg, msg_type_t t = MSG_PROGRESS, const char * func_id = NULL, int verbose_level = -1) {
	MessageHandler().msg(msg, t, func_id, verbose_level);
}

inline void MSG(const IBK::format_string& msg, msg_type_t t = MSG_PROGRESS, const char * func_id = NULL, int verbose_level = -1) {
	MessageHandler().msg(msg.str(), t, func_id, verbose_level);
}

/// @file IBK_messages.h
/// Contains prototypes for the message functions used in the IBK-lib.

} // namespace IBK

#endif // IBK_messages_h

