// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
// 
// Copyright (C) 2006 Andreas Nicolai, Andreas.Nicolai@gmx.net
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#include "IBK_configuration.h"

#ifdef __BORLANDC__
  #include "IBK_prehead.h"
  #pragma hdrstop
#else
  #include <cmath>
  #include <stdexcept>
  #include <limits>
  #include <utility>
#endif
#include "IBK_numerics.h"
#include "IBK_format.h"
#include "IBK_exception.h"
using namespace std;

namespace IBK {

const double GOLDEN_SECTION = 0.61803399;
const double GOLDEN_SECTION_1 = 1.61803399;
const double GOLDEN_SECTION_COMP = (1.0-GOLDEN_SECTION);
const double GOLDEN_LIMIT = 100.0;
const double TINY = std::numeric_limits<double>::epsilon();


// ROMBERG-Integration
// (algorithm found on http://www.voidware.com/rombint.htm)
//
// Approximates the numeric integral of `f' between `a' and `b' subject
// to a given `eps'ilon of error.
//
// Use Romberg method with refinement substitution, x = (3u-u^3)/2 which
// prevents endpoint evaluation and causes non-uniform sampling.
//
// max samples 2^(ROMB_MAX)+1
//
double romberg_integral(const DblFunction &f, double a, double b, double eps) {
	const int ROMB_MAX = 20;

    int n, i;
    double h, s;

    double c[ROMB_MAX+1];    
    int lim;
    double p;
    double t, u;
    double l;
    
    h = 2;
    c[0] = 0;
    l = 0;
    lim = 1;
    b -= a;
    for (n = 0; n < ROMB_MAX; ++n) {
        p = h/2 - 1;
        s = 0.0;
        for (i = 0; i < lim; i++) { 
            t = 1-p*p;
            u = p + t*p/2;
            u = (u*b + b)/2 + a;
            s = s + t*f(u);
            p += h;
        }

        p = 4;      
        t = c[0];
        c[0] = (c[0] + h*s)/2.0;
        for(i = 0; i <= n; i++) {
            u = c[i+1];
            c[i+1] = (p*c[i] - t)/(p-1);
            t = u;
            p *= 4;
        }

        if (fabs(c[i] - l) < eps*fabs(c[i])*16) break;
        l = c[i];
        lim <<= 1;
        h /= 2.0;
    }
    return c[i]*b*3/4;
}
// ----------------------------------------------------------------------------

// The bi-section root finding algorithm - simple but slowest.
double bisection(const DblFunction& f, double a, double  b, unsigned int& iterations, double eps) {
    double xm = a, f_m;
    double f_1 = f(a);
	if (f_1 == 0) return a;
    while (--iterations) {
        xm = (a+b)/2;
        f_m = f(xm);
        if (std::fabs(f_m) < eps)  return xm;
        if ((f_m<0 && f_1>0) || (f_m>0 && f_1<0))  b = xm;
        else                                       a = xm;
    }
    return xm;
}
// ----------------------------------------------------------------------------

// The regula falsi root finding algorithm - faster, but still not the fastest.
double regula_falsi(const DblFunction& f, double a, double b, unsigned int& iterations, double eps) {
    double xm = a, f_1, f_2, f_m;
    while (--iterations) {
        f_1 = f(a);
        f_2 = f(b);
        if (f_1==f_2) throw IBK_exception("f(a)==f(b)!","[regula_falsi]");
        xm = b-f_2*(b-a)/(f_2-f_1);
        f_m = f(xm);
        if (std::fabs(f_m) < eps) return xm;
        if ((f_m<0 && f_1>0) || (f_m>0 && f_1<0))  b = xm;
        else                                       a = xm;
    }
    return xm;
}
// ----------------------------------------------------------------------------

// Newtons root finding algorithm - fastest, but requires the derivative.
double newton_root(const DblFunction& f, DblFunction& df, double a, unsigned int& iterations, double eps) {
    double df_1, b = a, f_2;
    while (--iterations) {
        df_1 = df(a);
        if (df_1==0) throw IBK_exception("df(a)==0!","[newton_root]");
        b = a-f(a)/df_1;
        f_2 = f(b);
        if (std::fabs(f_2) < eps) return b;
        a = b;
    }
    return b;
}
// -----------------------------------------------------------------------------

// Function required by the bracketing algorithm
inline double sign(double a, double b) { return b >= 0.0 ? std::fabs(a) : -std::fabs(a); }
// -----------------------------------------------------------------------------

void bracket_minimum(const DblFunction& f, double& a, double& b, double& c) {
	double fa = f(a);
    double fb = f(b);
    // switch directions so that the downhill direction is a -> b
    if (fb > fa) {
    	std::swap(a, b);
        std::swap(fa, fb);
    }
    // if we don't have a given c, guess a first value for c
    if (c == -1)
    	c = b + GOLDEN_SECTION_1*(b-a);
    double fc = f(c);
    // now step downhill until fa > fb && fb < fc
    while (fb > fc) {
    	// compute parabolic extrapolation from a, b, c
    	double r = (b-a)*(fb-fc);
        double q = (b-c)*(fb-fa);
        double u = b - ( (b-c)*q - (b-a)*r)/
        	(2.0*sign(std::max<double>(fabs(q-r), TINY), q-r) );
        // calculate a limit that we will never overstep
        double ulim = b + GOLDEN_LIMIT*(c-b);
        // test various possibilities
        double fu;
        if ((b-u)*(u-c) > 0) {
        	// parabolic u might be between b and c, try it
            fu = f(u);
            if (fu < fc) {	// got a minimum between b and c,
            	a = b; // store our bracket and return
                b = u;
                return;
            }
            else if (fu > fb) { // got a minimum between a and u
            	c = u;
                return;
            }
            // parabolic fit didn't work, use standard magnification
            u = c + GOLDEN_SECTION_1*(c-b);
            fu = f(u);
        }
        else if ((c-u)*(u-ulim) > 0) {
        	// parabolic fit might be between u and its limit ulim
            fu = f(u);
            if (fu < fc) {
                // it is, so set new bracket to b, u, c
            	b = c; c = u; u = c + GOLDEN_SECTION_1*(c-b);
                fb = fc; fc = fu; fu = f(u);
            }
        }
        else if ((u-ulim)*(ulim-c) >= 0) {
        	// limit u to maximum allowed value
            u = ulim;
            fu = f(u);
        }
        else {
        	// parabolic fit didn't work, use standard magnification
            u = c + GOLDEN_SECTION_1*(c-b);
            fu = f(u);
        }
        // move bracket
        a = b; b = c; c = u;
        fa = fb; fb = fc; fc = fu;
    }
}
// -----------------------------------------------------------------------------

double golden_section(const DblFunction& f, double a, double b, double c,
	unsigned int& iterations, double eps)
{
	// required: a < b < c
    if (a > b || b > c)
    	throw IBK_exception("Invalid input. Required: a < b < c!", "[golden_section]");
	// also required: f(a) > f(b) && f(b) < f(c)
    if (f(a) < f(b) || f(b) > f(c))
    	throw IBK_exception("Invalid input. Required: f(a) > f(b) && f(b) < f(c)!", "[golden_section]");

	// golden section search algorithm, based on Numerical Recipes in C code
    // four points are considered i=0...3
	double x0 = a; // x0 is the left boundary of the bracket
    double x3 = c; // x3 is the right boundary of the bracket
    double x1, x2;
    // make x0-x1 the smaller segment
    if (fabs(c-b) > fabs(b-a)) {
    	x1 = b;
    	x2 = b + GOLDEN_SECTION_COMP*(c-b);
    }
    else {
    	x1 = b - GOLDEN_SECTION_COMP*(b-a);
    	x2 = b;
    }
    // initial function evaluations
    double f1 = f(x1);
    double f2 = f(x2);
    // now loop until we have found the minimum
    while (fabs(x3-x0) > eps*(fabs(x1)+fabs(x2))) {
    	// one of the possible outcomes
        if (f2 < f1) {
        	// shift the bracket to the left
            x0 = x1; x1 = x2; x2 = GOLDEN_SECTION*x1 + GOLDEN_SECTION_COMP*x3;
            // shift the function evaluations
            f1 = f2; f2 = f(x2);
        }
        else {
        	// shift bracket to the right
            x3 = x2; x2 = x1; x1 = GOLDEN_SECTION*x2 + GOLDEN_SECTION_COMP*x0;
            // shift the function evaluations
            f2 = f1; f1 = f(x1);
        }
        // check that our iteration limit wasn't exceeded
        if (--iterations == 0)  return 0.5*(x1+x2);
    }
    if (f1 < f2)  	return x1;
    else 			return x2;
}
// -----------------------------------------------------------------------------


/* 

  FIXME : this code is not standard C++

#ifdef __BORLANDC__

double InverseFunction::operator() (double y) const {
    unsigned int iterations = 200;
    InverseAdapter f_1(m_f, y);
    double result = IBK::bisection(f_1, m_a, m_b, iterations, m_eps);
	if (iterations != 0) return result;
	iterations = 200;
   	result = IBK::bisection(f_1, m_a, m_b, iterations, m_eps*10);
	if (iterations != 0) return result;
	iterations = 2000;
   	IBK::bisection(f_1, m_a, m_b, iterations, m_eps*100);
	throw std::IBK_exception( (format_string("[IBK::InverseFunction]  "
		"Could not find function value % in given search range %...%!")
		% y % m_a % m_b).str());
}
// -----------------------------------------------------------------------------
#endif // __BORLANDC__

*/

double scale(double x) {
	return 0.5 - 0.5*cos(x*3.141592654);
}
// -----------------------------------------------------------------------------

}  // namespace IBK

