// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
// 
// Copyright (C) 2006 Andreas Nicolai, Andreas.Nicolai@gmx.net
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#ifndef IBK_numerics_h
#define IBK_numerics_h

#include <functional>
#include "IBK_types.h"
#include "IBK_linearspline.h"

namespace IBK {

/// Generic linear interpolation function taking a sorting criterion for the x values.
template <typename T, class Compare>
T interpolate(const T * x_array, const T * y_array, unsigned int n, T x, Compare comp) {
    unsigned int left = 0;
    unsigned int m = n / 2;
    while (m!=left) {
        if (comp(x, x_array[m]))    n = m;      // adjust right border
        else                        left = m;   // adjust left border
        m = (left + n) / 2;
    }
    return y_array[m] + (y_array[m+1]-y_array[m])/(x_array[m+1]-x_array[m])*(x-x_array[m]);
}

/// Generic linear interpolation function (you don't have to specify how the x values are ordered).
template <typename T>
T interpolate(const T * x_array, const T * y_array, unsigned int n, T x) {
    if (x_array[0] < x_array[1])
        return interpolate<T>(x_array, y_array, n, x, std::less<T>());
    else
        return interpolate<T>(x_array, y_array, n, x, std::greater<T>());
}

/// Generic linear interpolation function taking a sorting criterion for the x values.
template <typename T, class Compare>
T interpolate(const T * x_array, const T * y_array, const T * slopes, unsigned int n, T x, Compare comp) {
    unsigned int left = 0;
    unsigned int m = n / 2;
    while (m!=left) {
        if (comp(x, x_array[m]))    n = m;      // adjust right border
        else                        left = m;   // adjust left border
        m = (left + n) / 2;
    }
    return y_array[m] + slopes[m]*(x-x_array[m]);
}

/// Generic linear interpolation function (you don't have to specify how the x values are ordered).
template <typename T>
T interpolate(const T * x_array, const T * y_array, const T * slopes, unsigned int n, T x) {
    if (x_array[0] < x_array[1])
        return interpolate<T>(x_array, y_array, slopes, n, x, std::less<T>());
    else
        return interpolate<T>(x_array, y_array, slopes, n, x, std::greater<T>());
}

/// Performs a numeric integration using the Romberg algorithm.
/// @param f 	The function object encapsulating the function.
/// @param a	The lower bound of the integration range.
/// @param b	The upper bound of the integration range.
/// @param eps  (optional) The allowed relative tolerance.
double romberg_integral(const DblFunction &f, double a, double b, double eps);

/// The bi-section root finding algorithm - simple but slowest.
/// Finds x so that f(x) = 0.
/// @param f 			The function object encapsulating the function f(x).
/// @param a			The lower bound of the search range.
/// @param b			The upper bound of the search range.
/// @param iterations	The maximum number of iterations to be performed. Must be a 
///                     reference variable. After algorithm has finished it contains 
///                     the number of iterations left. 
///                     If iterations==0 the returned solution is outside the 
///                     tolerance margin.
/// @param eps			(optional) The allowed relative tolerance.
double bisection(const DblFunction& f, double a, double  b, unsigned int& iterations, double eps);

/// The regula falsi root finding algorithm - faster than bisection, but still not the fastest.
/// Finds x so that f(x) = 0.
/// @param f 			The function object encapsulating the function f(x).
/// @param a			The lower bound of the search range.
/// @param b			The upper bound of the search range.
/// @param iterations	The maximum number of iterations to be performed. Must be a
///                     reference variable. After algorithm has finished it contains
///                     the number of iterations left.
///                     If iterations==0 the returned solution is outside the
///                     tolerance margin.
/// @param eps			(optional) The allowed relative tolerance.
double regula_falsi(const DblFunction& f, double a, double b, unsigned int& iterations, double eps);

/// Newtons root finding algorithm - fastest, but requires the derivative.
/// Finds x so that f(x) = 0.
/// @param f 			The function object encapsulating the function f(x).
/// @param df 			The function object encapsulating df/dx, the derivative of the function f(x).
/// @param a			The starting value for the search (should be already near the solution).
/// @param iterations	The maximum number of iterations to be performed. Must be a 
///                     reference variable. After algorithm has finished it contains 
///                     the number of iterations left. 
///                     If iterations==0 the returned solution is outside the 
///                     tolerance margin.
/// @param eps			(optional) The allowed relative tolerance.
double newton_root(const DblFunction& f, DblFunction& df, double a, unsigned int& iterations, double eps);

/// This function calculates a, b and c in such a way, that they form a valid minimization bracket.
/// A valid minimization bracket is found when a &lt; b &lt; c and f(a) &gt; f(b) and f(b) &lt; f(c).
/// @param f			The minimization function.
/// @param a            The left side of the guessed bracket.
/// @param b            Another point in the interval, where a minimum is probable.
/// @param c            (optional) The right side of the guessed bracket (the algorithm will not go past this point).
/// @return             The function does not return anything, but the values a, b, c will contain the bracket values.
void bracket_minimum(const DblFunction& f, double& a, double& b, double& c);

/// This function calculates the minimum of the function f in the interval a, c.
/// The function requires that the arguments a, b and c form a valid minimization bracket. 
/// A valid minimization bracket is given when a &lt; b &lt; c and f(a) &gt; f(b) and f(b) &lt; f(c).
/// @param f			The minimization function.
/// @param a            The left side of the guessed bracket.
/// @param b            Another point in the interval, where a minimum is probable.
/// @param c            The right side of the guessed bracket.
/// @param iterations   The maximum number of iterations to use (will hold the number of remaining 
///                     iterations after the function returned).
/// @param eps          The accuracy of the solution (don't set it to less then sqrt(numeric_limits&lt;double&rt;(epsilon)) ).
/// @return             Returns the minimum if found or the closest value to it.
double golden_section(const DblFunction& f, double a, double b, double c,
	unsigned int& iterations, double eps);


/*

This code is not standard C++ and need fixing...

#ifdef __BORLANDC__
/// This is an adapter class, that calculates g(x, y) = y - f(x), whereas f(x) can
/// be an arbitrary function encapsulated in a function object.
class InverseAdapter : public DblFunction {
  public:
	/// Constructor, creates the InverseAdapter function object.
    InverseAdapter(const DblFunction &f, double y) : m_y(y), m_f(f) {}
	/// Calculates the function g(x, y) = y - f(x).
	double operator()(double x) const { return m_y - m_f(x); }

	double 				 m_y;	///< The value y
    const DblFunction 	&m_f;	///< The function f(x)
};

/// InverseFunction is used to numerically approximate the inverse
/// x = g(y) of a function y = f(x)
class InverseFunction : public DblFunction {
  public:
	InverseFunction(const DblFunction& f, double a, double b, double eps)
		: m_f(f), m_a(a), m_b(b), m_eps(eps) {}
	double operator() (double y) const;

	const DblFunction  &m_f;		///< The function f(x).
	double				m_a;		///< The lower boundary of the search range.
	double				m_b;		///< The upper boundary of the search range.
	double				m_eps;		///< The maximum tolerance allowed.
};
#endif// __BORLANDC__

*/

class SplineAdapter : public DblFunction {
  public:
    SplineAdapter(const LinearSpline& spl) : m_spl(spl) {}
	double operator() (double x) const { return m_spl.value(x); }

	LinearSpline m_spl;
};

/// This function is a non-linear scaling function.
/// The function has the following defined values:
/// <ol><li>scale(x=0) = 0</li>
///     <li>scale(x=1) = 1</li>
///     <li>d(scale(x=0))/dx = 0</li>
///     <li>d(scale(x=1))/dx = 0</li></ol>
double scale(double x);

} // namespace IBK

#endif  // IBK_numerics_h
