// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
//
// Copyright (C) 2006 John Grunewald, John.Grunewald@gmx.de
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#ifndef IBK_para_test_h
#define IBK_para_test_h

#include <string>
#include <stdexcept>

#include "IBK_format.h"
#include "IBK_exception.h"


namespace IBK {

/// A class for simplified handling and particularly input of rational numbers.
///
/// This class is used to test limits of physical parameters that can be
/// of int, double, IBK:Parameter or any other arbitrary data type.
/// Tests of lower and upper limits are done, in addition test on equality
/// within a certain range +-eps is provided.
/// The template class para_test can be used as in the following example:
/// <pre>
///     // assume 'p1', "p2' and 'p3' are parameters of the type double or IBK::Parameter
///     // 'p1' must be > 100.0
///     // 'p2' must be in range of 0 .. 50 (equality allowed) and
///     // 'p3' must be == 10.0 +- 0.05
///     para_test<double> ptest1 = para_test<double>(para_test<double>::OP_GT, 100.0);
///     ptest1.test(p1);
///
///     para_test<double> ptest2 = para_test<double>(para_test<double>::OP_GE,  0.0,
///                                                  para_test<double>::OP_LE, 50.0);
///     ptest2.test(p2);
///
///     para_test<double> ptest3 = para_test<double>(para_test<double>::OP_EQ, 10.0,
///                                                  para_test<double>::OP_EQ, 0.05);
///     ptest3.test(p3);
/// </pre>
template <typename T>
class para_test {
  public:
    /// Operators used by para_test().
    enum oper_t {
        OP_LT,
        OP_LE,
        OP_EQ,
        OP_GE,
        OP_GT,
        OP_END
    };

  	/// Initialization constructor
    ///
    /// TODO : parameter description
  	para_test(oper_t min_op, T minval, oper_t max_op=OP_END, T maxval=T())
        : m_minOp(min_op), m_minVal(minval), m_maxOp(max_op), m_maxVal(maxval)
    {
        if (m_minOp == OP_EQ) {
            m_equVal = minval;
            m_eps = maxval;
        }
    }

    /// TODO : parameter description
    void test(const std::string& name, const T& val, const std::string& unit) const;

  private:

    /// TODO : parameter description
    void test(const std::string& name, oper_t op, const T& comp_val, const T& val,
              const std::string& unit) const;

    oper_t  m_minOp;     ///< descr
  	T       m_minVal;
    oper_t  m_maxOp;
  	T       m_maxVal;
  	T       m_equVal;
  	T       m_eps;
};


template <typename T>
void para_test<T>::test(const std::string& name, const T& value, const std::string& unit) const {
    //static const char * const FUNC_ID = "[para_test<T>::test]  ";

    // test min limit
    test(name, m_minOp, m_minVal, value, unit);
    // test max limit
    test(name, m_maxOp, m_maxVal, value, unit);
}

template <typename T>
void para_test<T>::test(const std::string& name, oper_t op, const T& comp_val, const T& val,
                        const std::string& unit) const {
    static const char * const FUNC_ID = "[para_test<T>::test]";

    switch (op) {
        case OP_GE :
            if (!(val >= comp_val)) {
				throw IBK::IBK_exception(format_string("Parameter '%' out of range (must be >= % %)") 
					% name % comp_val % unit, FUNC_ID);
            }
            break;

        case OP_GT :
            if (!(val > comp_val)) {
                throw IBK::IBK_exception(format_string("Parameter '%' out of range (must be > % %)") 
					% name % comp_val % unit, FUNC_ID);
            }
            break;

         case OP_EQ :
            if (!((val >= m_equVal-m_eps) && (val <= m_equVal+m_eps))) {
                throw IBK::IBK_exception(format_string("Parameter '%' out of range (must be == % % +- %)") 
					% name % m_equVal % unit % m_eps, FUNC_ID);
            }
            break;

        case OP_LE :
            if (!(val <= comp_val)) {
                throw IBK::IBK_exception( format_string("Parameter '%' out of range (must be <= % %)") 
					% name % comp_val % unit, FUNC_ID);
            }
            break;

        case OP_LT :
            if (!(val < comp_val)) {
				throw IBK::IBK_exception(format_string("Parameter '%' out of range (must be < % %)") 
					% name % comp_val % unit, FUNC_ID);
            }
            break;
        default : ;
    }
}


} // namespace IBK

/// @file IBK_para_test.h
/// @brief Contains the declaration of the class template para_test, a data type that
///        allows testing of lower and upper limits of int and double parameters.

#endif // IBK_para_test_h
