// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
// 
// Copyright (C) 2003 Andreas Nicolai, Andreas.Nicolai@gmx.net
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#include "IBK_configuration.h"

#ifdef __BORLANDC__
  #include "IBK_prehead.h"
  #pragma hdrstop
#else
  #include <stdexcept>
  #include <iostream>
  #include <iomanip>
#endif
#include "IBK_parameter.h"
#include "IBK_unitlistmanager.h"
#include "IBK_message_strings.h"
#include "IBK_exception.h"
using namespace std;

namespace IBK {

Parameter::Parameter(const std::string& str, double val, Unit unit)
    : name(str), value(val), IO_unit(unit)
{
    if (UnitList==NULL)  throw IBK_exception(IBK_UNITLIST_NOT_INITIALISED, "[Parameter::Parameter]");
    UnitList->convert( IO_unit, Unit(IO_unit.base_id()), value);
}
// ---------------------------------------------------------------------------

Parameter::Parameter(const std::string& str, double val, const std::string& ustr)
    : name(str), value(val), IO_unit(ustr)
{
    if (UnitList==NULL)  throw IBK_exception(IBK_UNITLIST_NOT_INITIALISED, "[Parameter::Parameter]");
    UnitList->convert( IO_unit, Unit(IO_unit.base_id()), value);
}
// ---------------------------------------------------------------------------

void Parameter::set(double val, Unit unit) {
    if (UnitList==NULL)  throw IBK_exception(IBK_UNITLIST_NOT_INITIALISED, "[Parameter::set]");
    value = val;
    IO_unit = unit;
    UnitList->convert( IO_unit, Unit(IO_unit.base_id()), value);
}
// ---------------------------------------------------------------------------

void Parameter::set(const std::string& str, double val, Unit unit) {
    name = str;
    set(val, unit);
}
// ---------------------------------------------------------------------------

void Parameter::set(const std::string& str, double val) {
    name = str;
    value = val;
    IO_unit.clear();
}
// ---------------------------------------------------------------------------

void Parameter::set(const std::string& str, const std::string& valstr) {
    stringstream lstrm(valstr);
    string ustr;
    lstrm >> value >> ustr;
    if (ustr.empty() || !lstrm)
        throw IBK_exception(format_string(IBK_PARAM_FROM_STRING_READ_ERROR) % valstr, "[Parameter::set]");
    set(str, value, ustr);
}
// ---------------------------------------------------------------------------

double Parameter::get_value(Unit unit) const {
    if (UnitList==NULL)  throw IBK_exception(IBK_UNITLIST_NOT_INITIALISED, "[Parameter::get_value]");
    double v=value; // copy parameter (don't change the actual base value)
    UnitList->convert(Unit(IO_unit.base_id()), unit, v);
    return v;
}
// ---------------------------------------------------------------------------

void Parameter::read(std::istream& in, bool no_name) {
    string name_, unit_;
    double val_;
    if (no_name)
        in >> val_ >> unit_;            // don't read name
    else
        in >> name_ >> val_ >> unit_;   // read name, too
    if (!in)
        throw IBK_exception( IBK_PARAMETER_READ_ERROR, "[Parameter::read]");
    set(val_, Unit(unit_));
    if (!no_name)  name = name_;        // assign read name
}
// ---------------------------------------------------------------------------

void Parameter::write(std::ostream& out, unsigned int indent, unsigned int paramWidth, bool writeName) const {
    if (UnitList==NULL)  throw IBK_exception(IBK_UNITLIST_NOT_INITIALISED, "[Parameter::write]");
    double v=value;
    UnitList->convert(Unit(IO_unit.base_id()), IO_unit, v);
    out << string(indent, ' ');
    if (name.size() && writeName) {
        if (paramWidth > name.size())
            out << setw(paramWidth) << left << name << " = ";
        else
            out << name << " = ";
    }
    out << v << " " << IO_unit.name() << endl;
}
// ---------------------------------------------------------------------------

void Parameter::clear() {
    name.clear();
    value=0;
    IO_unit=Unit(0);
}
//---------------------------------------------------------------------------

void Parameter::check(const std::string& ptxt,
	const std::string& unit, const IBK::para_test<double>& ptest) const
{
    const char * const FUNC_ID = "[Parameter::check]";

    // check parameter
    if (name.empty()) {
        std::string msg = ptxt + " is required";
        throw IBK_exception(msg, FUNC_ID);
    }
    // check unit
    if (IO_unit.base_id() != Unit(unit.c_str()).base_id()) {
        std::string msg = "Invalid or unrecognized unit (must be convertible into " + unit + ")";
        throw IBK_exception(msg, FUNC_ID);
    }
    // test limits
    ptest.test(name, value, unit);
}
// ---------------------------------------------------------------------------

}  // namespace IBK

