// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
// 
// Copyright (C) 2003 Andreas Nicolai, Andreas.Nicolai@gmx.net
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#ifndef IBK_parameter_h
#define IBK_parameter_h

#include <string>
#include "IBK_unit.h"
#include "IBK_para_test.h"

namespace IBK {

/// The class Parameter represents a physical parameter consisting of a descriptive
/// keyword, a value and unit used for input/output. Parameters can be used in
/// calculation programs for any kind of input parameters. The advantage of the
/// use of a parameter is, that the input data can be given in a unit that differs
/// from the SI base unit used for calculation. 
/// When created or set a parameter always converts the value given in the IO unit
/// into the internal SI base unit which can be retrieved via the 'value' member 
/// variable.
/// The member function get_value() can be used to retrieve the parameters value in 
/// the required unit.
class Parameter {
  public:
    /// Default constructor, creates an empty parameter with undefined 
    Parameter() : name(""), value(0), IO_unit(0) {};

    /// Constructor to initialise the parameter with name, value and IO unit.
    /// When you create a parameter like this the value 'val' will be converted into
    /// the base unit of 'unit', if available. For example when passing val=20 and 
    /// mm as unit the value of the parameter will be 0.02 afterwards.
    Parameter(const std::string& str, double val, const Unit unit=Unit(0));

    /// Constructor to initialise the parameter with name, value and IO unit.
    /// When you create a parameter like this the value 'val' will be converted into
    /// the base unit of 'unit', if available. For example when passing val=20 and 
    /// mm as unit the value of the parameter will be 0.02 afterwards.
    Parameter(const std::string& str, double val, const std::string& ustr);

    /// Sets value and unit with conversion.
    /// The value 'val' will be converted from 'unit' in the base unit of 'unit'.
    void set(double val, Unit unit);

    /// Sets name, value and unit with conversion.
    /// The value 'val' will be converted from 'unit' in the base unit of 'unit'.
    void set(const std::string& str, double val, Unit unit);

    /// Sets name and value without conversion.
    /// The value 'val' will not be converted and the unit will be set to 'undefined'.
    void set(const std::string& str, double val);

    /// Sets name, value and unit with conversion.
    /// The value 'val' will be converted from unit 'ustr' in the base unit of 'ustr'.
    void set(const std::string& str, double val, const std::string& ustr) { set(str, val, Unit(ustr)); }

    /// Sets the parameter by name and value, whereas valstr is composed of the actual value and the IO_unit.
    /// The value will be converted in the corresponding base unit. 
    void set(const std::string& str, const std::string& valstr);

    /// Returns the value of the parameter converted into its IO unit.
    double get_value() const { return get_value(IO_unit); }

    /// Returns the value of the parameter converted into 'unit'.
    /// The function throws a std::IBK_exception exception if the units cannot be related.
    double get_value(Unit unit) const;

    /// Returns the value of the parameter converted into unit 'ustr'.
    /// The function throws a std::IBK_exception exception if the units cannot be related.
    double get_value(const std::string& ustr) const { return get_value(Unit(ustr)); };

    /// Reads a parameter from the stream 'in'.
    /// The format for reading is: name value unit <br>
    /// When a parameter should be read without a name (usually a keyword), 
    /// pass 'true' as second parameter. 
    /// The value that has been read will be converted from the IO-unit to 
    /// the base unit.
    /// @param in       The input stream.
    /// @param no_name (optional) Flag that tells the read function to read only
    ///                value and unit.
    void read(std::istream& in, bool no_name=false);

    /// Writes a parameter into the output stream.
    /// The output format is: name value unit <br>
    /// If the parameter hasn't got a name, only the value and the unit is written.
    /// The value will be first converted into the IO unit and then written.
    /// The optional parameters can be used to specify the layout.
    /// @param out          The output stream.
    /// @param indent       (optional) Number of spaces, that should be put in front.
    /// @param paramwidth   (optional) Width of the parameter name (exclusive indentation), 
    ///                                a " = " is always appended
    /// @param writeName    (optional) Should the name (keyword) be written or not.
    void write(std::ostream& out, unsigned int indent=0, unsigned int paramWidth=0, bool writeName=true) const;

    /// Returns the current base unit of the parameter.
    Unit unit() const { return Unit(IO_unit.base_id()); };

    /// Clears the parameter.
    void clear();

    /// Checks the parameter.
    void check(const std::string& ptxt,
        const std::string& unit, const IBK::para_test<double>& ptest) const;

    // ****** member variables *************************************************

    /// The descriptive keyword/name of the parameter.
    std::string name;
    /// The value of the parameter for calculations (always in SI base unit).
    double value;
    /// Input and output unit for the parameter.
    Unit IO_unit;
};

}  // namespace IBK

/// @file IBK_parameter.h
/// @brief Contains the declaration of the class Parameter, a class for a physical parameter 
///        with name, value and assiciates IO unit.

#endif  // IBK_parameter_h
