// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
// 
// Copyright (C) 2003 Andreas Nicolai, Andreas.Nicolai@gmx.net
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#ifndef IBK_point_h
#define IBK_point_h

#include <ostream>

namespace IBK {

/// The struct point2D represents a 2D point consisting of an x and y coordinate.
template <typename T>
struct point2D {
    /// Default constructor.
    point2D() : x(T()), y(T()) {};
    /// Constructor (initialises with coordinates 'a' and 'b')
    point2D(T a, T b) : x(a), y(b) {};
    /// Sets both coordinates.
    void set(T a, T b) { x=a; y=b; };

    T x;   ///< The X-coordinate.
    T y;   ///< The Y-coordinate.
};

/// Returns 'true' if the points 'lhs' and 'rhs' have the same coordinates.
template <typename T>
inline bool operator==(const point2D<T>& lhs, const point2D<T>& rhs) { return lhs.x == rhs.x && lhs.y == rhs.y; };
/// Returns 'true' if the points 'lhs' and 'rhs' do not have the same coordinates.
template <typename T>
inline bool operator!=(const point2D<T>& lhs, const point2D<T>& rhs) { return lhs.x != rhs.x || lhs.y != rhs.y; };

/// Stream output operator for point2D objects.
/// The point is written into the output stream in the format: 'x,y'.
template <typename T>
inline std::ostream& operator<<(std::ostream& out, const point2D<T>& p) {
    return out << p.x << ',' << p.y;
};

// =============================================================================


/// The struct point3D represents a 3D point consisting of an x, y and z coordinate.
template <typename T>
struct point3D {
    /// Default constructor.
    point3D() : x(T()), y(T()), z(T()) {};
    /// Constructor (initialises with coordinates 'a', 'b' and 'c')
    point3D(T a, T b, T c) : x(a), y(b), z(c) {};
    /// Sets all coordinates. 
    void set(T a, T b, T c) { x=a; y=b; z=c; };

    T x;   ///< The X-coordinate.
    T y;   ///< The Y-coordinate.
    T z;   ///< The Z-coordinate.
};

} // namespace IBK

/// @file IBK_point.h
/// @brief Contains the declarations of the structs point2D and point3D.

#endif // IBK_point_h

