// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
// 
// Copyright (C) 2003 Andreas Nicolai, Andreas.Nicolai@gmx.net
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#ifndef IBK_rectangle_h
#define IBK_rectangle_h

#include <iostream>
#include <sstream>

namespace IBK {

/// A type that represents a rectangle. 
/// The class can be used whenever ccordinates describing a rectangle have to 
/// be stored together, such as for screen coordinates or when describing a 
/// geometrical shape.
template <class T>
class rectangle {
  public:
    /// Default contructor (initialises the rectangle with default value of type T).
    rectangle() : left(T()), top(T()), right(T()), bottom(T()) {};
    /// Constructor (initialises the rectangle with given parameters for the coordinates).
    rectangle(T l, T t, T r, T b) : left(l), top(t), right(r), bottom(b) {};
    /// Sets all four coordinates with one function call (essentially behaves like 'rect = rectangle(l,r,t,b)')
    void set(T l, T t, T r, T b) { left=l; top=t; right=r; bottom=b; };

    /// Returns the width (that is difference between the coordinates 'right' and 'left')
    T width()   const   { return right - left; };
    /// Returns the height (that is difference between the coordinates 'bottom' and 'top')
    T height()  const   { return bottom - top; };

    T left;     ///< The left coordinate.
    T top;      ///< The top coordinate.
    T right;    ///< The right coordinate.
    T bottom;   ///< The bottom coordinate.
};

/// Writes the coordinates of the rectangle 'rect' to the output stream 'out' 
/// (the coordinates are written in the following order: left, top, right, bottom).
template <class T>
inline std::ostream& operator<<(std::ostream& out, const rectangle<T>& rect) {
    std::stringstream strm;
    strm << rect.left << " " << rect.top << " " << rect.right << " " << rect.bottom;
    return out << strm.str();
};

/// Reads the coordinates of the rectangle 'rect' from the input stream 'in' 
/// (the coordinates are expected in the following order: left, top, right, bottom).
/// The coordinates of 'rect' are only modified if ALL coordinates could be read
/// correctly.
template <class T>
inline std::istream& operator>>(std::istream& in, rectangle<T>& rect) {
    rectangle<T> tmp;
    if (in >> tmp.left >> tmp.top >> tmp.right >> tmp.bottom) 
        std::swap(tmp,rect);
    return in;
};

/// Compares two rectangles 'lhs' and 'rhs' and returns 'true' if they have the 
/// same coordinates.
template <class T>
inline bool operator==(const rectangle<T>& lhs, const rectangle<T>& rhs) {
    return (lhs.left == rhs.left && lhs.top==rhs.top && lhs.right==rhs.right && lhs.bottom==rhs.bottom); }

} // namespace IBK

#endif  // IBK_rectangle_h
