// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
// 
// Copyright (C) 2003 Andreas Nicolai, Andreas.Nicolai@gmx.net
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#ifndef IBK_simplematrix_h
#define IBK_simplematrix_h

#include "IBK_configuration.h"

#include <vector>
#include <stdexcept>
#include <cmath>
#include "IBK_matrix.h"
#include "IBK_format.h"
#include "IBK_message_strings.h"

namespace IBK {

/// A simple matrix implementation.
/// This is an implementation for a general dense matrix and supports element access.
/// Resizing is possible, see resize() for details.
/// The data is internally stored row-based in a contiguous memory block, so the element
/// 5,2 has the index [2*cols + 5] in the linear data array accessible via the iterator
/// returned from the member function begin().
/// Since the data of a simple matrix is stored in a single continuous memory block, the
/// dimensions of the matrix is limited (depending on the capabilities of the operating
/// system). If you use large dense matrices, take the dense_matrix class from the
/// IBK library.
template <typename T>
class simple_matrix : public matrix<T> {
  public:
    typedef T                                           value_type;         ///< The type of the values.
    typedef T&                                          reference;          ///< Reference to one element.
    typedef const T&                                    const_reference;    ///< Constant reference to one element.
    typedef typename std::vector<T>::iterator           iterator;           ///< Iterator to one element.
    typedef typename std::vector<T>::const_iterator     const_iterator;     ///< Constant iterator to one element.

    /// Default construction.
    /// Creates an empty matrix.
    simple_matrix() : rows_(0), cols_(0) {};
    /// Initialisation constructor.
    /// Creates a matrix with the dimensions: cols x rows
    simple_matrix(std::size_t cols, std::size_t rows);
    /// Initialisation constructor with initial value.
    /// Creates a matrix with the dimensions: rows x cols and sets all elements to 'value'.
    simple_matrix(std::size_t cols, std::size_t rows, const T& value);

#ifndef IBK_ENABLE_RANGE_CHECKING
    /// Returns a reference to the matrix element at [col][row].
    /// If IBK_ENABLE_RANGE_CHECKING is defined it throws an out_of_range exception if the
    /// indices are invalid.
    reference  operator() (std::size_t col, std::size_t row) { return data_[row][col]; };
    /// Returns a constant reference to the matrix element at [col][row].
    /// If IBK_ENABLE_RANGE_CHECKING is defined it throws an out_of_range exception if the
    /// indices are invalid.
    const_reference  operator() (std::size_t col, std::size_t row) const { return data_[row][col]; };
#else
    /// Returns a reference to the matrix element at [col][row].
    /// If IBK_ENABLE_RANGE_CHECKING is defined it throws an out_of_range exception if the
    /// indices are invalid.
    reference  operator() (std::size_t col, std::size_t row);
    /// Returns a constant reference to the matrix element at [col][row].
    /// If IBK_ENABLE_RANGE_CHECKING is defined it throws an out_of_range exception if the
    /// indices are invalid.
    const_reference  operator() (std::size_t col, std::size_t row) const;
#endif // IBK_ENABLE_RANGE_CHECKING

    /// Returns a reference to the matrix element at [col][row].
    /// It throws an std::out_of_range exception when the indices are invalid.
    reference  at(std::size_t col, std::size_t row);
    /// Returns a constant reference to the matrix element at [col][row].
    /// It throws an std::out_of_range exception when the indices are invalid.
    const_reference  at(std::size_t col, std::size_t row) const;
    /// Returns the number of rows in the matrix (must be reimplemented in derived classes).
    std::size_t rows() const { return rows_; };
    /// Returns the number of columns in the matrix (must be reimplemented in derived classes).
    std::size_t cols() const { return cols_; };

    /// Resizes the matrix (all data is lost).
    void resize(std::size_t col, std::size_t row);
    /// Swaps the content of the matrix with the content of simple matrix 'mat'.
    void swap(simple_matrix<T>& mat);
    /// Empties the matrix and sets its dimensions to 0.
    void clear() { data_.clear(); rows_ = cols_ = 0; };
    /// Fills the matrix with the given value
    void fill(const T& value);
	/// Sets the matrix values to zero
	void setZero();

    /// Solves the linear system of equations A*x = d with the GAUSS algorithm (with pivoting) and
    /// stores the resultant x vector in the place of vector d. The matrix is hereby modified.
    /// The parameter workspace must point to a memory block of size n.
    /// @return The return codes are as follows: <ul>
    /// <li>Return value = 0 : All ok</li>
    /// <li>Return value = 1 : Error - Invalid dimensions</li>
    /// <li>Return value = 2 : Error - Matrix is singular</li>
    /// </ul>
    unsigned int solve_gauss_pivot(T * d, T * workspace);
    /// Convenience function, essentially behaves like the function above. Make sure the vector has
    /// the right size!
    unsigned int solve_gauss_pivot(std::vector<T>& d, std::vector<T>& workspace) {
        return solve_gauss_pivot(&d[0], &workspace[0]);
    };

  private:
    std::vector<std::vector<T> >  data_;     // the main data member, stores data as [row][col]
    std::size_t     rows_;     // the rows of the matrix
    std::size_t     cols_;     // the columns of the matrix
};
// ---------------------------------------------------------------------------

template <typename T>
simple_matrix<T>::simple_matrix(std::size_t cols, std::size_t rows)
	: data_(rows), rows_(rows), cols_(cols) 
{
	for (std::size_t i=0; i<rows_; ++i)
		data_[i].resize(cols_);
};

template <typename T>
simple_matrix<T>::simple_matrix(std::size_t cols, std::size_t rows, const T& value)
	: data_(rows), rows_(rows), cols_(cols) 
{
	for (std::size_t i=0; i<rows_; ++i)
		data_[i].resize(cols_,value);
};

#ifdef IBK_ENABLE_RANGE_CHECKING
template <typename T>
inline typename simple_matrix<T>::reference simple_matrix<T>::operator() (std::size_t col, std::size_t row) {
    if (col>=cols_ || row>=rows_)
        throw std::out_of_range("[IBK::simple_matrix::operator() ]  " +
            format_string(IBK_MATRIX_INDEX_OUT_OF_RANGE) % col % row % cols_ % rows_);
    return data_[row][col];
};

template <typename T>
inline typename simple_matrix<T>::const_reference  simple_matrix<T>::operator() (std::size_t col, std::size_t row) const {
    if (col>=cols_ || row>=rows_)
        throw std::out_of_range("[IBK::simple_matrix::operator() const]  " +
            format_string(IBK_MATRIX_INDEX_OUT_OF_RANGE) % col % row % cols_ % rows_);
    return data_[row][col];
};
#endif // IBK_ENABLE_RANGE_CHECKING

template <typename T>
inline typename simple_matrix<T>::reference simple_matrix<T>::at(std::size_t col, std::size_t row) {
    if (col>=cols_ || row>=rows_)
        throw std::out_of_range("[IBK::simple_matrix::at() ]  " +
            format_string(IBK_MATRIX_INDEX_OUT_OF_RANGE) % col % row % cols_ % rows_);
    return data_[row][col];
};

template <typename T>
inline typename simple_matrix<T>::const_reference  simple_matrix<T>::at(std::size_t col, std::size_t row) const {
    if (col>=cols_ || row>=rows_)
        throw std::out_of_range("[IBK::simple_matrix::at() const]  " +
            format_string(IBK_MATRIX_INDEX_OUT_OF_RANGE) % col % row % cols_ % rows_);
    return data_[row][col];
};

template <typename T>
void simple_matrix<T>::resize(std::size_t cols, std::size_t rows) {
    cols_ = cols;
    rows_ = rows;
    data_.resize(rows_);
	for (std::size_t i=0; i<rows_; ++i)
		data_[i].resize(cols_);
};

template <typename T>
inline void simple_matrix<T>::swap(simple_matrix<T>& mat) {
    std::swap(cols_, mat.cols_);
    std::swap(rows_, mat.rows_);
    data_.swap(mat.data_);    // this is faster then normal swapping!
};

template <typename T>
void simple_matrix<T>::fill(const T& value) {
	for (std::size_t i=0; i<rows_; ++i)
		std::fill(data_[i].begin(), data_[i].end(), value);
}

template <typename T>
void simple_matrix<T>::setZero() {
	for (std::size_t i=0; i<rows_; ++i)
		std::memset(&data_[i][0], 0, cols_*sizeof(T));
}

template <typename T>
unsigned int simple_matrix<T>::solve_gauss_pivot(T * d, T * workspace) {
	T * current_row;
	T pivot, tmp, sum, mult, pivot_1;
	if (cols_ != rows_) return 1;  // invalid dimensions
    // 1. elimination into upper triagonal matrix (n = cols_ = rows_)
    for (unsigned int i=0; i<cols_-1; ++i) {
        // find Pivot element
        unsigned int pivot_index = i;
		pivot = std::fabs(data_[i][i]);
        for (unsigned int p=i+1; p<cols_; ++p) {
            tmp = std::fabs(data_[p][i]);
            if (pivot < tmp) {
                pivot = tmp;
                pivot_index = p;
            }
        }
        // if Pivot element is zero, bail out
        if (pivot==0)   {
			return 2;  // matrix singular
        }
        // if Pivot index is different from current row, swap rows
        if (pivot_index != i) {
            unsigned int bytes = static_cast<unsigned int>((cols_-i)*sizeof(T));
            T * src = &data_[pivot_index][i];
            pivot = *src;
            memcpy(workspace, src, bytes);
            memcpy(src, &data_[i][i], bytes);
            memcpy(&data_[i][i], workspace, bytes);
            std::swap(d[pivot_index], d[i]);
        }
        // now modify all rows below the i-th row
		pivot_1 = 1.0/pivot;
        for (unsigned int j=i+1; j<rows_; ++j) {
			current_row = &data_[j][0];
            mult = current_row[i]*pivot_1;  // must be smaller then 1 now
            current_row[i]=0;
            if (mult==0) continue;        // row remains unchanged
            for (unsigned int k=i+1; k<cols_; ++k)
                current_row[k] -= data_[i][k]*mult;
            d[j] -= d[i]*mult;
        }
    }
    // 2. backward substitution
    for (int i=static_cast<int>(cols_)-1; i>=0; --i) {
		current_row = &data_[i][0];
        if (!current_row[i])  return 2;  // matrix singular
        sum = 0;
        for (unsigned int j=i+1; j<cols_; ++j)
            sum += d[j]*current_row[j];
        d[i] = (d[i] - sum)/current_row[i];
    }
    return 0;
}
// ---------------------------------------------------------------------------

} // namespace IBK

#endif // IBK_simplematrix_h

/// @file IBK_simplematrix.h
/// @brief Contains the class template simple_matrix for general dense matrices.
