// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
// 
// Copyright (C) 2003 Andreas Nicolai, Andreas.Nicolai@gmx.net
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#include "IBK_configuration.h"

#ifdef __BORLANDC__
  #include "IBK_prehead.h"
  #pragma hdrstop
#else
  #include <stdexcept>
  #include <ctime>
  #include <cmath>
  #include <iostream>
#endif
#include "IBK_splinegeneration.h"
#include "IBK_exception.h"
using namespace std;

namespace IBK {

int generate_spline(std::vector<double>& x, std::vector<double>& y, DblFunction& f,
	double x1, double x2, double tolerance, double dx, unsigned int max_points,
	unsigned int iterations, bool cont_iter, bool clean_up)
{
	if (!cont_iter) {
        // empty the vectors and reserve enough memory to avoid reallocations
        x.clear();
        y.clear();
        x.reserve(max_points);
        y.reserve(max_points);

        // *** PART 1 ***
        // at first generate a spline using constant step sizes
        if (x1 > x2)  std::swap(x1, x2);
        if (dx <= 0)
            dx = (x2 - x1)/10;
        double cx = x1;  // "current x" -> cx
        while (cx < x2) {
            x.push_back( cx );
            y.push_back( f(cx) );
            cx += dx;
        }
        x.push_back( x2 );
        y.push_back( f(x2) );
	}
    
    // *** PART 2 ***
    // now we start with the first iteration cycle where we add new points whereever the
    // tolerance is exceeded
    size_t n=iterations; // maximum iterations
    double xm, ym, ym_approx, dy;
    std::vector<double> nx, ny;
	nx.reserve(max_points);
	ny.reserve(max_points);
    int result_code = 0;
	clock_t next_notify = clock() + CLOCKS_PER_SEC;
    while (--n && result_code==0) {
		double max_dy = tolerance;
        nx.clear();
        ny.clear();
        // now first introduce new points where necessary
        unsigned int start_index = 1;
        while (x[start_index] < x1 && start_index < x.size()-1) {
        	nx.push_back(x[start_index-1]);
        	ny.push_back(y[start_index-1]);
        	++start_index;
        }
        if (start_index==x.size()-1) return 3; // out of range for continuation
        unsigned int end_index = static_cast<unsigned int>(x.size()-1);
        while (x[end_index] > x2 && end_index > 1) --end_index;
        if (end_index==1) return 3; // out of range for continuation
        for (size_t i=start_index; i<=end_index; ++i) {
            // add the last point (left point of the interval)
            nx.push_back(x[i-1]);
            ny.push_back(y[i-1]);
            // calculate the exact and approximated y values at the middle 
            xm = (x[i-1]+x[i])/2;
            ym = f(xm);
            dx = x[i]-x[i-1];
            if (dx==0)
                throw IBK_exception("Division by zero while adding spline points!","[generate_spline]");
            ym_approx = y[i-1] + (y[i]-y[i-1])/dx*(xm-x[i-1]);
           	dy = fabs(ym - ym_approx);
            // if tolerances are exceeded add the middle point, but only if dx is not becoming
			// too small
            if (dy > tolerance) {
				max_dy = std::max(dy, max_dy);
                nx.push_back(xm);
                ny.push_back(ym);
				if (nx.size() >= max_points)  {
                	result_code = 1;
                }
            }
			if (clock() > next_notify) {
				cout << '.';
				next_notify += CLOCKS_PER_SEC;
			}

        }
        // add values outside the interval
        for (unsigned int i=end_index; i<x.size(); ++i) {
        	nx.push_back(x[i]);
        	ny.push_back(y[i]);
        }
        // store the last value
        //nx.push_back(x.back());
        //ny.push_back(y.back());

        // if nothing changed since last iteration cycle, break: we are finished
        if (equal(nx.begin(), nx.end(), x.begin()) &&
            equal(ny.begin(), ny.end(), y.begin()) )
        {
            break;
        }
        x.swap(nx);
        y.swap(ny);
		cout.flush();
    }

    if (!clean_up) {
    	if (n==0)  return 2;
		return result_code;
	}
    // *** PART 3 ***
    // now remove all points that could be calculed via linear interpolation
    // between the neighbouring points within the allowed tolerances
    nx.clear();
    ny.clear();
    nx.push_back(x.front());
    ny.push_back(y.front());
    size_t xsize = x.size()-1;
    double y_approx;
    for (size_t i=1; i<xsize; ++i) {
        // calculate approximated value at point i using linear interpolation between
        // the last stored point and point i+1
        dx = x[i+1]-nx.back();
        if (dx==0)
            throw IBK_exception("Division by zero while skipping spline points!","[generate_spline]");
        y_approx = ny.back() + (y[i+1]-ny.back())/dx*(x[i+1]-x[i]);
      	dy = fabs(f(x[i]) - y_approx);
        // if tolerances are exceeded we must keep the point
        if (dy > tolerance) {
            nx.push_back(x[i]);
            ny.push_back(y[i]);
        }
    }
    nx.push_back(x.back());
    ny.push_back(y.back());
    x.swap(nx);
    y.swap(ny);
    if (n==0)  return 2;
	return result_code;
}

}  // namespace IBK

