// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
// 
// Copyright (C) 2003 Andreas Nicolai, Andreas.Nicolai@gmx.net
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#ifndef IBK_spline_generation_h
#define IBK_spline_generation_h

#include <vector>
#include "IBK_types.h"

namespace IBK {

/// Generates the tabulated data for a linear spline using an arbitrary function.
/// The resulting spline data is stored in the vectors x and y. The spline 
/// generation is a 3 step process. At first points are calculated using the 
/// original function at regular intervals given by the parameter 'dx'. 
/// Then values in the middle of each interval are calculated and compared with 
/// the original values. If the difference exceeds the allowed tolerance a new point 
/// is calculated and inserted at the middle of the interval. This procedure is 
/// repeated for all intervals, until the tolerance is nowhere exceeded. In the
/// 3rd and final step all intervals are checked again and all points are removed
/// than can also be calculated via linear interpolation within the given tolerance.
/// The result is a spline that contains just as many points as necessary to 
/// calculate any point in the given range of the spline with the required
/// tolerance. If you set the initial step size to 0, the function will start with
/// a dx = (x1 - x2)/10.
/// @param x			The vector for the generated x-coordinates.
/// @param y			The vector for the generated y-coordinates.
/// @param f            The orignal function encapsulated in a function object derived from DblFunction.
/// @param x1			The lower bound for the range of x coordinates.
/// @param x2			The upper bound for the range of x coordinates.
/// @param tolerance    The maximum relative tolerance allowed in the final spline.
/// @param dx           Initial step size for calculating the first spline values.
/// @param max_points   What is the maximum number of points in the final spline. If exceeded
///                     the function will return with the return code 1.
/// @param iterations   This indicates how many iterations should be performed to 
///                     find the optimal solution. 100 iterations is normally a
///                     good value. Depending on the evaluation speed of the function object f
///                     it is recommanded to start with a lower value. If the iteration limit 
///                     is exceeded the function will return with the return code 2.
/// @return				If the algorithm fails the function will return an error code: 0=all ok,
///						1=exceeded max spline points, 2=exceeded iteration limit.
int generate_spline(std::vector<double>& x, std::vector<double>& y, DblFunction& f,
	double x1, double x2, double tolerance, double dx, unsigned int max_points,
	unsigned int iterations, bool cont_iter = false, bool clean_up = true);

}  // namespace IBK

/// @file IBK_splinegeneration.h
/// @brief Contains the linear spline generation function.

/// @example splinegeneration_example.cpp
/// This is an example of how to use the function IBK::generate_spline to create 
/// optimized linear splines.

#endif  // IBK_spline_generation_h
