// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
//
// Copyright (C) 2003 Andreas Nicolai, Andreas.Nicolai@gmx.net
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#ifndef IBK_stopwatch_h
#define IBK_stopwatch_h

#include <string>
#include <ctime>    // for std::clock()

namespace IBK {

/// The StopWatch class can be used to measure the time used during execution of
/// certain program parts.
/// Every StopWatch object remembers its time and state. So you can use several
/// stop watches simultaneously (for instance a global watch and a watch for
/// certain subroutines).
class StopWatch {
  public:
    /// Default constructor, creates a stop watch and starts it.
    StopWatch() : start_(std::clock()), stop_(0) {};
    /// Restarts the clock.
    std::clock_t start();
    /// Stops the clock and remembers the elapsed time.
    double stop();
    /// Returns the time so far in milli seconds.
    double difference() const;
    /// Returns the time so far as a string with unit (either in milli seconds or seconds).
    const std::string diff_str() const;
    /// Returns the time so far as a string but allows the specification of the width of
    /// the resulting string. The time value will be align to the right.
    const std::string diff_str(std::size_t width) const;

  private:
    std::clock_t start_, stop_;
};

} // namespace IBK

/// @file IBK_stopwatch.h
/// @brief Contains the declaration of the class StopWatch.

/// @example stopwatch_example.cpp
/// This is an example of how to use the class IBK::StopWatch.

#endif  // IBK_stopwatch_h
