// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
//
// Copyright (C) 2003 Andreas Nicolai, Andreas.Nicolai@gmx.net
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#include "IBK_configuration.h"

#ifdef __BORLANDC__
  #include "IBK_prehead.h"
  #pragma hdrstop
#else
  #include <iostream>
  #include <iomanip>
  #include <sstream>
  #include <algorithm>
#endif
#include "IBK_stringutils.h"
#include "IBK_unit.h"
#include "IBK_unitlistmanager.h"
#include "IBK_messages.h"
#include "IBK_keyword.h"
#include "IBK_exception.h"
using namespace std;

namespace IBK {

void split_filename(std::string fullname, std::string& name, std::string& dir) {
    std::replace(fullname.begin(), fullname.end(), '\\', '/');
    size_t pos=fullname.rfind('/');
    if (pos!=string::npos) {
        dir=fullname.substr(0,++pos);
        if (pos<fullname.size()) name=fullname.substr(pos);
        else name="";
    } else {
        name=fullname;
        dir="";
    }
}
// ---------------------------------------------------------------------------

/// Trims the string on both sides (removes tabs and spaces)
std::string& trim(std::string& str) {
    size_t f = str.find_first_not_of(" \t\r");
    size_t l = str.find_last_not_of(" \t\r");
    if (f!=string::npos && l!=string::npos)
        str = str.substr(f,l-f+1);
    else
        str.clear();
    return str;
}
// ---------------------------------------------------------------------------

/// Removes comments from the str (removes everything after and including the ; )
std::string& remove_comment(std::string& str) {
    std::size_t pos = str.find(";");
    if ((pos!=std::string::npos) && (pos<str.length()))
        str.erase(pos);
    return str;
}
// ---------------------------------------------------------------------------

/// Compares the strings case in-sensitve.
bool string_nocase_compare(const std::string& lhs, const std::string& rhs) {
    if (lhs.size()!=rhs.size()) return false;
    const char* s1=lhs.data();
    const char* s2=rhs.data();
    size_t n=lhs.size()+1;
    while (tolower(*s1++)==tolower(*s2++) && --n);
    return n==0;
}
// ---------------------------------------------------------------------------

/// Finds the string 'substr' in the string 'str' (case insensitive).
/// Returns either the position of the substr or string::npos.
std::size_t string_nocase_find(const std::string& str, const std::string& substr) {
    if (str.empty())    return string::npos;
    if (substr.empty()) return 0;
    const char* s1=str.data();
    const char* s2=substr.data();
    size_t size1=str.size();
    size_t pos = 0;
    while (pos<size1) {
        if (tolower(*s1)==tolower(*s2)) {
            size_t size2=substr.size();
            const char* s3=s1;
            while (--size2 && tolower(*(++s3))==tolower(*(++s2)));
            if (!size2) return pos;
            s2 = substr.data(); // prepare for next trial
        };
        ++s1;
        ++pos;
    }
    if (pos==size1) return string::npos;
    else            return pos;
}
// ---------------------------------------------------------------------------

/// Extracts keyword and value from string "keyword = value"
bool extract_keyword(const std::string& line, std::string& keyword, std::string& value) {
    size_t pos = line.find('=');
    if (pos==string::npos) {
        value = line;
        keyword.clear();
        return false;
    };
    keyword = line.substr(0, pos-1);
    trim(keyword);
    value = line.substr(pos+1);
    trim(value);
    return true;
}
// ---------------------------------------------------------------------------

/// Returns the relative path of the file 'filename' to the relative path 'relpath'.
std::string relative_path(const std::string& filename, const std::string& relpath) {
    string fname(filename), name, fpath;
    trim(fname);
    if (fname.empty())
        throw IBK_exception("Got empty file path!","[relative_path]");
    replace(fname.begin(), fname.end(), '\\', '/');
    split_filename(fname, name, fpath);
    if (fpath.rfind('/')==fpath.size()-1)  fpath.erase(fpath.size()-1);
    string path(relpath);
    trim(path);
    if (path.empty())
        throw IBK_exception("Got empty path!","[relative_path]");
    replace(path.begin(), path.end(), '\\', '/');
    if (path.rfind('/')==path.size()-1)  path.erase(path.size()-1);
    // make a backup copy of the old string
    string oldfpath = fpath;
    while (string_nocase_find(path, fpath)==string::npos) {
        size_t pos = fpath.rfind('/');
        if (pos!=string::npos)
            fpath.erase(pos);
        else
            // not possible to find relative path, probably different drives
            return filename;
    };
    string postStr = oldfpath.substr(fpath.size())+'/'+name;
    // now go back the path until path is equal to fpath
    string preStr;
    while (!string_nocase_compare(path, fpath)) {
        size_t pos = path.rfind('/');
        if (pos!=string::npos) {
            path.erase(pos);
            preStr+="../";
        }
        else
            // not possible to find relative path, probably different drives
            return filename;
    };
    if (!preStr.empty())
        preStr.erase(preStr.size()-1);
    return preStr + postStr;
}
// ---------------------------------------------------------------------------

std::string remove_extension(const std::string& filename) {
	size_t pos = filename.rfind(".");
    return filename.substr(0, pos);
}
// ---------------------------------------------------------------------------

void explode(const std::string& str, std::list<std::string>& tokens, char delim) {
    string tmp;
    tokens.clear();
    for (std::string::const_iterator it=str.begin(); it!=str.end(); ++it) {
        if (*it!=delim)
            tmp+=*it;
        else {
            if (tmp.empty()) continue;
            tokens.push_back(tmp);
            tmp.clear();
        }
    }
    if (tmp.size())
        tokens.push_back(tmp);
}
// ---------------------------------------------------------------------------

void explode(const std::string& str, std::list<std::string>& tokens,
	std::string& delims, const std::string& delimiter)
{
    string tmp;
    tokens.clear();
    delims.clear();
    int delim_count = 0;
    delims.resize(500); // 500 delimiters should be enough, right?
    for (std::string::const_iterator it=str.begin(); it!=str.end() && delim_count < 500; ++it) {
    	bool delim_found = false;
    	for (std::string::const_iterator tit = delimiter.begin(); tit != delimiter.end(); ++tit) {
        	if (*it==*tit) {
            	tokens.push_back(tmp);
                delims[delim_count++] = *tit;
            	tmp.clear();
                delim_found = true;
                break;
        	}
        }
        if (!delim_found) tmp += *it;
    }
    if (!tmp.empty()) {
        tokens.push_back(tmp);
        delims[delim_count] = ' ';
    }
}
// ---------------------------------------------------------------------------

void explode_csv(const std::string& str, std::list<std::string>& tokens) {
    string tmp;
    tokens.clear();
    bool in_quotes = false;
    for (std::string::const_iterator it=str.begin(); it!=str.end(); ++it) {
    	if (*it == '\"')  {
        	in_quotes = !in_quotes;
            tmp += *it;
        }
        else if (!in_quotes && *it == ',') {
            tokens.push_back(tmp);
            tmp.clear();
        }
        else {
            tmp+=*it;
        }
    }
    if (tmp.size())
        tokens.push_back(tmp);
}
// ---------------------------------------------------------------------------

void explode_sections(const std::string& str, const std::vector<std::string>& section_titles,
	std::vector<std::string>& section_data)
{
    size_t current_section = section_titles.size();
    section_data.clear();
    section_data.resize(section_titles.size());
    stringstream strm(str);
    string line, keyword;
    while (getline(strm, line)) {
        keyword = trim(remove_comment(line)); // 'line' will be changed here!
        trim_keyword(keyword);
       	vector<string>::const_iterator it = find(section_titles.begin(), section_titles.end(), keyword);
       	// check for new keyword
        if (!line.empty() && it != section_titles.end()) {
       		current_section = distance(section_titles.begin(), it);
        }
        else {
            // skip empty lines
            if (line.empty())  continue;
            // if no list keyword has been found yet, throw exception
            if (current_section==section_titles.size())
                throw IBK_exception("Expected section header in first non-empty line:\n'" + line + "'","[IBK::explode]");
            // add string line to current part
            section_data[current_section] += line + '\n';
        }
    }
}
// ---------------------------------------------------------------------------

void explode_section(const std::string& str, const std::string& section_title,
	std::vector<std::string>& section_data)
{
    stringstream in(str);
    string line, keyword;
    section_data.clear();
    while (getline(in, line)) {
        keyword = trim(remove_comment(line)); // 'line' will be changed here!
        if (line.empty()) continue;
        trim_keyword(keyword);
        if (keyword==section_title)
            section_data.push_back(""); // add a new section
        else {
            if (section_data.empty())
                throw IBK_exception("Expected keyword/section title!","[explode_section]");
            // add string line to current section data
            section_data.back() += line + '\n';
        }
    }
}
// ---------------------------------------------------------------------------

void explode_section(const std::string& data, std::vector<std::string>& section_data)
{
	stringstream in(data);
    string line;
    section_data.clear();
    while (getline(in, line)) section_data.push_back(line + '\n');
}
// ---------------------------------------------------------------------------

std::string format_double(double val, int precision, ios_base::fmtflags fmt) {
	stringstream strm;
	strm.setf(fmt);
	strm << setprecision(precision) << val;
	return strm.str();
}
// ---------------------------------------------------------------------------

std::string format_time_duration(unsigned int t, bool shortformat) {
    stringstream strm;
    unsigned int tmp;
    if (shortformat) {
        tmp = t / 86400;
        if (tmp > 0) {
            strm << setw(4) << right << tmp << " d";
            t %= 86400;
        }
        tmp = t / 3600;
        if (tmp > 0) {
            if (!strm.str().empty())  strm << "  ";
            strm << setw(2) << right << tmp << " h";
            t %= 3600;
        }
        tmp = t / 60;
        if (tmp > 0) {
            if (!strm.str().empty())  strm << "  ";
            strm << setw(2) << right << tmp << " min";
            t %= 60;
        }
        if (t > 0) {
            if (!strm.str().empty())  strm << "  ";
            strm << setw(2) << right << t << " s";
        }
    }
    else {
        tmp = t / 86400;
        if (tmp > 0) {
            strm << setw(4) << right << tmp << " day" << (tmp==1 ? " " : "s");
            t %= 86400;
        }
        tmp = t / 3600;
        if (tmp > 0) {
            if (!strm.str().empty())  strm << ", ";
            strm << setw(2) << right << tmp << " hour" << (tmp==1 ? " " : "s");
            t %= 3600;
        }
        tmp = t / 60;
        if (tmp > 0) {
            if (!strm.str().empty())  strm << ", ";
            strm << setw(2) << right << tmp << " minute" << (tmp==1 ? " " : "s");
            t %= 60;
        }
        if (t > 0) {
            if (!strm.str().empty())  strm << " and ";
            strm << setw(2) << right << t << " second" << (t==1 ? " " : "s");
        }
    }
    return strm.str();
}
// ---------------------------------------------------------------------------

std::string format_time_difference(const std::string& ustr, double delta_t) {
    if (UnitList==NULL)
        throw IBK_exception("Unitlist not initialised!","[format_time_difference]");
    UnitList->convert(Unit("s"), Unit(ustr), delta_t);
    stringstream strm;
    strm << delta_t;
    return strm.str();
}
// ---------------------------------------------------------------------------

std::string format_time_difference(double delta_t) {
    stringstream strm;
    strm << fixed << setprecision(3) << setw(6) << right;
    if (delta_t<60)
        strm << delta_t << " s";
    else if (delta_t<3600)
        strm << delta_t/60.0 << " min";
    else if (delta_t<86400)
        strm << delta_t/3600.0 << " h";
    else if (delta_t<86400*365.0)
        strm << delta_t/86400.0 << " d";
    else 
        strm << delta_t/(86400.0*365) << " a";
    return strm.str();
}
// ---------------------------------------------------------------------------

void replace_path_placeholders(std::string& str,
	const std::string& install_dir, const std::string& project_dir)
{
	string line;
    stringstream strm(str);
    str.clear();
    while (getline(strm,line)) {
        size_t pos = line.find("$(INSTALL_DIR)");
        if (pos != string::npos)
            line = line.substr(0, pos) + install_dir + line.substr(pos+15,string::npos);
    	pos = line.find("$(PROJECT_DIR)");
        if (pos != string::npos)
            line = line.substr(0, pos) + project_dir + line.substr(pos+15,string::npos);
        str += line + "\n";
	}
    // remove trailing \n
    if (!str.empty())
    	str.erase(str.end()-1, str.end());
}
// ---------------------------------------------------------------------------

void break_string(const std::string& msg, std::vector<std::string>& lines,
	int line_width)
{
	// first tokenize the string
    list<string> tokens;
    string delims;
	explode(msg, tokens, delims, " \n");
   	string current_line;
   	lines.clear();
    string::const_iterator delit = delims.begin();
    for (list<string>::iterator it = tokens.begin(); it != tokens.end() && delit != delims.end(); ++it, ++delit) {
    	if (current_line.size() + it->size() + 1 > static_cast<unsigned int>(line_width)) {
        	lines.push_back( current_line );
    		current_line.clear();
        }
        current_line += *it;
        if (*delit == '\n') {
        	lines.push_back( current_line );
    		current_line.clear();
        }
        else
        	current_line += ' ';
    }
   	lines.push_back(current_line); // if current_line is empty, msg was appended with a \n
}
// ---------------------------------------------------------------------------

// utility function to parse error/progress/debug message strings
void parse_msg_string(const std::string& msg, std::string& fname,
	std::vector<std::string>& lines, int line_width, bool include_fname)
{
	// first attempt to extract the file name
	size_t pos1 = msg.find("[");
	size_t pos2 = msg.find("]");
    string pure_msg;
    fname.clear();
	if (pos1 == 0 && pos2 != string::npos && pos1 < pos2) {
		size_t pos3 = msg.find_first_not_of(" \t", pos2+1);
		if (pos3 != string::npos) {
			fname = msg.substr(0, pos3);
            pure_msg = msg.substr(pos3, string::npos);
		}
    }
    if (fname.empty())  	pure_msg = msg;
    // now fname contains the function name, e.g. "[parse_string]"
    // and pure_msg contains the message itself.

    // now we correct the line_width if the fname should be included
    if (!fname.empty()) {
    	line_width = std::max<int>(static_cast<int>(fname.size()) + 20, line_width - static_cast<int>(fname.size()) - 2);
    }

    // now break the pure msg into substrings
    break_string(pure_msg, lines, line_width);

    if (include_fname && !fname.empty()) {
        for (unsigned int i=0; i<lines.size(); ++i) {
        	// if line is empty do not prepend name
            if (i != lines.size()-1 || !lines[i].empty())
        		lines[i] = fname + "  " + lines[i];
        }
    }
}
// ---------------------------------------------------------------------------

int count_tokens(const std::string& str) {
	stringstream strm(str);
    string dummy;
    int count = 0;
    while (strm >> dummy) ++count;
    return count;
}
// ---------------------------------------------------------------------------

}  // namespace IBK

