// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
// 
// Copyright (C) 2003 Andreas Nicolai, Andreas.Nicolai@gmx.net
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#ifndef IBK_stringutils_h
#define IBK_stringutils_h

#include <string>
#include <sstream>
#include <iomanip>
#include <stdexcept>
#include <list>
#include <vector>
#include <typeinfo>

#include "IBK_linearspline.h"
#include "IBK_exception.h"

namespace IBK {

/// Converts the value 'val' to a string.
template <class T>
std::string val2string(const T val) {
    std::stringstream strm;
    strm << val;
    return strm.str();
}

/// Converts the value 'val' to a string using special format qualifiers (total string width and fill chars)
template <class T>
std::string val2string(const T val, const std::size_t width, const char fillchar) {
    std::stringstream strm;
    strm.fill(fillchar);
    strm << std::setw(width) << val;
    return strm.str();
}

/// Attempts to extract a numerical value from a string
template <class T>
T string2val(const std::string& str) {
    T val;
    std::stringstream strm(str);
    if (!(strm >> val))
        throw IBK::IBK_exception("Could not convert '"+str+"' into '"+typeid(T).name()+"'!","[string2val]");
    return val;
}

/// Tries to read a vector of numbers of type T out of a string. 
/// If the number of read values is less then 2, a runtime error is thrown.
template <class T>
void string2vector(const std::string& data, std::vector<T>& vec) {
    std::stringstream strm(data);
    double para;
    vec.clear();
    while (strm >> para)
        vec.push_back(para);
    if (vec.size() < 2)  throw IBK::IBK_exception("Couldn't read a vector of '"+
        std::string(typeid(T).name())+"' from string '"+data+"'!","[string2vector]");
}

/// Splits a string containing a full file name like "/home/user1/file.txt" into the 
/// path /home/user1/ and the filename "file.txt".
void split_filename(std::string fullname, std::string& name, std::string& dir);

/// Splits a line like " keyword = value" into "keyword" and "value".
/// Returns 'true' if successful, or 'false' if not. In this case value==line
bool extract_keyword(const std::string& line, std::string& keyword, std::string& value);

/// Trims the string on both sides (removes tabs and spaces)
std::string& trim(std::string& str);
/// Removes comments from the str (removes everything including the //)
std::string& remove_comment(std::string& str);

/// Compares two string case insensitive
bool string_nocase_compare(const std::string& lhs, const std::string& rhs);

/// Finds the string 'substr' in the string 'str' (case insensitive).
/// Returns either the position of the substr or string::npos.
std::size_t string_nocase_find(const std::string& str, const std::string& substr);

/// Returns the relative path of the file 'filename' to the relative path 'relpath'.
/// Throws an exception if an empty file path or empty relative path has been passed.
std::string relative_path(const std::string& filename, const std::string& relpath);

/// Removes the extension from a filename and returns the filename without extension
std::string remove_extension(const std::string& filename);

/// Explodes the string 'str' into substrings (stored in the string list
/// 'tokens'), which were separated by 'delim' characters.
void explode(const std::string& str, std::list<std::string>& tokens, char delim);

/// Explodes the string 'str' into substrings (stored in the string list
/// 'tokens'), which were separated by any of the characters in the delimiter string.
/// The actual delimeters are stored in the string delims.
void explode(const std::string& str, std::list<std::string>& tokens,
	std::string& delims, const std::string& delimiter);

/// Explodes the CSV (comma-separated-values) string 'str' into substrings stored
/// in the string list 'tokens'.
void explode_csv(const std::string& str, std::list<std::string>& tokens);

/// Explodes a string using sequential keywords/section titles.
void explode_sections(const std::string& str, const std::vector<std::string>& section_titles,
	std::vector<std::string>& section_data);

/// Explodes a string using the same keyword/section title.
void explode_section(const std::string& str, const std::string& section_title,
	std::vector<std::string>& section_data);

/// Explodes a string into several sections where each section is stored in a single line
/// Essentially this function simply splits up a string into several lines which
/// are stored in the vector section_data.
void explode_section(const std::string& data, std::vector<std::string>& section_data);

/// Returns a string with the formatted double number. 
/// Pass either ios_base::scientific or ios_base::fixed as optional third argument
/// to specify fixed or exponetial format.
std::string format_double(double val, int precision, std::ios_base::fmtflags fmt = std::ios_base::floatfield);

/// Formats a time duration given in seconds into a string of the
/// format: xxxx days, xx hours, xx minutes and xx seconds.
/// If the second parameter 'shortformat' is 'true', the duration will be
/// printed in short format, that is: xxxx d  xx h  xx min  xx s
std::string format_time_duration(unsigned int t, bool shortformat = false);

/// Converts a relative time point in seconds into the output unit 'ustr' and
/// returns a completely formatted string
std::string format_time_difference(const std::string& ustr, double delta_t);

/// Creates a suitable output string for a time difference by automatically
/// converting the value delta_t (in seconds).
std::string format_time_difference(double delta_t);

/// Replaces the strings $(INSTALL_DIR) and $(PROJECT_DIR) in the string 'str'
/// with the appropriate strings.
void replace_path_placeholders(std::string& str,
	const std::string& install_dir, const std::string& project_dir);

/// Breaks the string 'msg' appart at the spaces and creates lines with
/// max 'line_width' characters.
void break_string(const std::string& msg, std::vector<std::string>& lines, int line_width);

/// Parses a message string of format '[myfunc]  my message text' into
/// the function name and several lines of max line_width chars.
void parse_msg_string(const std::string& msg, std::string& fname,
	std::vector<std::string>& lines, int line_width, bool include_fname);

/// Counts the number of whitespace separated tokens in the string
int count_tokens(const std::string& str);

}   // namespace IBK

/// @file IBK_stringutils.h
/// @brief Contains utility functions for string operations.

#endif  // IBK_stringutils_h
