// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
// 
// Copyright (C) 2003 Andreas Nicolai, Andreas.Nicolai@gmx.net
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#include "IBK_configuration.h"

#ifdef __BORLANDC__
  #include "IBK_prehead.h"
  #pragma hdrstop
#else
  #include <iostream>
  #include <iomanip>
  #include <sstream>
  #include <stdexcept>
#endif
#include "IBK_time.h"
#include "IBK_stringutils.h"  // for trim() and string2val
#include "IBK_messages.h"
#include "IBK_exception.h"
using namespace std;

namespace IBK {

const unsigned int SECS_PER_DAY = 86400;

void Time::set(unsigned long int sec) {
    while (sec>=SECS_PER_DAY)
        sec-=SECS_PER_DAY;
    sec_ = sec;
}
// ---------------------------------------------------------------------------

void Time::set(unsigned int hour, unsigned int min, unsigned int sec) {
    if (hour>23)    throw IBK_exception("Hours > 23!","[Time::set]");
    if (min>59)     throw IBK_exception("Minutes > 59!","[Time::set]");
    if (sec>59)     throw IBK_exception("Seconds > 59!","[Time::set]");
    sec_ = hour*3600 + min*60 + sec;
}
// ---------------------------------------------------------------------------

void Time::set(const std::string& timestr) {
    string str = timestr;
    for (string::iterator it=str.begin(); it!=str.end(); ++it)
        if (*it==':')    *it=' ';
    stringstream strm(str);
    unsigned int hour, min, sec;
    if (!(strm >> hour >> min))
        throw IBK_exception("Invalid time string '" + timestr + "'!","[Time::set]");
    if (!(strm >> sec))
        sec = 0;
    set(hour, min, sec);
}
// ---------------------------------------------------------------------------

Time& Time::operator+=(unsigned int secs) {
    sec_ += secs;
    while (sec_>= SECS_PER_DAY)
        sec_ -= SECS_PER_DAY;
    return *this;
}
// ---------------------------------------------------------------------------

Time& Time::operator-=(unsigned int secs) {
    while (sec_ < secs)
        sec_ += SECS_PER_DAY;
    sec_ -= secs;
    return *this;
}
// ---------------------------------------------------------------------------

const string& Time::as_string() const {
    unsigned int h = sec_ / 3600;
    unsigned int m = sec_ / 60 - h*60;
    unsigned int s = sec_ % 60;
    static string str(8,':');
    str[0] = (h >= 10) ? static_cast<char>(h / 10 + '0') : ' ';
    str[1] = static_cast<char>(h % 10 + '0');
    str[3] = static_cast<char>(m / 10 + '0');
    str[4] = static_cast<char>(m % 10 + '0');
    str[6] = static_cast<char>(s / 10 + '0');
    str[7] = static_cast<char>(s % 10 + '0');
    return str;
}
// ---------------------------------------------------------------------------

std::ostream& operator<<(std::ostream& out, const Time& t) {
    return out << t.as_string();
}
// ---------------------------------------------------------------------------

std::istream& operator>>(std::istream& in, Time& t) {
    string tmp;
    if (in >> tmp) {
        try { t.set(tmp); }
        catch (...) { in.setstate(std::ios_base::badbit); }
    }
    return in;
}
// ---------------------------------------------------------------------------

} // namespace IBK
