// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
// 
// Copyright (C) 2003 Andreas Nicolai, Andreas.Nicolai@gmx.net
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#ifndef IBK_time_h
#define IBK_time_h

#include <string>
#include <iosfwd>   // only include the iostream prototypes

namespace IBK {

/// The class Time represents a certain time point or a time difference.
class Time {
  public:
    /// Constructor
    Time() : sec_(0) {}
    /// Constructor taking the seconds since midnight.
    explicit Time(unsigned long int sec) { set(sec); };
    /// Constructor taking the time in hours, mins and secons
    /// Throws an exception if time greater then 23:59:59.
    Time(unsigned int hour, unsigned int min, unsigned int sec) { set(hour, min, sec); };
    /// Constructor taking a string containing the time.
    Time(const std::string& str) { set(str); };

    /// Sets the time from seconds since midnight.
    void set(unsigned long int sec);
    /// Sets the time.
    void set(unsigned int hour, unsigned int min, unsigned int sec);
    /// Sets the time from time string, for instance 12:40:12 or 12:30
    void set(const std::string& full_time);

    /// Adds 'secs' seconds to the current time (time will be wrapped at a full day).
    Time& operator+=(unsigned int secs);
    /// Subtracts 'secs' seconds from the current time (time will be wrapped at a full day).
    Time& operator-=(unsigned int secs);

    /// Adds the time 'rhs' to the current time (time will be wrapped at a full day).
    Time& operator+=(const Time& rhs) { return *this += rhs.sec_; };
    /// Subtracts the time 'rhs' from the current time (time will be wrapped at a full day).
    Time& operator-=(const Time& rhs) { return *this -= rhs.sec_; };

    /// Returns the current hour (0 to 23)
    unsigned int hour() const;
    /// Returns the current minute of the hour (0 to 59)
    unsigned int min() const;
    /// Returns the current second of the minute (0 to 59)
    unsigned int seconds() const;
    /// Returns the current time as string, for instance 12:30:25.
    const std::string& as_string() const;

  private:
    unsigned int sec_;   ///< the seconds since 00:00 (midnight)

    friend Time operator+(const Time& lhs, const Time& rhs);
    friend Time operator-(const Time& lhs, const Time& rhs);
    friend bool operator>(const Time& lhs, const Time& rhs);
    friend bool operator<(const Time& lhs, const Time& rhs);
    friend bool operator==(const Time& lhs, const Time& rhs);
    friend bool operator!=(const Time& lhs, const Time& rhs);
};

/// Writes the time into the stream 'out' using the format hh:mm:ss.
std::ostream& operator<<(std::ostream& out, const Time& t);
/// Reads the time from the stream 'in' using the format hh:mm:ss or hh:mm.
std::istream& operator>>(std::istream& in, Time& t);

inline Time operator+(const Time& lhs, const Time& rhs) { return Time(lhs.sec_)+=rhs; };
inline Time operator-(const Time& lhs, const Time& rhs) { return Time(lhs.sec_)-=rhs; };

inline bool operator==(const Time& lhs, const Time& rhs) { return lhs.sec_ == rhs.sec_; };
inline bool operator!=(const Time& lhs, const Time& rhs) { return lhs.sec_ != rhs.sec_; };
inline bool operator>(const Time& lhs, const Time& rhs) { return lhs.sec_ > rhs.sec_; };
inline bool operator<(const Time& lhs, const Time& rhs) { return lhs.sec_ < rhs.sec_; };
inline bool operator>=(const Time& lhs, const Time& rhs) { return !(lhs < rhs); };
inline bool operator<=(const Time& lhs, const Time& rhs) { return !(lhs > rhs); };

/// @file IBK_time.h
/// @brief Contains the declaration of the class Time.

/// @example time_example.cpp
/// This small example demonstrates the use of the class IBK::Time.

} // namespace IBK

#endif // IBK_time_h
