// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
// 
// Copyright (C) 2003 Andreas Nicolai, Andreas.Nicolai@gmx.net
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#ifndef IBK_tridiagmatrix_h
#define IBK_tridiagmatrix_h

#include "IBK_configuration.h"

#include <vector>
#include <stdexcept>
#include "IBK_matrix.h"
#include "IBK_format.h"
#include "IBK_message_strings.h"
#include "IBK_exception.h"

namespace IBK {

/// The template class tridiag_matrix stores the data of a tridiagonal matrix while allowing easy and
/// convenient access to the elements.
/// The class provides iterator support which allows to store the tridiagonal matrix data (consisting of
/// three linear bands) in a single linear container, such as the std::vector.
template <typename T>
class tridiag_matrix : public IBK::matrix<T> {
  public:
    typedef T                                           value_type;         ///< The type of the values.
    typedef T&                                          reference;          ///< Reference to one element.
    typedef const T&                                    const_reference;    ///< Constant reference to one element.
    typedef typename std::vector<T>::iterator           iterator;           ///< Iterator to one element.
    typedef typename std::vector<T>::const_iterator     const_iterator;     ///< Constant iterator to one element.

    /// Default constructor (creates an empty matrix)
    tridiag_matrix() : size_(0) {};
    /// Constructor (creates a matrix of dimension n)
    explicit tridiag_matrix(std::size_t n) : size_(n), data_(3*n) {};
    /// Constructor (creates a matrix of dimension n and initialises all elements with the value 'value')
    tridiag_matrix(std::size_t n, const T& value) : size_(n), data_(3*n, value) {};
	/// Assignment operator
	const tridiag_matrix<T>& operator=(const tridiag_matrix<T>& rhs) {
		if (size_ == rhs.size_) 
			copy(rhs.data_.begin(), rhs.data_.end(), data_.begin());
		else {
			data_ = rhs.data_;
			size_ = rhs.size_;
		}
		return *this;
	}

    /// Returns an iterator to the first element
    iterator        begin()         { return data_.begin(); };
    /// Returns a constant iterator to the first element
    const_iterator  begin() const   { return data_.begin(); };
    /// Returns an iterator to the element after the last
    iterator        end()           { return data_.end(); };
    /// Returns a constant iterator to the element after the last
    const_iterator  end() const     { return data_.end(); };


#ifdef IBK_ENABLE_RANGE_CHECKING
    /// Returns a reference to the element at the coordinates [col,row].
    /// This function throws a std::IBK_exception exception when the indices 
    /// are not part of the three bands and therefore invalid for a tridiagonal
    /// matrix.
    reference  operator() (std::size_t col, std::size_t row);

    /// Returns a constant reference to the element at the coordinates [col,row].
    /// Other than the non-constant operator() this function does not throw an 
    /// exception when an element is requested, that is not part of the 
    /// three bands. It will return a constant reference to a static variable 
    /// containing the default value of the template type T. Therefore
    /// a tridiag matrix can accessed (read-only) using normal matrix 
    /// access functions.
    const_reference  operator() (std::size_t col, std::size_t row) const;

#else // IBK_ENABLE_RANGE_CHECKING

    /// Returns a reference to the element at the coordinates [col,row].
    /// If IBK_ENABLE_RANGE_CHECKING is defined, this function throws 
    /// a std::IBK_exception exception when the indices 
    /// are not part of the three bands and therefore invalid for a tridiagonal
    /// matrix.
	reference  operator() (std::size_t col, std::size_t row) {
	    return data_[row + size_*(col-row+1)];
	}

    /// Returns a constant reference to the element at the coordinates [col,row].
	const_reference  operator() (std::size_t col, std::size_t row) const {
	    return data_[row + size_*(col-row+1)];
	}

#endif // IBK_ENABLE_RANGE_CHECKING

    /// Returns a reference to the element at the coordinates [col,row].
    /// If the coordinates are invalid (outside the three bands) a runtime error
    /// exception will be thrown.
    reference  at(std::size_t col, std::size_t row);
    /// Returns a constant reference to the element at the coordinates [col,row].
    /// If the coordinates are invalid (outside the three bands) a runtime error
    /// exception will be thrown.
    const_reference  at(std::size_t col, std::size_t row) const;

    /// Returns an iterator to the first element in the lower diagonal (which 
    /// has the coordinates (-1, 0) and is therefore actually not part of the
    /// matrix.
    iterator            lower()         { return data_.begin(); };
    /// Returns a constant iterator to the first element in the lower diagonal (which 
    /// has the coordinates (-1, 0) and is therefore actually not part of the
    /// matrix.
    const_iterator      lower() const   { return data_.begin(); };
    /// Returns an iterator to the first element in the main diagonal.
    iterator            main()          { return data_.begin()+size_; };
    /// Returns a constant iterator to the first element in the main diagonal.
    const_iterator      main()  const   { return data_.begin()+size_; };
    /// Returns an iterator to the first element in the upper diagonal.
    iterator            upper()         { return data_.begin()+2*size_; };
    /// Returns a constant iterator to the first element in the upper diagonal.
    const_iterator      upper() const   { return data_.begin()+2*size_; };

	void set_lower(T value)  { std::fill(lower(), lower()+size_, value); }
	void set_main(T value)   { std::fill(main(),  main()+size_, value); }
	void set_upper(T value)  { std::fill(upper(), upper()+size_, value); }

    double *            direct_lower()  { return &data_[0]; }
    double *            direct_main()   { return &data_[0] +   size_; }
    double *            direct_upper()  { return &data_[0] + 2*size_; }

    const tridiag_matrix<T>& operator+=(T value);
    const tridiag_matrix<T>& operator-=(T value);
    const tridiag_matrix<T>& operator*=(T value);
    const tridiag_matrix<T>& operator/=(T value);

    /// Resizes the matrix (all data will be lost).
    void resize(std::size_t n) { data_.resize(3*n); size_=n; };
    /// Returns the number of rows in the matrix.
    std::size_t  rows() const { return size_; };
    /// Returns the number of columns in the matrix.
    std::size_t  cols() const { return size_; };
    /// Clears the matrix.
    void clear() { data_.clear(); size_=0; }

    /// Adds the matrix 'other' to this matrix.
    tridiag_matrix<T>& operator+=(const tridiag_matrix<T>& other) {
        std::transform(data_.begin(), data_.end(), other.begin(), 
                       data_.begin(), std::plus<double>() );
        return *this;
    };

    /// Subtracts the matrix 'other' from this matrix.
    tridiag_matrix<T>& operator-=(const tridiag_matrix<T>& other) {
        std::transform(data_.begin(), data_.end(), other.begin(), 
                       data_.begin(), std::minus<double>() );
        return *this;
    };

    /// Solves the linear system of equations A*x = rhs and stores the resultant x vector 
    /// in the place of vector rhs.
    /// In this solver routine the tridiagonal matrix is changed because this algorithm
    /// works in-place. The advantage is, that is does not require additional workspace
    /// and is the fastest code (on single processor machines) without using vectorization 
    /// (which is not supported for C++ compilers anyway, yet).
    /// @return The return codes are as follows: <ul>
    /// <li>Return value = 0 : All ok</li>
    /// <li>Return value = 1 : Error - Zero element in main diagonal</li>
    /// <li>Return value = 2 : Error - Matrix is singular or becomes singular by transformation</li>
    /// </ul>
    unsigned int solve_equation_system(double * rhs);
    /// Convenience function, essentially behaves like the above function (make sure the
    /// vector has the right size!)
    unsigned int solve_equation_system(std::vector<double>& rhs) {
        return solve_equation_system(&rhs.front());
    };

  private:
    // Return value for all elements that are not part of the three bands of the matrix.
    static T  zero_element_;
    // order of the matrix 
    std::size_t size_;
    // contains the data of the three diagonals in the following 
    // order: main diagonale (m_size), lower band (m_size) first not used, upper band (m_size) last not used
    std::vector<T> data_;
};
// ---------------------------------------------------------------------------

template <typename T>
T tridiag_matrix<T>::zero_element_ = T();    // static member for a zero element

#ifdef IBK_ENABLE_RANGE_CHECKING

template <typename T>
typename tridiag_matrix<T>::reference tridiag_matrix<T>::operator() (std::size_t col, std::size_t row) {
    if (col>=size_ || row>=size_)
		throw IBK::IBK_exception(format_string(IBK_MATRIX_INDEX_OUT_OF_RANGE) % col % row % size_ % size_,
			"[IBK::tridiag_matrix::operator() ]");
    int diff=static_cast<int>(col);
    diff-=static_cast<int>(row);
    if (diff==-1)   return data_[row];
    if (diff==0)    return data_[row + size_];
    if (diff==1)    return data_[row + 2*size_];
	throw IBK::IBK_exception(format_string(IBK_MATRIX_INDEX_OUT_OF_RANGE) % col % row % size_ % size_,
		"[IBK::tridiag_matrix::operator() ]");
}
// ---------------------------------------------------------------------------

template <typename T>
typename tridiag_matrix<T>::const_reference tridiag_matrix<T>::operator() (std::size_t col, std::size_t row) const {
    if (col>=size_ || row>=size_)
        throw IBK::IBK_exception(
            (format_string(IBK_MATRIX_INDEX_OUT_OF_RANGE) 
			% static_cast<unsigned int>(col) 
			% static_cast<unsigned int>(row)
			% static_cast<unsigned int>(size_) 
			% static_cast<unsigned int>(size_)),
			"[IBK::tridiag_matrix::operator() const]");
    int diff=static_cast<int>(col);
    diff-=static_cast<int>(row);
    if (diff==-1)   return data_[row];
    if (diff==0)    return data_[row + size_];
    if (diff==1)    return data_[row + 2*size_];
    return zero_element_;
}
// ---------------------------------------------------------------------------

#endif // IBK_ENABLE_RANGE_CHECKING

template <typename T>
inline typename tridiag_matrix<T>::reference tridiag_matrix<T>::at(std::size_t col, std::size_t row) {
    if (col>=size_ || row>=size_)
        throw IBK::IBK_exception(format_string(IBK_MATRIX_INDEX_OUT_OF_RANGE) % col % row % size_ % size_,
		"[IBK::tridiag_matrix::at()]");
    int diff=static_cast<int>(col);
    diff-=static_cast<int>(row);
    if (diff==-1)   return data_[row];
    if (diff==0)    return data_[row + size_];
    if (diff==1)    return data_[row + 2*size_];
    throw IBK::IBK_exception(format_string(IBK_MATRIX_INDEX_OUT_OF_RANGE) % col % row % size_ % size_,
		"[IBK::tridiag_matrix::at()]");
}
// ---------------------------------------------------------------------------

template <typename T>
inline typename tridiag_matrix<T>::const_reference tridiag_matrix<T>::at(std::size_t col, std::size_t row) const {
    if (col>=size_ || row>=size_)
        throw IBK::IBK_exception(format_string(IBK_MATRIX_INDEX_OUT_OF_RANGE) % col % row % size_ % size_,
			"[IBK::tridiag_matrix::at() const]");
    int diff=static_cast<int>(col);
    diff-=static_cast<int>(row);
    if (diff==-1)   return data_[row];
    if (diff==0)    return data_[row + size_];
    if (diff==1)    return data_[row + 2*size_];
    return zero_element_;
}
// ---------------------------------------------------------------------------

template <>
unsigned int tridiag_matrix<double>::solve_equation_system(double * rhs);

// ---------------------------------------------------------------------------

template <class T>
const tridiag_matrix<T>& tridiag_matrix<T>::operator+=(T value) {
	iterator it=begin();
	const_iterator eit=end();
	while (it!=eit)
		*it++ += value;
	return *this;
}
// ---------------------------------------------------------------------------

template <class T>
const tridiag_matrix<T>& tridiag_matrix<T>::operator-=(T value) {
	iterator it=begin();
	const_iterator eit=end();
	while (it!=eit)
		*it++ -= value;
	return *this;
}
// ---------------------------------------------------------------------------

template <class T>
const tridiag_matrix<T>& tridiag_matrix<T>::operator*=(T value) {
	iterator it=begin();
	const_iterator eit=end();
	while (it!=eit)
		*it++ *= value;
	return *this;
}
// ---------------------------------------------------------------------------

template <class T>
const tridiag_matrix<T>& tridiag_matrix<T>::operator/=(T value) {
	if (value == 0) throw IBK::IBK_exception("Division by zero!","[IBK::tridiag_matrix::operator/=]");
	iterator it=begin();
	const_iterator eit=end();
	while (it!=eit)
		*it++ /= value;
	return *this;
}
// ---------------------------------------------------------------------------

} // namespace IBK

#endif // IBK_tridiagmatrix_h

/// @file IBK_tridiagmatrix.h
/// @brief Contains the class template tridiag_matrix for tridiagonal matrices.
/// @example tridiagmatrix_example.cpp
///          This small example demonstrates the use of the IBK::tridiag_matrix for 
///          solving an equation system.
