// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
// 
// Copyright (C) 2003 Andreas Nicolai, Andreas.Nicolai@gmx.net
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#include "IBK_configuration.h"

#ifdef __BORLANDC__
  #include "IBK_prehead.h"
  #pragma hdrstop
#endif
#include "IBK_undo.h"
using namespace std;

namespace IBK {

UndoListManager *UndoList = NULL;

// Destructor, free allocated memory of the undo actions
UndoListManager::~UndoListManager() {
    clear_undo();
    clear_redo();
}

// Adds a new undo action
bool UndoListManager::add(UndoAction* ptr) {
    if (ptr==NULL) return false;
    if (busy) {     // if an undo action is currently executed, do NOT add another.
        delete ptr;
        return false;
    };
    clear_redo();  // adding of new undo actions clears redo list.
    undolist_.push_back(ptr);
    while( undolist_.size() >= max_undo_) {
        delete undolist_.front();
        undolist_.pop_front();  
    }
    return true;
}

// Executes last (current) undo action.
void UndoListManager::undo_last() {
    if ( undolist_.size() > 0) {
        busy=true;
        ActionList::iterator it=undolist_.end()-1;
        // In case this undo action can be reverted, push it to the redo list
        // or free its memory otherwise
        if ((*it)->undo())
            redolist_.push_back(*it);
        else
            delete *it;
        undolist_.pop_back();
        busy=false;
    }
}

// Reverts the last undo action (Redo).
void UndoListManager::redo_last() {
    if ( redolist_.size() > 0) {
        busy=true;
        ActionList::iterator it=redolist_.end()-1;
        (*it)->redo();             // redo changes
        undolist_.push_back(*it);  // add action to undolist
        redolist_.pop_back();      // remove action from redolist
        busy=false;
    }
}

// Sets max number of undo steps
void UndoListManager::set_maxundo(size_t num) {
    if (num<1) return; // max_undo_ has to be at least 1
    max_undo_=num;
    while( undolist_.size() >= max_undo_) {
        delete undolist_[0];
        undolist_.pop_front();
    }
}

// Returns the descriptive message of the current undo action.
std::string UndoListManager::undo_message() const {
    if (undolist_.empty())  return "";
    else return undolist_.back()->m_description;
}

// Returns the descriptive message of the current redo action.
std::string UndoListManager::redo_message() const {
    if (redolist_.empty())  return "";
    else return redolist_.back()->m_description;
}


//************************** PRIVATE *********************************************
void UndoListManager::clear_undo() {
    for (ActionList::iterator it=undolist_.begin(); it!=undolist_.end(); ++it)
        delete *it;
    undolist_.clear();
}

void UndoListManager::clear_redo() {
    for (ActionList::iterator it=redolist_.begin(); it!=redolist_.end(); ++it)
        delete *it;
    redolist_.clear();
}

}  // namespace IBK

