// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
// 
// Copyright (C) 2003 Andreas Nicolai, Andreas.Nicolai@gmx.net
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#ifndef IBK_undo_h
#define IBK_undo_h

#include <deque>
#include <string>

namespace IBK {

/// The class UndoAction is the abstract base class for all undo/redo actions.
/// When deriving a class from UndoAction you have to implement the member 
/// functions undo() and redo(). The member function undo() has to return a 
/// boolean indicating, whether the action (that has been undone by the 
/// function) can be redone using redo().
class UndoAction {
  public:
    /// Default Constructor.
    UndoAction() {};
    /// Constructor, takes an optional string describing the undo action.
    explicit UndoAction(const std::string& desc) : m_description(desc) {};
    /// Destructor, so that the descructor of the derived class will be called.
    virtual ~UndoAction() {};
    /// Implement this member function in your own specialised undo/redo class.
    virtual bool undo() = 0;
    /// Implement this member function in your own specialised undo/redo class.
    virtual void redo() = 0;
    
    std::string m_description;  ///< Explainatory text of the undo action.
};
// ****************************************************************************


/// The class UndoListManager encapsulated the undo- und redo-lists and the 
/// undo/redo-functionality.
/// This class is the major framework of the undo-system. It stores the 
/// undo/redo classes and is very convenient to be build into graphical user
/// interfaces.
class UndoListManager {
  public:
    /// Default constructor
    UndoListManager() : max_undo_(20), busy(false) {};
    /// Destructor, frees the allocated space of the undo/redo classes
    ~UndoListManager();

    /// Adds a new undo-action to the list.
    /// If there are more then max_undo() items in the last the oldest 
    /// undo-action will be removed from the list.
    /// Whenever a new undo action will be added to the list, the redo-list
    /// will be emptied.
    /// @param ptr A base class pointer to a undo/redo-class derived from 
    ///            UndoAction.
    /// @return Returns 'true' if the undo action has been added successfully, or
    ///         'false' if not (this may happen when a undo-action is currently 
    ///         being executed).
    bool add(UndoAction* ptr);

    /// Executes the last undo action. 
    /// If that undo action returns 'true' it will be added to the redo-list.
    void undo_last();

    /// Executes the last redo action.
    /// The executed redo-action will be added to the undo-list so that it can be
    /// undone again.
    void redo_last();

    /// Sets the maximum number of undo actions in the list.
    void set_maxundo(std::size_t num);
    /// Returns the maximum number of undo actions in the list.
    std::size_t get_maxundo() { return max_undo_; };
    /// Sets the current number of undo actions in the list.
    std::size_t undo_size() const { return undolist_.size(); };
    /// Sets the current number of redo actions in the list.
    std::size_t redo_size() const { return redolist_.size(); };
    /// Clears both the undo and redo list.
    void clear() { clear_undo(); clear_redo(); };

    /// Returns the descriptive message of the current undo action 
    /// (or an empty string).
    std::string undo_message() const;
    /// Returns the descriptive message of the current redo action
    /// (or an empty string).
    std::string redo_message() const;

  private:
    typedef std::deque<UndoAction*> ActionList;

    ActionList undolist_;   ///< the undo list
    ActionList redolist_;   ///< the redo list

    void clear_undo();
    void clear_redo();

    size_t max_undo_;       ///< the maximum number of undo steps

    bool busy;  // stops the function add_undo() from adding new undo actions
                // during the execution of an undo/redo action
};

/// Pointer to the global undo list. 
/// This is a pointer to allow convenient access to the undo list from
/// any place of inside your program. Make sure you initialise this pointer
/// before you start using it!
extern UndoListManager *UndoList;

} // namespace IBK


/// @file IBK_undo.h
/// @brief Contains the declarations of the classes UndoAction and UndoListManager.

/// @example undo_example.cpp
/// This example demonstrates the use of UndoList and how UndoListManager are written.

#endif  // IBK_undo_h
