// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
// 
// Copyright (C) 2003 Andreas Nicolai, Andreas.Nicolai@gmx.net
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#ifndef IBK_unit_h
#define IBK_unit_h

#include <string>
#include <iosfwd>

namespace IBK {

/// The class Unit represents a scientific unit.
/// A Unit object represents a scientific unit, such as 'g/mm3'. The units can
/// be accessed either with their unique ID-number or their name. 
/// Every unit that can be related to another unit has a SI base unit. That base
/// unit's id-number can be retrieved via the member function base_id().
/// @warning Never use Unit before you initialised the unit list!
class Unit {
  public:
    /// Default constructor (creates an undefined unit).
    Unit() : id_(0) {};
    /// Constructor (creates a unit with the id number 'unitid').
    /// This function throws an out_of_range exception if the 'unitid' is invalid.
    explicit Unit(unsigned int unitid);
    /// Constructor (creates a unit with the name 'name').
    /// This function throws a IBK_exception exception if the unit 'name' is not in the unitlist.
    explicit Unit(const std::string& name);
    /// Changes the unit to the unit with id number 'unitid' (same as 'unit = CUnit(unitid)').
    /// This function throws an out_of_range exception if the 'unitid' is invalid.
    void set(unsigned int unitid);
    /// Changes the unit to the unit 'name'.
    /// This function throws a IBK_exception exception if the unit 'name' is not in the unitlist.
    void set(const std::string& name);
	/// Clears the unit (resets it to an undefined state).
	void clear() { id_ = 0; };
    /// Returns the id number of the unit.
    unsigned int id() const { return id_; };
    /// Returns the name of the unit.
    const std::string& name() const;
    /// Returns the id number of the matching SI base unit.
    /// The base unit will be returned as id number so that comparing the 
    /// base-id-numbers of two different units will yield whether the
    /// units are related to each other or not.
    unsigned int base_id() const;
    /// Retrieves the calculation factor/summand and the operation necessary to relate a value
    /// of the unit to the unit given by 'unitid'.
    void relate_to(unsigned int unitid, double& fact, unsigned int& op);
    /// Retrieves the calculation factor/summand and the operation necessary to relate a value
    /// of the unit to the unit given by 'name'.
    void relate_to(const std::string& name, double& fact, unsigned int& op);
    /// Retrieves the calculation factor/summand and the operation necessary to relate a value
    /// of the unit to the unit given by 'unit'.
    void relate_to(Unit unit, double& fact, unsigned int& op);
    /// Returns true when the unit is undefined.
    bool operator!() { return id_==0; }

  private:
    unsigned int id_;

    friend bool operator==(const Unit& lhs, const Unit& rhs);
    friend bool operator!=(const Unit& lhs, const Unit& rhs);
};

/// Comparison operator== tests, whether two units are equal.
inline bool operator==(const Unit& lhs, const Unit& rhs) { return lhs.id_==rhs.id_; }
/// Comparison operator!= tests, whether two units are not equal.
inline bool operator!=(const Unit& lhs, const Unit& rhs) { return lhs.id_!=rhs.id_; }
/// Input stream operator>> reads a unit via name from the stream 'in'.
std::istream& operator>>(std::istream& in, Unit& unit);
/// Output stream operator<< writes a unit via name into the stream 'out'.
std::ostream& operator<<(std::ostream& out, const Unit& unit);

}  // namespace IBK

/// @file IBK_unit.h
/// @brief Contains the declaration of class Unit, a data type that represents a scientific unit.

/// @example unit_example.cpp
/// This small example demonstrates the use of the classes IBK::Unit and IBK::UnitList.

#endif  // IBK_unit_h
