// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
// 
// Copyright (C) 2003 Andreas Nicolai, Andreas.Nicolai@gmx.net
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#ifndef IBK_unitdata_h
#define IBK_unitdata_h

#include <string>
#include <iosfwd>

namespace IBK {

/// The class UnitData contains the information for a single unit.
/// This class is internally used by the UnitList and shouldn't be
/// used directly.
class UnitData {
  public:
    /// Default constructor, creates an undefined unit.
    UnitData() : id_(0), name_("undefined"), base_id_(0), factor_(0), operation_(0)  {}

    /// Initialisation constructor.
    UnitData(const unsigned int unitid, const std::string& name,
              const unsigned int baseunit, double fact, const unsigned int op)
      :  id_(unitid), name_(name), base_id_(baseunit), factor_(fact), operation_(op) {}

    /// Returns the ID number of the unit.
    const unsigned int  id()          const     { return id_; };
    /// Returns the unit string (or name).
    const std::string&  name()        const     { return name_; };
    /// Returns the ID of the appropriate base unit.
    const unsigned int  base_id()     const     { return base_id_; };
    /// Returns the factor or summand for conversion into the base unit.
    const double        factor()      const     { return factor_; };
    /// Returns the operation for converting the unit into the base unit.
    const unsigned int  operation()   const     { return operation_; };

  private:
    unsigned int id_;           ///< ID number.
    std::string  name_;         ///< String constant, unit name.
    unsigned int base_id_;      ///< Assigned base unit it.
    double       factor_;       ///< Conversion factor/summand.
    unsigned int operation_;    ///< Conversion operation.

    /// Returns true, if two units are equal (have the same ID number).
    friend bool operator==(const UnitData& lhs, const UnitData& rhs);
    /// Returns true, if two units are not equal (do not have the same ID number).
    friend bool operator!=(const UnitData& lhs, const UnitData& rhs);
};
//*****************************************************************************

inline bool operator==(const UnitData& lhs, const UnitData& rhs) {
                        return (lhs.id_==rhs.id_); }
inline bool operator!=(const UnitData& lhs, const UnitData& rhs) {
                        return (lhs.id_!=rhs.id_); }

}  // namespace IBK

/// @file IBK_unitdata.h
/// @brief Contains declaration of class UnitData.

#endif  // IBK_unitdata_h
