// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
//
// Copyright (C) 2003 Andreas Nicolai, Andreas.Nicolai@gmx.net
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#include "IBK_configuration.h"

#ifdef __BORLANDC__
  #include "IBK_prehead.h"
  #pragma hdrstop
#else
  #include <iomanip>
  #include <iostream>
  #include <stdexcept>
  #include <fstream>
  #include <sstream>
  #include <algorithm>
  #include <cmath>
#endif
#include "IBK_unitlistmanager.h"
#include "IBK_unitdata.h"
#include "IBK_unit.h"
#include "IBK_stringutils.h"
#include "IBK_messages.h"
#include "IBK_math.h"
#include "IBK_exception.h"
using namespace std;

namespace IBK {

UnitListManager* UnitList=NULL;

/*****************************************************************************
 * Public methods
 *****************************************************************************/

// Reads the unit list from a file
bool UnitListManager::read_file(const std::string& filename, bool overwrite) {
    ifstream in(filename.c_str());
  	return read(in, overwrite);
}
// ---------------------------------------------------------------------------

// Reads the unitlist from an input filestream
bool UnitListManager::read(std::istream& stream, bool overwrite) {
    if (!overwrite && !empty())  return true; // do not read the list again
    if (!stream)  return false;

    unsigned int base_index=0, current_index=0;
    clear();                        // Clear list

    while (stream.good()) {
        stringstream ins;           // Input-String-Stream
        string str;                 // Unit string
        getline(stream,str);
        ins << str;                 // fill string stream
        char op=' ';                // conversion operator
        unsigned int op_id;         // enum value of operator
        double factor;              // conversion factor

        // read first unit in line
        ins >> str;
        if (ins.good()) {
            // add base unit to the list (no operation
            add(new UnitData(current_index, str, base_index, 1.0, OP_NONE));

            // Try reading the next unit
            ins >> op >> factor >> str;
            while (!ins.fail()) {
                current_index++;
                switch (op) {
                    case '+' : op_id = OP_ADD; break;
                    case '-' : op_id = OP_SUB; break;
                    case '*' : op_id = OP_MUL; break;
                    case '/' : op_id = OP_DIV; break;
                    case '%' : op_id = OP_SPECIAL; break;
                    default  : op_id = OP_NONE;
                }
                add(new UnitData(current_index, str, base_index, factor, op_id));
                ins >> op >> factor >> str;
            }
            current_index++;          // next unit
            base_index=current_index; // is also the base unit
        }
    }
    if (empty()) return false;
//        throw IBK_exception( "[UnitListManager::read]  Could not read any units!");
    return true;
}
// ---------------------------------------------------------------------------

bool UnitListManager::read(const std::string& str, bool overwrite) {
    string tmp(str);
    replace(tmp.begin(), tmp.end(), ';', '\n');
    stringstream lstrm(tmp);
    return read(lstrm, overwrite);
};
// ---------------------------------------------------------------------------

bool UnitListManager::write(const std::string& filename) const {
    if (empty()) return false;
//        throw IBK_exception( "[UnitListManager::write]  Unit list is empty!");
    ofstream out(filename.c_str());
    // write first base unit
    out << setw(10) << left << list_.front()->name() << "   ";
    // remember last base unit index
    unsigned int last_base=0;
    const_iterator endit=end();
    const_iterator it=begin();
    ++it;
    for (; it!=endit; ++it) {
        // get next unit data
        UnitData* unit=*it;
        // do we have a new base unit?
        if (unit->base_id()!=last_base) {
            // new line and new base unit
            out << ";" << endl << setw(10) << left << unit->name() << "   ";
            last_base=unit->base_id();
        } else {
            stringstream ustream;
            switch (unit->operation()) {
                case OP_ADD : ustream << "+ "; break;
                case OP_SUB : ustream << "- "; break;
                case OP_MUL : ustream << "* "; break;
                case OP_DIV : ustream << "/ "; break;
                case OP_SPECIAL : ustream << "% "; break;
                default     : ustream << "| "; break;
            }
            ustream << unit->factor() << " " << unit->name() << " ";
            out << setw(20) << left << ustream.str();
        }
    }
    out << ";" << endl;    // finish last line
    return out.good();
//    if (!out)        throw IBK_exception( "[UnitListManager::write]  Error writing unit list to file '"+filename+"'!");
}
// ---------------------------------------------------------------------------

bool UnitListManager::convertible_units(std::vector<Unit>& list, const Unit& src,
                                  const bool remove) const
{
    unsigned int baseid( src.base_id() );
    list.clear();           // empty list
    unsigned int id=src.id();
    const_iterator endit=end();
    const_iterator it=begin();
    while (it!=endit) {
        // get pointer to current unit and increase pointer afterwards
        const UnitData* unit( *it++ );
        // 1. test whether base units are the same
        if (unit->base_id()==baseid) {
            // 2. get unit id
            unsigned int uid=unit->id();
            // 3. if units to not match and remove is == true, do NOT add unit
            if (remove && uid==id) continue;
            // 4. add unit to the list
            list.push_back( Unit(uid) );
        };
    };
    if (list.empty()) return false;
    return true;
}
// ---------------------------------------------------------------------------

bool UnitListManager::base_units(std::vector<Unit>& list) {
    if (empty())  return false;
    list.clear();
    const_iterator endit=end();
    const_iterator it=begin();
    while (it!=endit) {
        // get pointer to current unit and increase pointer afterwards
        const UnitData* unit( *it++ );
        if (unit->id()==unit->base_id())
            list.push_back( Unit(unit->id()) );
    };
    return true;
}
// ---------------------------------------------------------------------------

const UnitData* UnitListManager::retrieve(unsigned int unitid) const {
    if (unitid >= size())
        throw out_of_range("[UnitListManager::retrieve]  Unit index "+val2string(unitid)+
            " out of range 0.."+val2string(static_cast<int>(size())-1)+"!");
    return operator[](unitid);
}
// ---------------------------------------------------------------------------

const UnitData* UnitListManager::retrieve(const string &str) const {
    const_iterator endit=end();
    const_iterator it=begin();
    while (it!=endit) {
        const UnitData* unit( *it );
        if ( unit->name()==str)  return unit;
        ++it;
    };
    return NULL;
}
// ---------------------------------------------------------------------------

// Calculates conversion factor and operation to relate the src and target units
void UnitListManager::relate_units(const UnitData* src, const UnitData* target,
                             double& fact, unsigned int& op) const
{
    // Are the units the same?
    if (src->id()==target->id())  {
        fact=1;
        op=OP_NONE;
    }
    // is it actually possible to convert the units?
    if (src->base_id()!=target->base_id())
        throw IBK_exception("Can't relate unit ["+
            src->name()+"] to ["+target->name()+ "]!","[UnitListManager::relate_units]");
    // check if src unit or target unit require special operations
    if (src->operation()==OP_SPECIAL || target->operation()==OP_SPECIAL) {
    	fact = 1;
        op = OP_SPECIAL;
        return;
    }
    // check if src unit is already a base unit (no conversion necessary)
    if (src->operation()==OP_NONE) {
        // return only + or -
        // NOTE: if we convert backward, we need to get the opposite conversion operation
        switch (target->operation()) {
            case OP_ADD :
                op=OP_ADD;
                fact=target->factor();
                break;

            case OP_SUB :
                op=OP_ADD;
                fact=-target->factor();
                return;

            case OP_MUL :
                op=OP_MUL;
                fact=target->factor();
                return;

            case OP_DIV :
                op=OP_MUL;
                if (target->factor()==0)
                    throw IBK_exception("Div by zero! (tried to convert ["+src->name()+
						"] to ["+target->name()+ "] )!","[UnitListManager::relate_units]");
                fact=1.0/target->factor();
                return;

            default :
                fact=1;
                op=OP_NONE;
                return;
        }
    }

    // test + or - operation
    if (src->operation()==OP_ADD || src->operation()==OP_SUB) {
        // DO NOT modifiy the member variable, create copy of variable.
        double own_factor=src->factor();
        if (src->operation()==OP_ADD)
            own_factor=-own_factor;
        // calculate conversion factors
        switch (target->operation()) {
            case OP_ADD :
                op=OP_ADD;
                fact=own_factor+target->factor();
                return;

            case OP_SUB :
                op=OP_ADD;
                fact=own_factor-target->factor();
                return;

            case OP_MUL :    // both multiplication and division aren't allowed
            case OP_DIV :
                throw IBK_exception("* and / operations "
                    "not allowed when using + and - operations! "
                    "(tried to convert ["+src->name()+"] to ["+target->name()+ "] )!",
					"[UnitListManager::relate_units]");

            default :
                fact=src->factor();
                op=OP_ADD;
                return;
        }
    }
    // test * or / operation
    if (src->operation()==OP_MUL || src->operation()==OP_DIV) {
        double own_factor=src->factor();
        // convert to base unit
        if (src->operation()==OP_MUL) {
            if (own_factor==0)
                throw IBK_exception("Div by zero! "
                    "(tried to convert ["+src->name()+"] to ["+target->name()+ "] )!",
					"[UnitListManager::relate_units]");
            own_factor=1.0/own_factor;
        }
        // Calculate conversion factors
        switch (target->operation()) {
            case OP_ADD : // addition and subtraction are not allowed in combination with
            case OP_SUB : // * and /
                throw IBK_exception("+ and - operations "
                    "not allowed when using * and / operations! "
                    "(tried to convert ["+src->name()+"] to ["+target->name()+ "] )!",
					"[UnitListManager::relate_units]");

            case OP_MUL :
                op=OP_MUL;
                fact=own_factor*target->factor();
                return;

            case OP_DIV :
                op=OP_MUL;
                if (target->factor()==0)
                    throw IBK_exception("Div by zero! "
                        "(tried to convert ["+src->name()+"] to ["+target->name()+ "] )!",
						"[UnitListManager::relate_units]");
                fact=own_factor/target->factor();
                return;

            default :
                fact=own_factor;
                op=OP_MUL;
                return;
        }
    }
}
// ---------------------------------------------------------------------------

void UnitListManager::convert(Unit src, Unit target, double& val) {
     const UnitData* s( retrieve(src.id()) );
     const UnitData* t( retrieve(target.id()) );
     convert(s,t,val);
}
// ---------------------------------------------------------------------------

void UnitListManager::convert(const UnitData* src, const UnitData* target, double& val) {
    double fact;
    unsigned int op;
    relate_units(src, target, fact, op);
    if (op==OP_ADD) {
        val+=fact;
    }
    else if (op==OP_MUL) {
        val*=fact;
    }
    else if (op==OP_SPECIAL) {
    	convert_special(src, target, val);
    }
}
// ---------------------------------------------------------------------------

void UnitListManager::convert_special(const UnitData* src, const UnitData* target, double& val) {
    // first convert value from src unit to base unit, store result in base_val
    const UnitData* base( retrieve(src->base_id()) );
    double base_val = val;
    if (src->operation()==OP_SPECIAL) {
    	if (src->name() == "sqrt(s)") {
        	base_val = val*val;        	          // from sqrt(s) -> s
        }
        else if (src->name() == "sqrt(h)") {
        	base_val = val*val*3600.0;	          // from sqrt(h) -> s
        }
    	else if (src->name() == "log(kg/m3)") {
        	base_val = IBK::f_pow10(val);         // from log(kg/m3) -> kg/m3
        }
    	else if (src->name() == "log(g/m3)") {
        	base_val = IBK::f_pow10(val)/1e+3;    // from log(g/m3) -> kg/m3
        }
    	else if (src->name() == "log(mg/m3)") {
        	base_val = IBK::f_pow10(val)/1e+6;    // from log(mg/m3) -> kg/m3
        }
    	else if (src->name() == "log(g/m3)") {
        	base_val = IBK::f_pow10(val)/1e+9;    // from log(g/m3) -> kg/m3
        }
    }
    else {
         convert(src, base, base_val);
    }

    // then convert from base unit to target unit
    if (target->operation()==OP_SPECIAL) {
    	if (target->name() == "sqrt(s)") {
        	if (base_val < 0)
            	throw IBK_exception("Trying to take square root of negative number!",
					"[UnitListManager::convert_special]");
        	val = std::sqrt(base_val);            // from s -> sqrt(s)
        }
        else if (target->name() == "sqrt(h)") {
        	if (base_val < 0)
            	throw IBK_exception("Trying to take square root of negative number!",
					"[UnitListManager::convert_special]");
        	val = sqrt(base_val/3600.0);          // from s -> sqrt(h)
        }
    	else if (target->name() == "log(kg/m3)") {
        	if (base_val <= 0)
            	throw IBK_exception("Trying to take logarithm of zero or negative number!",
					"[UnitListManager::convert_special]");
        	val = IBK::f_log10(base_val);         // from kg/m3 -> log(kg/m3)
        }
    	else if (target->name() == "log(g/m3)") {
        	if (base_val <= 0)
            	throw IBK_exception("Trying to take logarithm of zero or negative number!",
					"[UnitListManager::convert_special]");
        	val = IBK::f_log10(base_val*1e+3);    // from kg/m3 -> log(g/m3)
        }
    	else if (target->name() == "log(mg/m3)") {
        	if (base_val <= 0)
            	throw IBK_exception("Trying to take logarithm of zero or negative number!",
					"[UnitListManager::convert_special]");
        	val = IBK::f_log10(base_val*1e+6);    // from kg/m3 -> log(mg/m3)
        }
    	else if (target->name() == "log(g/m3)") {
        	if (base_val <= 0)
            	throw IBK_exception("Trying to take logarithm of zero or negative number!",
					"[UnitListManager::convert_special]");
        	val = IBK::f_log10(base_val*1e+9);    // from kg/m3 -> log(g/m3)
        }
    }
    else {
    	val = base_val;
         convert(base, target, val);
    }
}
// ---------------------------------------------------------------------------


// The units defined per default in the unitlist
const char * const DEFAULT_UNITS =
"undefined    ;"
"m            * 1e+03 mm          * 1e+02 cm          * 1e+01 dm       ;"
"m2           * 1e+06 mm2         * 1e+04 cm2         * 1e+02 dm2      ;"
"m3           * 1e+09 mm3         * 1e+06 cm3         * 1e+03 dm3      ;"
"s            / 60 min            / 3600 h            / 86400 d           / 3.1536e+07 a      % 0       sqrt(s)   % 0       sqrt(h)  ;"
"m/s          * 100 cm/s          * 360000 cm/h       * 8.64e+06 cm/d     ;"
"m2/s         * 10000 cm2/s       * 3600 m2/h         * 3.6e+07 cm2/h     ;"
"m3/s         ;"
"m/s2         ;"
"s/m          ;"
"s2/m2        ;"
"kg           * 1000 g            * 1e+06 mg          ;"
"kg/ms        ;"
"kg/s         * 3600 kg/h         * 86400 kg/d        * 8.64e+07 g/d      * 1e6 mg/s      * 1e9 g/s    ;"
"kg/m2        / 100 kg/dm2        * 10 g/dm2          / 10 g/cm2          * 1e6 mg/m2     ;"
"kg/m2s       * 1000 g/m2s        * 3.6e+06 g/m2h     * 3600 kg/m2h       * 1e6 mg/m2s    * 1e9 g/m2s  * 3.6e9 mg/m2h    * 3.6e12 g/m2h ;"
"kg/m3        / 1000 kg/dm3       * 1 g/dm3           / 1000 g/cm3        * 1000 g/m3     * 1e6 mg/m3   * 1e9 g/m3       % 0      log(kg/m3)    % 0    log(g/m3)    % 0    log(mg/m3)    % 0    log(g/m3);"
"kg/m3s       * 1000 g/m3s        * 3.6e+06 g/m3h     * 3600 kg/m3h       * 1e6 mg/m3s    * 1e9 g/m3s  * 3.6e9 mg/m3h    * 3.6e12 g/m3h ;"
"kg/m         * 1000 g/m          * 1 g/mm            / 1000 kg/mm        ;"
"kg/kg        * 1000 g/kg         * 1e6 mg/kg         ;"
"J            / 1000 kJ           / 1e+06 MJ          ;"
"J/m2         / 1000 kJ/m2        / 1e+06 MJ/m2       / 1e+09 GJ/m2       / 100 J/dm2         / 10000 J/cm2       ;"
"J/m3         / 1000 kJ/m3        / 1e+06 MJ/m3       / 1000 J/dm3        / 1e+06 J/cm3       ;"
"J/m3s        / 1000 kJ/m3s       / 1e+06 MJ/m3s      / 1000 J/dm3s       / 1e+06 J/cm3s       ;"
"J/m3K        ;"
"J/s          * 3600 J/h          * 86400 J/d         * 86.4 kJ/d         ;"
"J/kg         / 1000 kJ/kg        ;"
"J/kgK        / 1000 kJ/kgK       ;"
"W            / 1000 kW           / 1e+06 MW          ;"
"W/m2         * 1    J/m2s        / 1000 kW/m2        / 1e+06 MW/m2       / 100 W/dm2         / 10000 W/cm2       ;"
"W/m3         / 1000 kW/m3        / 1e+06 MW/m3       / 1000 W/dm3        / 1e+06 W/cm3       / 1e+09 W/mm3       ;"
"W/mK         / 1000 kW/mK        ;"
"W/m2K        ;"
"W/mK2        ;"
"l/m2s        * 3600 l/m2h        * 86400 l/m2d       * 86400 mm/d        ;"
"l/m3s        * 3600 l/m3h        ;"
"m3/m2s       * 3600 m3/m2h       * 1000 dm3/m2s      * 3.6e+06 dm3/m2h   ;"
"m3/m3        * 100 Vol%          ;"
"m3/m3d       * 100 Vol%/d        ;"
"---          * 100 %             ;"
"---/d        * 100 %/d           ;"
"K            - 273.15 C          ;"
"1/K          ;"
"K/m          ;"
"m2s/kg       ;"
"Pa           / 100 hPa           / 1000 kPa          ;"
"1/Pa         ;"
"Pa/m         / 1000 kPa/m        ;"
"Lux          / 1000 kLux         ;"
"Rad          * 57.2958 Deg       ;"
"m2/kg        ;"
"m2K/W        ;"
"1/m          / 100 1/cm          ;"
"logcm        ;"
"logm         ;"
"logPa        ;"
"K/Pa         ;"
"mol/kg       / 1000 mol/g        ;"
"kg/mol       * 1000 g/mol        ;"
"J/mol        / 1000 kJ/mol       ;"
"kg/m2s05     * 60 kg/m2h05       ;"
"1/logcm      ;"
"mol/m3       / 1000 mol/ltr      / 1000 mol/dm3      / 1e+06 mol/cm3     ;"
"mol          * 1e+03 mmol        ;"
"-            ;"
"mm/mK        ;"
"mm/m         ;"
"1/s          * 60 1/min          * 3600 1/h ;";

}  // namespace IBK

