// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
// 
// Copyright (C) 2003 Andreas Nicolai, Andreas.Nicolai@gmx.net
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#ifndef IBK_unitlistmanager_h
#define IBK_unitlistmanager_h

#include <string>
#include <vector>
#include <iosfwd>
#include <iostream>
#include "IBK_ptrlist.h"
#include "IBK_unitdata.h"

namespace IBK {

class Unit;

/// The class UnitListManager is the central class of the unit system.
/// You can use multiple unit lists in the program, however, it is recommanded that
/// you only create a single unit list and make use of the global unit list pointer
/// (create an instance of UnitList on the heap and assign UnitList to it).
/// The unit list can be initialised using either of the two read() functions. A unit
/// list can be written to a file using the write function. The member function 
/// convertible_units() can be used to create a vector of units that can be converted in
/// or from each other. The member functions relate_units() and convert() are useful
/// if you want to implement a container, such as UnitVector.
class UnitListManager : public ptr_list<UnitData> {
  public:
    /// Reads the unit list from the file 'filename'.
    /// The unit list will be usually read from a file, but can be read from a stream
    /// (e.g. stringstream) also. Take a look at the example unit list file for info
    /// about the format.
    /// If the second parameter is 'true', the unit list will be erased before new
    /// units are read from the file. Otherwise read units are just added to the list.
    /// The function throws an exception if a syntax error is encountered in the unit list file.
    /// If the unit list cannot be found or opened, the function simply returns 'false' and 
    /// does not throw an exception.
    /// If the unit list can be read without problems the function returns 'true'.
    bool read_file(const std::string& filename, bool overwrite=false);

    /// Reads the unit list from a stream.
    /// Works essentially like the function above, only that the unit list data is expected to
    /// be in the input stream 'stream'.
    bool read(std::istream& stream, bool overwrite=false);

    /// Reads the unitlist from a string.
    /// Works essentially like the function above, only that the unit list data is expected to
    /// be in the string (wherein each line is separated by a ';'). 
    bool read(const std::string& str, bool overwrite=false);

    /// Writes the unit list into the file 'filename'.
    /// This function creates a new file with the name 'filename' and writes the unit list into 
    /// that file. Returns 'false' if writing fails.
    bool write(const std::string& filename) const;

    /// Creates a list of units, that can be converted in the unit given by 'src'.
    /// The function clears the vector (first argument) and add all units to it, that
    /// can be converted into 'src'. If the last (optional) parameter is true, the unit
    /// 'src' will not be contained in the vector (per default the unit 'src' will be 
    /// included in the vector).
    /// @return Returns 'false' if the vector is empty (meaning no convertible units 
    ///         could be found), otherwise 'true'. If the parameter 'remove' is false 
    ///         or not given the function will always return true.
    bool convertible_units(std::vector<Unit>& list, const Unit& src, 
                           const bool remove=false) const;

    /// Creates a vector with all base units contained in the unit list.
    bool base_units(std::vector<Unit>& list);

    /// Determines the conversion operation and the factor/summand for converting from 
    /// unit 'src' to 'target'. 
    /// Throws an exceptions if the unit 'src' cannot be converted to 'target'.
    void relate_units(const UnitData* src, const UnitData* target,
                      double& fact, unsigned int& op) const;

    /// Converts the value 'val' from unit 'src' to unit 'target'.
    void convert(Unit src, Unit target, double& val);

    /// Operations returned by relate_units().
    enum UnitOperations {
        OP_NONE,
        OP_ADD,
        OP_SUB,
        OP_MUL,
        OP_DIV,
        OP_SPECIAL,
        OP_END
    };

  private:
    void convert(const UnitData* src, const UnitData* target, double& val);
    void convert_special(const UnitData* src, const UnitData* target, double& val);

	const UnitData* retrieve(unsigned int unitid) const;
    const UnitData* retrieve(const std::string& str) const;

    friend class Unit;
};
//*****************************************************************************

extern UnitListManager * UnitList;			///< Global pointer to the unit list.

extern const char * const DEFAULT_UNITS;	///< Contains list of default units.

}  // namespace IBK 

/// @file IBK_unitlistmanager.h
/// @brief Contains declaration of class UnitList.

#endif  // IBK_unitlistmanager_h
