// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
// 
// Copyright (C) 2003 Andreas Nicolai, Andreas.Nicolai@gmx.net
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#include "IBK_configuration.h"

#ifdef __BORLANDC__
  #include "IBK_prehead.h"
  #pragma hdrstop
#else
  #include <iostream>
  #include <iomanip>
  #include <iterator>
  #include <sstream>
#endif
#include "IBK_unitvector.h"
#include "IBK_unitlistmanager.h"
#include "IBK_messages.h"
#include "IBK_format.h"
#include "IBK_message_strings.h"
#include "IBK_math.h"
#include "IBK_exception.h"
using namespace std;

namespace IBK {

void UnitVector::read(const std::string& str) {
    stringstream strm(str);
    m_data.clear();
    // read the data
    std::copy(istream_iterator<double>(strm), istream_iterator<double>(), back_inserter(m_data) );
    if (m_data.empty()) 
        throw IBK_exception( "Error reading vector data!","[UnitVector::read]");
    // reset stream state to read unit
    strm.clear();
    string ustr;
    strm >> ustr;
    if (ustr.empty())
        throw IBK_exception( "Error reading vector unit!","[UnitVector::read]" );
    try {  m_unit = Unit(ustr);  }
    catch (...) {
        throw IBK_exception( "Invalid vector unit '"+ustr+"'!","[UnitVector::read]" );
    };
    convert( m_unit, Unit(m_unit.base_id()) );   // convert data to base unit
}
// ---------------------------------------------------------------------------

void UnitVector::write(std::ostream& out, unsigned int indent) const {
    try {
        // create temporary vector
        UnitVector tmp(*this);
        // convert from base unit to IO-unit
        tmp.convert( Unit(m_unit.base_id()), m_unit );
        // write data
        out << string(indent,' ');
        out.precision(8);
        std::copy(tmp.m_data.begin(), tmp.m_data.end(), ostream_iterator<double>(out, " "));
        out << m_unit.name() << endl;
    }
    catch (std::exception& ex) {
        MSG(ex.what(), MSG_ERROR, "[UnitVector::write]");
        throw IBK_exception("Error writing unit vector!","[UnitVector::write]");
    }
}
// ---------------------------------------------------------------------------

std::string UnitVector::as_string() const {
    stringstream strm;
    write(strm,0);
    return strm.str();
}
// ---------------------------------------------------------------------------

void UnitVector::set(std::size_t n, double value, const std::string& ustr) {
    m_data.resize(n);
    std::fill(m_data.begin(), m_data.end(), value);
    m_unit.set(ustr);
}
// ---------------------------------------------------------------------------

void UnitVector::clear() {
    m_data.clear();
    m_unit.set(0);
}
// ---------------------------------------------------------------------------

void UnitVector::convert(Unit src_unit, Unit target_unit) {
    if (src_unit==target_unit)  return;       
    double fact;
    unsigned int op;
    src_unit.relate_to(target_unit, fact, op);
    if (op==UnitListManager::OP_ADD) {
        storage_member_type::iterator endIt = m_data.end();
        for (storage_member_type::iterator it=m_data.begin(); it!=endIt; ++it)
            *it += fact;
    }
    else if (op==UnitListManager::OP_MUL) {
        storage_member_type::iterator endIt = m_data.end();
        for (storage_member_type::iterator it=m_data.begin(); it!=endIt; ++it)
            *it *= fact;
    }
    else if (op==UnitListManager::OP_SPECIAL) {
        storage_member_type::iterator endIt = m_data.end();
        for (storage_member_type::iterator it=m_data.begin(); it!=endIt; ++it)
    	    convert_special(src_unit, target_unit, *it);
    }
}
// ---------------------------------------------------------------------------

void UnitVector::convert_special(Unit src_unit, Unit target_unit, double & val) {
    if (src_unit == Unit("log(kg/m3)")) {
        if (val <= -50) val = 1e-50;
        else val = IBK::f_pow10(val);         // from log(kg/m3) -> kg/m3
    }
    else if (src_unit == Unit("log(g/m3)")) {
        if (val <= -50) val = 1e-50;
        else val = IBK::f_pow10(val)/1e+3;    // from log(g/m3) -> kg/m3
    }
    else if (src_unit == Unit("log(mg/m3)")) {
        if (val <= -50) val = 1e-50;
        else val = IBK::f_pow10(val)/1e+6;    // from log(mg/m3) -> kg/m3
    }
    else if (src_unit == Unit("log(g/m3)")) {
        if (val <= -50) val = 1e-50;
        else val = IBK::f_pow10(val)/1e+9;    // from log(g/m3) -> kg/m3
    }

    if (target_unit == Unit("log(kg/m3)")) {
        if (val <= 1e-50) val = -50;
        else val = IBK::f_log10(val);         // from kg/m3 -> log(kg/m3)
    }
    else if (target_unit == Unit("log(g/m3)")) {
        if (val <= 1e-50) val = -50;
        else val = IBK::f_log10(val*1e+3);    // from kg/m3 -> log(g/m3)
    }
    else if (target_unit == Unit("log(mg/m3)")) {
        if (val <= 1e-50) val = -50;
        else val = IBK::f_log10(val*1e+6);    // from kg/m3 -> log(mg/m3)
    }
    else if (target_unit == Unit("log(g/m3)")) {
        if (val <= 1e-50) val = -50;
        else val = IBK::f_log10(val*1e+9);    // from kg/m3 -> log(g/m3)
    }
}
// ---------------------------------------------------------------------------

#ifdef IBK_ENABLE_RANGE_CHECKING
UnitVector::reference  UnitVector::operator[] (std::size_t i) {
	if (i > m_data.size())
        throw std::out_of_range("[IBK::UnitVector::operator[] ]  " + 
            format_string(IBK_INDEX_OUT_OF_RANGE) % i % 0 % (m_data.size()-1));
	return m_data[i];
}
// ---------------------------------------------------------------------------

UnitVector::const_reference  UnitVector::operator[] (std::size_t i) const {
	if (i > m_data.size())
        throw std::out_of_range("[IBK::UnitVector::operator[] ]  " +
            format_string(IBK_INDEX_OUT_OF_RANGE) % i % 0 % (m_data.size()-1));
	return m_data[i];
}
// ---------------------------------------------------------------------------

#endif // IBK_ENABLE_RANGE_CHECKING

}  // namespace IBK 
