// IBK Library - Collection of classes, functions and algorithms
//               for development of numeric simulations
// 
// Copyright (C) 2003 Andreas Nicolai, Andreas.Nicolai@gmx.net
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

#ifndef IBK_unitvector_h
#define IBK_unitvector_h

#include "IBK_configuration.h"

#include <iosfwd>
#include <string>
#include <vector>
#include "IBK_unit.h"

namespace IBK {

/// A data vector with associated unit.
/// The IBK::UnitVector encapsulates a double vector whose elements 
/// are physical values with an unit. In addition to that the vector
/// contains the input/output unit of the vector. The data inside the
/// vector is always in the base SI unit.
class UnitVector {
  public:
    typedef std::vector<double>                     storage_member_type;

    typedef double                                  value_type;
    typedef storage_member_type::iterator           iterator;
    typedef storage_member_type::const_iterator     const_iterator;
    typedef storage_member_type::reference          reference;
    typedef storage_member_type::const_reference    const_reference;

    // ****** member functions ***********************************************

    /// Default constructor.
    UnitVector() {};
    /// Constructor (creates a vector with 'size' elements).
    explicit UnitVector(std::size_t n)
        : m_data(n) {};
    /// Constructor (creates a vector with 'size' elements and the IO-unit 'ioUnit').
    UnitVector(std::size_t n, Unit IO_unit)
        : m_unit(IO_unit), m_data(n) {};
    /// Constructor (creates a vector with 'size' elements, initial values 'value' and the IO-unit 'ioUnit').
    UnitVector(std::size_t n, double value, Unit IO_unit)
        : m_unit(IO_unit), m_data(n, value) {};
    /// Constructor (creates a vector from the given range and the optional IO-unit 'IO_unit').
    template<typename inIt>
    UnitVector(inIt first, inIt last, Unit IO_unit=Unit())
        : m_unit(IO_unit), m_data(first, last) {};

    /// Reads a unit vector (data and unit) from a string.
    void read(const std::string& str);
    /// Writes a unit vector into the stream using 'indent' spaces as indentation.
    void write(std::ostream& out, unsigned int indent) const;
    /// Returns the unit vector written in a string.
    std::string as_string() const;
    /// Resizes the vector to 'n' elements, fill it with 'value' and sets the IO unit to 'ustr'.
    void set(std::size_t n, double value, const std::string& ustr);
    /// Resets the vector to its initial state.
    void clear();

    /// Converts the data in the vector from unit 'src_unit' to unit 'target_unit'.
    void convert(Unit src_unit, Unit target_unit);
    void convert_special(Unit src_unit, Unit target_unit, double & val);

    /// Returns the size of the vector (convenience function).
    std::size_t size() const { return m_data.size(); };
    /// Returns whether the vector is empty or not (convenience function).
    bool empty() const { return m_data.empty(); };

#ifndef IBK_ENABLE_RANGE_CHECKING
    /// Returns a reference to the vector element at index i.
    /// If IBK_ENABLE_RANGE_CHECKING is defined it throws an out_of_range exception if the 
    /// indices are invalid.
    reference  operator[] (std::size_t i) { return m_data[i]; };
    /// Returns a constant reference to the vector element at index i.
    /// If IBK_ENABLE_RANGE_CHECKING is defined it throws an out_of_range exception if the 
    /// indices are invalid.
    const_reference  operator[] (std::size_t i) const { return m_data[i]; };
#else
    /// Returns a reference to the vector element at index i.
    /// If IBK_ENABLE_RANGE_CHECKING is defined it throws an out_of_range exception if the 
    /// indices are invalid.
    reference  operator[] (std::size_t i);
    /// Returns a constant reference to the vector element at index i.
    /// If IBK_ENABLE_RANGE_CHECKING is defined it throws an out_of_range exception if the 
    /// indices are invalid.
    const_reference  operator[] (std::size_t i) const;
#endif // IBK_ENABLE_RANGE_CHECKING


    // ****** member variables ***********************************************
    Unit                    m_unit;     ///< The input/output unit of the vector.
    storage_member_type     m_data;     ///< Contains the actual data of the vector.
};


} // namespace IBK

/// @file IBK_unitvector.h
/// @brief Contains the class UnitVector.

#endif  // IBK_unitvector_h
